package AIspace.XMLReader;

import java.util.ArrayList;
import java.util.StringTokenizer;

/**
 * @author Joseph Roy Santos
 *
 * This class implements an n-ary tree structure for use with the XML reader
 * This still needs a lot of work - specifically encapsulation with regards to 
 * XMLBlock (the children nodes are not private for traversal). The XMLTree class
 * should be extended to fit the needs of the application that uses this.
 * Also, this is very intolerant of spaces between property-value pairs. 
 * Comments have not been implemented.
 * 
 * It is probable that the XMLTree will never read DTDs. This is why it is recommended
 * that the XMLTree be extended for specific uses.
 */
public class XMLTree {
	
	/** the head node of the tree. It is an XMLBlock
	 */
	protected XMLBlock headNode;
	
	/** constructor. Creates a new head node
	 */
	public XMLTree() {
		headNode = new XMLBlock();
	}
	
	/** function to read in a String which represents an XML file.
	 * @throws XMLParseException 
	 */
	public void readString(String input) throws XMLParseException {
		readString(stripComments(input), headNode); 	
	}

	/**
	 * Function to find a subtree in this XMLTree with the specified tag.
	 * Note that the tag is case-insensitive.
	 */
	public XMLBlock findNetworkTree(String tag) {
		ArrayList<XMLBlock> thisTree = new ArrayList<XMLBlock>();
		thisTree.add(headNode);
		return findNetworkTree(thisTree, tag);
	}

	/**
	 * Private helper function that finds a subtree in this XMLTree 
	 * with the specified tag. This does a depth-first search and returns the
	 * first tree that it encounters.
	 */
	private XMLBlock findNetworkTree(ArrayList<XMLBlock> currLevel, String tag) {
		// do a search on the current level of the tree. If the tag is found,
		// return that subtree. If not, recurse on that node's subtree.
		for (int i = 0; i < currLevel.size(); i++) {
			XMLBlock currBlock = currLevel.get(i);
			if (currBlock.getTag().equalsIgnoreCase(tag)) {
				return currBlock;	
			}
			XMLBlock returnBlock = findNetworkTree(currBlock.childBlocks, tag);
			if (returnBlock != null) {
				return returnBlock;	
			}
		}	
		return null;	
	}
	
	/**
	 * helper function that removes all comments, processing directives
	 * and DTDs from the input string
	 */
	private String stripComments(String s) {
		String input = s;
		// find and remove comments
		int commentStart;
		while ((commentStart = input.indexOf("<!--")) != -1) {
			input = input.substring(0,commentStart)+input.substring(input.indexOf("-->")+3);
		}
		// find and remove processing directives
		while ((commentStart = input.indexOf("<?")) != -1) {
			input = input.substring(0,commentStart)+input.substring(input.indexOf("?>")+2);
		}	
		// find and remove DTDs
		while ((commentStart = input.indexOf("<!DOCTYPE")) != -1) {
			input = input.substring(0,commentStart)+input.substring(input.indexOf("]>")+2);
		}
		
		// chop off all whitespace at the start & end of file
		return input.trim();	
	}
	
	/** private function to read in an XML string. This is called by readString
	 * 	The currentNode is the XMLBlock that is being parsed
	 *	Make sure that the string is a valid XML string (ie <TAG>some information</TAG>)
	 *	This calls itself recursively to build a tree structure.
	 *	This parser has a special case - it does not allow tags to start with a -. This
	 *	is to prevent CILog text from being read as a tag. For example, <-TAG> is not
	 *	allowed, but <TAG> is. 
	 * @throws Exception 
	 */
	private void readString(String input, XMLBlock currentNode) throws XMLParseException {
		// find the start of the first tag
		// ignore <-
		int	startOfTag = input.indexOf("<");
		int errorTag = input.indexOf("<-");
		while ((startOfTag == errorTag)) {
			if (startOfTag == -1) {
				return;	
			}
			startOfTag = input.indexOf("<", errorTag);		
			errorTag = input.indexOf("<-", errorTag);
		}
		
		// initialize the tag name
		int	endOfTagName = input.indexOf(" ");
		
		// finds the end of the tag name (<TAG> or <TAG someproperty = I>)
		// in this case, the tag name is TAG
		// input.indexOf(" ") > 0 makes sure that -1 is not allowed when there
		// are no spaces
		if ((input.indexOf(" ") < input.indexOf(">")) && (input.indexOf(" ") > 0)) {
			endOfTagName = input.indexOf(" ");
		} else {
			endOfTagName = input.indexOf(">");
		}
	
		// finds the end of the first tag
		int	endOfStartTag = input.indexOf(">");
		
		// gets the string that represents the first tag
		// for example, <HELLO someproperty = I> would give
		// a tag name of HELLO
		String tag = input.substring(startOfTag+1,endOfTagName); 
		
		//sets the currentNode tagName to the tag obtained above 
		currentNode.setTag(tag);
		
		// find the end of the whole XML "block"
		int endOfTag = input.indexOf("</"+tag+">");
		if(endOfTag == -1) {
			throw new XMLParseException("End of tag \"<" + tag + ">\" not found.");
		}
		
		// checks to see if there is any text between the open and close tags.
		// if there is, set the currentNode text to the text.
		// the check works by checking to see if there are additional tags
		// between the open and close tags (which would start with a "<")
		// It then compares this index to the endOfTag index. If they are the same,
		// then there is only text between the open and close tags 
		if ((input.indexOf("<",endOfStartTag+1) == endOfTag) || tag.equalsIgnoreCase("logicprogram")) {
			currentNode.setText((input.substring(endOfStartTag+1,endOfTag)));
		}
		
		// This checks whether or not there are properties specified for the tag
		// it does this by finding out if the end of the start tag is actually
		// the same as the end of the tag name. So <TAG> would be false, but 
		// <TAG someproperty=0> would be true. 
		if ((endOfStartTag-endOfTagName) > 0) {
			// the tokenizer is used to find all the properties.
			StringTokenizer startTag = new StringTokenizer (input.substring(endOfTagName+1,endOfStartTag).trim(), " \"=", true);
			String tok = startTag.nextToken();
		
			while (startTag.hasMoreTokens()) {
				String propertyName = "";
				String propertyValue ="";
				while (tok.equals(" ")) {
					tok = startTag.nextToken();	
				}
				propertyName = tok;
				tok = startTag.nextToken();
				while (tok.equals(" ") || tok.equals("=")) {
					tok = startTag.nextToken();	
				}	
				if (!tok.equals("\"")) {
					propertyValue = tok;
					currentNode.addProperty(propertyName, propertyValue);						
				} else {
					propertyValue = startTag.nextToken();
					tok = startTag.nextToken();
					while (!tok.equals("\"")) {
						propertyValue += tok;
						tok = startTag.nextToken();	
					}	
					currentNode.addProperty(propertyName, propertyValue);
				}
				if (startTag.hasMoreTokens()) {
					tok = startTag.nextToken();	
					while (tok.equals(" ") && startTag.hasMoreTokens()) {
						tok = startTag.nextToken();	
					}					
				} else {
					break;	
				}
			}
		}
		
		XMLBlock newChildNode;
		
		// chop off the open and close tags
		String remainingString = input.substring(endOfStartTag+1, endOfTag);
		
		// figure out if there are additional blocks between the open and close tags.
		// since remainingString does not have the first open and close tags, there
		// will only be additional blocks if there is greater than symbol (<).
		// the while loop then parses through each one, recursively calling readString,
		// and adds the created XMLBlock as a child node.
		while ((remainingString.indexOf("<")) != -1) {
			// find the first open tag
			int startOfNewTag = remainingString.indexOf("<");

			if (!tag.equalsIgnoreCase("logicprogram")) {	
				int endOfNewTagName = startOfNewTag;
			
				// finds the end of the tag name (<TAG> or <TAG someproperty = I>)
				// in this case, the tag name is TAG
				// input.indexOf(" ") > 0 makes sure that -1 is not allowed when there
				// are no spaces.
				// Note that the indexOf(" ") search starts from the open bracket.
				// This is to make sure that spaces before the bracket are ignored					
				if (((remainingString.indexOf(" ",startOfNewTag)) < (remainingString.indexOf(">", startOfNewTag))) 
						&& (remainingString.indexOf(" ",startOfNewTag) > 0)) {
					endOfNewTagName = remainingString.indexOf(" ", startOfNewTag);
				} else {
					endOfNewTagName = remainingString.indexOf(">", startOfNewTag);
				}
			
				// gets the string that represents the first tag
				// for example, <HELLO someproperty = I> would give
				// a tag name of HELLO			
				String newTag = remainingString.substring(startOfNewTag+1,endOfNewTagName);

				// find the end of the first tag
				int endOfNewTag = remainingString.indexOf("</"+newTag+">");
				if(endOfNewTag == -1) {
					throw new XMLParseException("End of tag \"<" + newTag + ">\" not found.");
				}
				// initialize a new XMLBlock
				newChildNode = new XMLBlock();
			
				// copy the first block into childString.
				// note that this includes the open and close tags
				String childString = remainingString.substring(startOfNewTag,endOfNewTag)+"</"+newTag+">";
			
				// remove this block from remainingString.
				// In this way, the while loop removes a single block everytime it loops.
				remainingString = remainingString.substring(endOfNewTag+("/"+newTag+">").length()+1);
			
				// recursively add the first block (childString) to newChildNode
				readString(childString, newChildNode);
				// add newChildNode (which now has data) into currentNode as a child node.
				currentNode.addChild(newChildNode);	
			} else {
				// create a lowercase copy of the remainingString string
				String copyOfRemainingString = new String(remainingString).toLowerCase();
				// chop off the logicprogram text 
				remainingString = remainingString.substring(copyOfRemainingString.indexOf("</logicprogram>")+15);			
			}
		}
			
		return;
	}
	
	/** function to print the contents of the tree into the console. Used
	 * for testing purposes only and is very crude.
	 */
	public void print() {
		printAll(headNode,0);
	}
	
	/** private helper function to print the contents of a Node and its children into 
	 * the console.
	 */
	private void printAll(XMLBlock currNode, int spaces) {
		String s = "";
		for (int i = 0; i < spaces; i++) {
			s += "	 ";	
		}
		System.out.println(s+"node tag: " + currNode.getTag());
		for (int i = 0; i < currNode.getProperties().size(); i++) {
			System.out.println(s+"Property: " + (currNode.getProperties().get(i)).name);	
		}
		System.out.println(s+"Text: " + currNode.getText());
		System.out.println(s+"Children: ");
		for (int i = 0; i < currNode.childBlocks.size(); i++) {
			printAll(currNode.childBlocks.get(i), spaces+1);
		}
	}
	
	/**
	 * returns the headNode
	 */
	public XMLBlock getHead() {
		return headNode;	
	}
}