/**
 * File: Graph.java
 * Jul 5, 2000
 */

package AIspace.graphToolKit;

import java.awt.Color;
import java.awt.Graphics;
import java.util.ArrayList;
import java.util.Iterator;

import AIspace.graphToolKit.elements.Edge;
import AIspace.graphToolKit.elements.EdgeLabel;
import AIspace.graphToolKit.elements.Entity;
import AIspace.graphToolKit.elements.Node;
import AIspace.graphToolKit.elements.Point;

/**
 * A general-purpose graph class.
 * @author Shinjiro Sueda
 */
public class Graph {
	public ArrayList<Node> nodes;				// ArrayList of nodes
	public ArrayList<Edge> edges;				// ArrayList of edges
	protected ArrayList<Node> selectedNodes;	// ArrayList of selected nodes
	protected ArrayList<Edge> selectedEdges;	// ArrayList of selected edges
	private float scale;				 	// "bigness" scale
	public GraphCanvas canvas;				// canvas containing this graph

	private int lineWidth;
	private float arrowWidth;
	private float arrowHeight;


	//
	// Methods
	//

	/**
	 * Default Constructor.
	 */
	public Graph( GraphCanvas canvas ) {
		this.canvas = canvas;
		nodes = new ArrayList<Node>();
		edges = new ArrayList<Edge>();
		selectedNodes = new ArrayList<Node>();
		selectedEdges = new ArrayList<Edge>();
		setScale(1);
		setLineWidth(1);
	}

	public void setLineWidth( int width ) {
		lineWidth = width;
		calculateArrowSize();
	}

	public int getLineWidth() {
		return lineWidth;
	}

	public float getArrowWidth() {
		return arrowWidth;
	}

	public float getArrowHeight() {
		return arrowHeight;
	}

	private void calculateArrowSize() {
		arrowWidth = lineWidth*2;
		if( arrowWidth < 6 ) arrowWidth = 6;
		if( arrowWidth > 15 ) arrowWidth = 15;
		arrowHeight = arrowWidth*2.5f;
	}

	public void showMessage( String title, String msg ) {
		canvas.showMessage( title, msg );
	}

	public boolean isInSolveMode() {
		return canvas.getMode() == GraphConsts.SOLVE;
	}

	/**
	 * Returns the rectangular bound of the graph.
	 */
	public Point[] getBound() {
		// If the number of nodes <= 1, return a bogus value
		if( numNodes() == 0 ) {
			Point[] tmp = {new Point(0, 0), new Point(0, 0)};
			return tmp;
		}

		Iterator<Node> nodeItr = nodes.iterator();
		Node curNode = nodeItr.next();
		float minx = curNode.pos.x - curNode.width/2;
		float miny = curNode.pos.y - curNode.height/2;
		float maxx = curNode.pos.x + curNode.width/2;
		float maxy = curNode.pos.y + curNode.height/2;

		while(nodeItr.hasNext()) {
			curNode = nodeItr.next();
			if( curNode.pos.x - curNode.width/2 < minx ) minx = curNode.pos.x - curNode.width/2;
			if( curNode.pos.x + curNode.width/2 > maxx ) maxx = curNode.pos.x + curNode.width/2;
			if( curNode.pos.y - curNode.height/2 < miny ) miny = curNode.pos.y - curNode.height/2;
			if( curNode.pos.y + curNode.height/2 > maxy ) maxy = curNode.pos.y + curNode.height/2;
		}

		Iterator<Edge> edgeItr = edges.iterator();
		Point[] bounds;
		while(edgeItr.hasNext()) {
			bounds = edgeItr.next().eLabel.getBounds();
			if( bounds != null ) {
				if( bounds[0].x < minx ) minx = bounds[0].x;
				if( bounds[0].y < miny ) miny = bounds[0].y;
				if( bounds[1].x > maxx ) maxx = bounds[1].x;
				if( bounds[1].y > maxy ) maxy = bounds[1].y;
			}
		}
		
		float pad = 30;
		Point[] tmp = {new Point( minx - pad, miny - pad ), new Point( maxx + pad, maxy + pad )};
		return tmp;
	}
	
	public final void setPromptLabel(String s) {
		canvas.setPromptLabel(s);
	}

	public final void setPromptColor(Color col) {
		canvas.setPromptColor(col);
	}

	public final String getPromptLabel() {
		return canvas.getPromptLabel();
	}

	/**
	 * Called when font is changed.
	 */
	public void updateNodeSize() {
		for(int i = 0; i < nodes.size(); ++i) {
			nodes.get(i).updateSize();
		}
	}

	public void updateEdgeSize() {
		for(int i = 0; i < edges.size(); ++i) {
			edges.get(i).updateLength();
		}
	}

	/**
	 * Updates a single node's size.
	 */
	public void updateNodeSize( Node n ) {
		n.updateSize();

		// update lengths of all connected edges.
		ArrayList<Edge>[] e = n.getAllEdges();
		Iterator<Edge> itr = e[0].iterator();
		while(itr.hasNext()) {
			itr.next().updateLength();
		}
		itr = e[1].iterator();
		while(itr.hasNext()) {
			itr.next().updateLength();
		}
	}

	public void resetLabels() {
		Iterator<Edge> itr = edges.iterator();
		while(itr.hasNext()) {
			itr.next().eLabel.reset();
		}
	}

	public void resetLabel( Edge e ) {
		e.eLabel.reset();
	}

	public void repaint() {
		 canvas.repaint();
	}

	/**
	 * Draws the components.
	 */
	public void draw( Graphics g, boolean moving ) {
		updateEdgeSize();
		// Draw the edges and edge labels
		Edge curEdge;
		for(int i = 0; i < edges.size(); i++) {
			curEdge = edges.get(i);
			curEdge.draw( g, moving );
			curEdge.eLabel.draw(g, moving);
		}

		// Draw the nodes
		for(int i = 0; i < nodes.size(); i++) {
			nodes.get(i).draw(g, moving);
		}
	}	
	
	public void drawStructure( Graphics g, boolean moving) {
		for(Edge e : edges) {
			e.updateLengthStructure();
			e.drawStructure(g, moving);
		}

		// Draw the nodes
		for(Node n : nodes) {
			n.drawStructure(g, moving);
		}
	}

	/**
	 * Deletes currently selected entities.
	 */
	public void deleteSelected() {
		// Remove the selected edges
		Iterator<Edge> edgeItr = selectedEdges.iterator();
		Edge curEdge;
		while(edgeItr.hasNext()) {
			curEdge = edgeItr.next();
			//Remove itself from all connected nodes
			curEdge.removeFromNodes();
			//remove from edges in graph
			edges.remove(curEdge);
		}
		selectedEdges.clear();

		Iterator<Node> nodeItr = selectedNodes.iterator();
		Node curNode;
		ArrayList[] allNodeEdges;
		//remove selected nodes
		while(nodeItr.hasNext()) {
			curNode = nodeItr.next();
			//get all connected edges
			allNodeEdges = curNode.getAllEdges();
			//removing edges in
			while (allNodeEdges[0].size() > 0) {
				curEdge = (Edge) allNodeEdges[0].get(0);
				//Remove itself from all connected nodes
				curEdge.removeFromNodes();
				//remove from edges in graph
				edges.remove(curEdge);
			}
			//removing edges out
			while (allNodeEdges[1].size() > 0) {
				curEdge = (Edge) allNodeEdges[1].get(0);
				//Remove itself from all connected nodes
				curEdge.removeFromNodes();
				//remove from edges in graph
				edges.remove(curEdge);
			}
			nodes.remove(curNode);
		}
		selectedNodes.clear();
	}


	/**
	 * adds a node to the graph at the next index
	 * @param newNode - the node to be added
	 */
	public void addNode( Node newNode ) {
		newNode.setIndex( getNextIndex() );
		nodes.add(newNode.index, newNode);
	}

	/**
	 * removes a node from the graph
	 * @param node - the node to be removed
	 */
	public void removeNode(Node node){
		nodes.remove(node);
	}

	/**
	 * Modifies a node.
	 */
	public void modifyNode( Node n, String l, Point p, Color c, int s ) {
		n.set( l, p, c, s );
	}

	/**
	 * Returns the node at the given index.
	 */
	public Node nodeAt( int index ) {
		return nodes.get( index );
	}


	/**
	 * Returns the number of nodes.
	 */
	public int numNodes() {
		return nodes.size();
	}

	/**
	 * Adds a new edge to the graph.
	 * @param label - the label for the node
	 * @param start - the starting node for the edge
	 * @param end - the ending node for the edge
	 * @param color - the color of the edge
	 * @param edgeType - the type of edge (DIRECTIONAL, etc)
	 */
	public void addEdge( String label, Node start, Node end, Color color, int edgeType ) {
		edges.add( new Edge( this, label, start, end, color, edgeType ) );
	}

	/**
	 * Adds a new edge to the graph.
	 */
	public void addEdge( Edge newEdge ) {
		edges.add( newEdge );
	}

	/**
	 * Modifies an edge.
	 */
	public void modifyEdge( Edge edge, String l, Node s, Node e, Color c, int t ) {
		edge.set( l, s, e, c, t );
	}

	/**
	 * Returns the edge at the given index.
	 */
	public Edge edgeAt( int index ) {
		return edges.get( index );
	}

	/**
	 * Returns the number of edges.
	 */
	public int numEdges() {
		return edges.size();
	}
	
	/**
	 * Moves a node and its edges so that it appears to be on top of other entities.
	 * Warning: this changes the order of nodes in the "nodes" collection.
	 * @param node the node to bring to the top.
	 * @author Byron Knoll
	 */
	public void moveNodeToTop (Node node)
	{
		// Only move the node to the top in create mode. This is to avoid problems caused by
		// changing node order while the applet is in solve mode.
		if (canvas.mode==GraphConsts.CREATE)
		{
			// Check whether the node is already on top.
			if (nodes.indexOf(node) != nodes.size()-1)
			{
				// Move the node to the end of the collection (so that it appears on top).
				nodes.remove(node);
				nodes.add(node);
				// Update the indexes of all the nodes.
				for (int count = 0; count < nodes.size(); count++)
				{
					nodes.get(count).setIndex(count);
				}
			}
			
			// Iterate through edges going into the node and bring them to the top.
			Iterator<Edge> itrIn = node.getAllEdges()[0].iterator();
			Edge curEdge;
			while(itrIn.hasNext()) 
			{
				curEdge = itrIn.next();
				if (edges.indexOf(curEdge) != edges.size()-1)
				{
					// Move the edge to the end of the collection (so that it appears on top).
					edges.remove(curEdge);
					edges.add(curEdge);
				}
			}
			
			// Iterate through edges going out of the node and bring them to the top.
			Iterator<Edge> itrOut = node.getAllEdges()[1].iterator();
			while(itrOut.hasNext()) 
			{
				curEdge = itrOut.next();
				if (edges.indexOf(curEdge) != edges.size()-1)
				{
					// Move the edge to the end of the collection (so that it appears on top).
					edges.remove(curEdge);
					edges.add(curEdge);
				}
			}
		}
	}
	
	/**
	 * Searches through nodes and edges to find the entity at a given point on the canvas. 
	 * If multiple entities are overlapping, the top entity will be returned. If different
	 * types of entities are overlapping, nodes get priority over edge labels, and edge labels
	 * get priority over edges.
	 * @param	p a point on the canvas to search for entities.
	 * @return	the entity at the given point (or null if there are no entities there).
	 * @author	Byron Knoll
	 */
	public Entity searchEntities(Point p) 
	{
		// Keep track of the entity found at point p.
		Entity found = null;

		Iterator<Edge> itr = edges.iterator();
		Edge curEdge;
		// Iterate through the edges and edge labels to see if p is contained in one of them.
		while(itr.hasNext()) 
		{
			curEdge = itr.next();
			if(curEdge.eLabel.contains(p))
				found = curEdge.eLabel;
			else if(curEdge.contains(p))
				found = curEdge;
		}
		
		Iterator<Node> nodeItr = nodes.iterator();
		Node curNode;
		// Iterate through the nodes to see if p is contained in one of them.
		while(nodeItr.hasNext()) 
		{
			curNode = nodeItr.next();
			if(curNode.contains(p))
				found = curNode;
		}
		
		// Return the entity (null if no entity was found).
		return found;
	}


	/**
	 * Searches through nodes to see which one of them lie within
	 * the rectangle bounded by x1, y1, x2, y2.
	 */
	public ArrayList<Entity> searchEntities( Point a, Point b ) {
		ArrayList<Entity> selected = new ArrayList<Entity>();
		
		Iterator<Node> nodeItr = nodes.iterator();
		Node curNode;
		//iterate through the nodes to see if p is contained in one of them
		while(nodeItr.hasNext()) {
			curNode = nodeItr.next();
			if(curNode.isInRect(a, b))
				selected.add( curNode );
		}
		Iterator<Edge> itr = edges.iterator();
		Edge curEdge;
		//iterate through the edges/edgeLabels to see if p is contained in one of them
		while(itr.hasNext()) {
			curEdge = itr.next();
			if(curEdge.isInRect( a, b ))
				selected.add( curEdge );
		}
		return selected;
	}

	/**
	 * Adds ent to the offsetPoint "selected".  Returns true.
	 */
	public boolean select( Entity ent ) {
		if( ent.type == GraphConsts.EDGE_LABEL ) {
			return true;
		} else if( !ent.isSelected ) {
			if( ent.type == GraphConsts.NODE ) 
				selectedNodes.add( (Node) ent );
			else if( ent.type == GraphConsts.EDGE )
				selectedEdges.add( (Edge)ent );
			ent.isSelected = true;
		}
		return true;
	}

	/**
	 * Removes ent from the offsetPoint "selected".  If the last entity
	 * is deselected, returns false.  Otherwise returns true.
	 */
	public boolean deselect( Entity ent ) {
		if( ent.type == GraphConsts.EDGE_LABEL ) {
			return false;
		} else if( ent.isSelected ) {
			if( ent.type == GraphConsts.NODE ) 
				selectedNodes.remove( ent );
			else if( ent.type == GraphConsts.EDGE )
				selectedEdges.remove( ent );
			ent.isSelected = false;
		}
		return ( selectedNodes.size() > 0 || selectedEdges.size() > 0 );
	}

	/**
	 * Deselects all edges and nodes.  Returns false.
	 */
	public boolean deselectAll() {
		// deselect all nodes
		Iterator<Node> nodeItr = selectedNodes.iterator();
		while(nodeItr.hasNext()) {
			nodeItr.next().isSelected = false;
		}
		selectedNodes.clear();

		// deselect all edges
		Iterator<Edge> edgeItr = selectedEdges.iterator();
		while(edgeItr.hasNext()) {
			edgeItr.next().isSelected = false;
		}
		selectedEdges.clear();

		return false;
	}

	/**
	 * Deselects all edges and nodes except ent.  Returns true.
	 */
	public boolean deselectAllButThis( Entity ent ) {
		deselectAll();
		select( ent );
		return true;
	}

	/** Moves all entities by dx and dy. */
	public void translateAll( float dx, float dy ) {
		// Move all nodes in graph
		for(Node n : nodes) {
			n.translate(dx, dy);
		}
	}
	
	/**
	 * Moves the node n to a new point.  Moves other selected entities.
	 */
	public void moveEntities( Point to, Node n ) {
		float dx = to.x - n.pos.x;
		float dy = to.y - n.pos.y;
		n.move( to );
		
		Iterator<Node> nodeItr = selectedNodes.iterator();
		Node curNode;
		while(nodeItr.hasNext()) {
			curNode = nodeItr.next();
			if(curNode.type == GraphConsts.NODE && curNode != n)
				curNode.translate(dx, dy);
		}
	}

	/**
	 * Moves the edge label.
	 */
	public void moveLabel( Point to, EdgeLabel label ) {
		label.move( to );
	}

	/**
	 * Sets the drawing scale.
	 */
	public void setScale( float newScale ) {
		if( newScale == 0 ) {
			// reset
			scale = 1;
			return;
		} else if( newScale < 0 ) 
			return;

		for(Node n : nodes) {
			n.pos.scale(newScale/scale);
		}
		for(Edge e : edges) {
			e.updateLength();
		}
		scale = newScale;
	}

	/** Gets the current scaling factor. */
	public float getScale() {
		return scale;
	}

	/** Resets the positions of the edge labels. */
	public void resetEdgeLabels() {
		Iterator<Edge> edgeItr = edges.iterator();
		while(edgeItr.hasNext()) {
			edgeItr.next().eLabel.reset();
		}
	}

	public Iterator<Edge> getEdges() {
		return edges.iterator();
	}

	public Iterator<Node> getNodes() {
		return nodes.iterator();
	}

	/**
	 * get the edge connecting nodes specified by start and end indexes
	 * @param startIndex
	 * @param endIndex
	 * @return - the edge connecting the two nodes
	 */
	public Edge getEdge( int startIndex, int endIndex ) {
		Iterator<Edge> edgeItr = edges.iterator();
		Edge curEdge;
		while(edgeItr.hasNext()) {
			curEdge = edgeItr.next();
			if((curEdge.getStartNode().getIndex() == startIndex) && (curEdge.getEndNode().getIndex() == endIndex))
				return curEdge;
		}
		return null;
	}

	/**
	 * returns true if the given index is assigned to a node
	 * @param index - the index to check
	 */
	protected boolean validNodeIndex( int index ) {
		Iterator<Node> nodeItr = nodes.iterator();
		while(nodeItr.hasNext()) {
			if(nodeItr.next().getIndex() == index)
				return true;
		}
		return false;
	}

	/**
	 * return the last index of the nodes
	 */
	public int getLastIndex() {
		if( numNodes() == 0 ) 
			return -1;
		return nodeAt( numNodes() - 1 ).getIndex();
	}

	/**
	 * gets the next node index
	 */
	public int getNextIndex() {
		int index = 0;
		for( int i = 0; i < numNodes(); i++ ) {
			if( index != nodeAt(i).getIndex() ) 
				return index;
			index = nodeAt(i).getIndex() + 1;
		}
		return index;
	}

	/**
	 * returns the node at the given index
	 * @param index
	 * @return - node at index, null if no node at the index
	 */
	public Node nodeFromIndex( int index ) {
		Iterator<Node> nodeItr = getNodes();
		Node tmpNode;
		while (nodeItr.hasNext()) {
			tmpNode = nodeItr.next();
			if( index == tmpNode.getIndex() )
				return tmpNode;
		}
		return null;
	}

	/**
	 * returns true if name of node is allowed
	 * @param name - the possible name of the node
	 * @param node - the node that will be set to name
	 * @return - true if the name of the node is allowed
	 */
	public boolean isNodeNameAllowed( String name, Node node ) {
		Iterator<Node> nodeItr = nodes.iterator();
		Node curNode;
		while(nodeItr.hasNext()) {
			curNode = nodeItr.next();
			//check to make sure node isn't same as node to check, and name isn't same
			if(!curNode.equals(node) && curNode.getLabel().equals(name))
				return false;
		}
		return true;
	}

	public String toString() {
		if(numNodes() > 0) {
			StringBuffer graphText = new StringBuffer("Graph:\n");
			Iterator<Node> nodeItr = nodes.iterator();
			int i = 0;
			while(nodeItr.hasNext()) {
				graphText.append(i++).append(" - ").append(nodeItr.next().getLabel()).append("\n");
			}
			return graphText.toString();
		}
		return "Empty graph";
	}
}