/**
 * File: GraphCanvas.java
 * Aug 22, 2000
 */

package AIspace.graphToolKit;

import java.awt.Color;
import java.awt.Container;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterException;
import java.awt.print.PrinterJob;
import java.util.ArrayList;

import javax.swing.BorderFactory;
import javax.swing.JFrame;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;

import AIspace.graphToolKit.dialogs.EdgeDialog;
import AIspace.graphToolKit.dialogs.FontDialog;
import AIspace.graphToolKit.dialogs.MessageDialog;
import AIspace.graphToolKit.dialogs.MessageFrame;
import AIspace.graphToolKit.dialogs.NodeDialog;
import AIspace.graphToolKit.elements.Edge;
import AIspace.graphToolKit.elements.EdgeLabel;
import AIspace.graphToolKit.elements.Entity;
import AIspace.graphToolKit.elements.Node;
import AIspace.graphToolKit.elements.Point;

/**
 * A class for visualizing the Graph class.
 * @author Shinjiro Sueda
 */
public class GraphCanvas extends JPanel implements MouseListener, MouseMotionListener, Printable, KeyListener, ActionListener {

	//
	// Fields
	//

	public Container parent; 		// parent Frame of this canvas
	public Graph graph; 			// Graph object to be displayed

	public int WIDTH, HEIGHT; 		// width and height of this Canvas
	protected Font font; 			// font used by the graph

	protected Point mClickedPos;
	protected Point mDraggedPos;
	protected Point mMovedPos;
	protected Point mPressedPos;
	protected Point mReleasedPos;
	protected Point mOriginalPos;

	protected boolean isDragging; 	// currently dragging an entity
	protected boolean isMoving; 	// Currently moving a canvas entity
	protected boolean isClicked; 	// Current piece was already clicked
	protected boolean isSelected; 	// A piece is currently selected
	protected Entity hover; 		// An entity being hovered
	protected Node from; 			// starting node of an edge
	protected Node to; 				// ending node of an edge

	protected int makingEdge; 		// Currently making an edge
	protected final static int NO = 0;
	protected final static int YES = 1;
	protected final static int DONE = 2;

	protected int submode; 			// Current submode.  create/select/delete/etc
	protected int mode; 			// Current mode.  create/solve
	protected int rmode; 			// Current right-click mode.  pan/zoom
	
	public int tempMode;			// no idea why there are two of these...could possibly 
	protected int tmpMode;			// be moved into one

	protected NodeDialog nDialog;
	protected EdgeDialog eDialog;
	
	protected float[] panZoomRect; 	// rectangle shown when panning/zooming

	public boolean pause = true;
	public boolean inline;

	public JPopupMenu pop;
	protected MessageDialog md;
	protected MessageFrame mf;

	protected Entity entClicked;
	protected Entity tempEnt;
	
	// Used to keep track of the difference between the center of a node and the position 
	// of the cursor.
	protected float draggingDifX;
	protected float draggingDifY;

	//
	// Methods
	//

	/**
	 * Default Constructor.
	 */
	public GraphCanvas(Container parent, boolean inline) {
		this.parent = parent;
		this.inline = inline;
		graph = new Graph(this);

		init();
	}

	/**
	 * Initialization common to both constructors.
	 */
	public void init() {
		font = new Font("SansSerif", Font.PLAIN, 12);
		mode = GraphConsts.CREATE;
		submode = GraphConsts.C_SELECT;
		rmode = GraphConsts.ZOOM;
		addMouseListener(this);
		addMouseMotionListener(this);
		addKeyListener(this);
		setBackground(Color.white);
		panZoomRect = new float[4];
		clearMouse();
		pause = false;
		pop = new JPopupMenu();
		add(pop);
		tmpMode = -1;
		setCanvasSize();
		setVisible(true);
	}

	/**
	 * Make the popup menu for white space (canvas).
	 */
	protected void popupCanvas(int x, int y) {
		pop.removeAll();
		pop.setBorder(BorderFactory.createTitledBorder("Canvas Options"));

		JMenuItem mItem;

		if (mode == GraphConsts.CREATE) {
			mItem = new JMenuItem("Create Node");
			mItem.setActionCommand(mItem.getText());
			mItem.addActionListener(this);
			pop.add(mItem);
			pop.addSeparator();
		}

		mItem = new JMenuItem("Autoscale");
		mItem.setActionCommand(mItem.getText());
		mItem.addActionListener(this);
		pop.add(mItem);
		mItem = new JMenuItem("Reset Edge Labels");
		mItem.setActionCommand(mItem.getText());
		mItem.addActionListener(this);
		pop.add(mItem);
		
		pop.show(this, x, y);
	}

	/**
	 * Make the popup menu for node
	 */
	protected void popupNode(int x, int y) {
		// If in solve mode, display canvas popup
		if (mode == GraphConsts.SOLVE) {
			popupCanvas(x, y);
			return;
		}

		pop.removeAll();
		pop.setBorder(BorderFactory.createTitledBorder("Node Options"));

		JMenuItem mItem = new JMenuItem("Create Edge");
		mItem.setActionCommand(mItem.getText());
		mItem.addActionListener(this);
		pop.add(mItem);
		mItem = new JMenuItem("Set Properties");
		mItem.setActionCommand(mItem.getText());
		mItem.addActionListener(this);
		pop.add(mItem);
		mItem = new JMenuItem("Delete Node");
		mItem.setActionCommand(mItem.getText());
		mItem.addActionListener(this);
		pop.add(mItem);

		pop.show(this, x, y);
	}

	/**
	 * Make the popup menu for edge
	 */
	protected void popupEdge(int x, int y) {
		pop.removeAll();
		pop.setBorder(BorderFactory.createTitledBorder("Edge Options"));

		JMenuItem mItem = new JMenuItem("Set Properties");
		mItem.setActionCommand(mItem.getText());
		mItem.addActionListener(this);
		pop.add(mItem);
		mItem = new JMenuItem("Reset Edge Label");
		mItem.setActionCommand(mItem.getText());
		mItem.addActionListener(this);
		pop.add(mItem);
		mItem = new JMenuItem("Delete Edge");
		mItem.setActionCommand(mItem.getText());
		mItem.addActionListener(this);
		pop.add(mItem);
		pop.addSeparator();

		mItem = new JMenuItem("Autoscale");
		mItem.setActionCommand(mItem.getText());
		mItem.addActionListener(this);
		pop.add(mItem);
		
		pop.show(this, x, y);
	}

	public void setCanvasSize() {
		WIDTH = getSize().width;
		HEIGHT = getSize().height;
	}

	public void reset() {
		graph = new Graph(this);
		clearMouse();
	    repaint();
	}

	public void showMessage(String title, String msg) {
		pause();
		if (inline) {
			if (mf == null) {
				mf = new MessageFrame();
				mf.addActionListener(this);
			}
			mf.open(title, msg);
		} else {
			md = new MessageDialog((JFrame) parent);
			md.addActionListener(this);
			md.open(title, msg);
		}
		unpause();
	}

	/**
	 * Clears all the mouse variables.
	 */
	public void clearMouse() {
		isDragging = false;
		isMoving = false;
		isClicked = false;
		isSelected = false;
		mClickedPos = new Point();
		mDraggedPos = new Point();
		mMovedPos = new Point();
		mPressedPos = new Point();
		mReleasedPos = new Point();
		mOriginalPos = new Point();
		makingEdge = NO;
		from = null;
		to = null;
		panZoomRect[0] = 0;
		panZoomRect[1] = 0;
		panZoomRect[2] = 0;
		panZoomRect[3] = 0;
		entClicked = null;
	}

	/**
	 * Changes the graph.
	 */
	public void setGraph(Graph graph) {
		clearMouse();
		this.graph = graph;
		moveGraphToMiddle();
		zoom(maximizingScale());
	}

	public void moveGraphToPoint(Point point) {
		Point min = graph.getBound()[0];
		Point max = graph.getBound()[1];

		// If the number of entities in the graph is 0, return
		if (min.equals(0, 0) || max.equals(0, 0)) {
			return;
		}

		//find middle of graph
		float midX = ((Math.abs(max.x - min.x)) / 2) + min.x;
		float midY = ((Math.abs(max.y - min.y)) / 2) + min.y;
		
		graph.translateAll(point.x - midX, point.y - midY);
	}

	/**
	 * Changes its font.  Also changes graph's font.
	 */
	public void changeFont(int size) {
		if (size == -1) {
			new FontDialog(this);
		} else {
			font = new Font("SansSerif", Font.PLAIN, size);
		
			((GraphWindow)parent).setMessageCanvasFontSize( size );
		}
		
		graph.updateNodeSize();
		graph.updateEdgeSize();
		repaint();
	}

	public int getFontSize() {
		return font.getSize();
	}
	
	/**
	 * centers the graph panel over a given node
	 * @param node - the node to center over
	 */
	public void centerOverNode(Node node) {
		if(node == null) {
			return;
		}
		if (!inline)
		{
			Rectangle windowBounds = ((GraphWindow) parent).scrollPanel.getBounds();
			((GraphWindow) parent).setHorizontalBarPosition((int)node.pos.x - windowBounds.width/2);
			((GraphWindow) parent).setVerticalBarPosition((int)node.pos.y - windowBounds.height/2);
		}
	}

	/**
	 * Gets the font.
	 */
	@Override
	public Font getFont() {
		return font;
	}

	/**
	 * Returns the FontMetrics for the font.
	 */
	public FontMetrics getFontMetrics() {
		return getFontMetrics(font);
	}
	
	public FontMetrics getFontMetricsForPrinting() {
		return getFontMetrics(new Font("sans serif", Font.PLAIN, 9));
	}

	/**
	 * Updates the node size and edge lengths.
	 */
	public void updateGraph() {
		graph.updateNodeSize();
		graph.updateEdgeSize();
		repaint();
	}

	/**
	 * Changes the width of lines used in the graph.
	 */
	public void changeLineWidth() {
		graph.updateNodeSize();
		repaint();
	}

	public void resetLabels() {
		graph.resetLabels();
		repaint();
	}
	
	public int getSubmode(){
		return submode;
	}

	public int getMode() {
		return mode;
	}
	
	public String getPromptLabel() {
		if (!inline) {
			return ((GraphWindow) parent).getPromptLabel();
		}
		else
		{
			return ((InlineApplet) parent).getPromptLabel();
		}
	}
	
	public void setPromptLabel(String s) {
		if (!inline) {
			((GraphWindow) parent).setPromptLabel(s);
		}
		else
		{
			((InlineApplet) parent).setPromptLabel(s);
		}
	}
	
	public final void setPromptColor(Color col) {
		if (!inline) {
			((GraphWindow) parent).setPromptColor(col);
		}
	}

	/** Finds the maximizing scale. */
	protected float maximizingScale() {
		Point[] bounds = graph.getBound();
		Point min = bounds[0];
		Point max = bounds[1];

		// If the number of entities in the graph is 0, return
		if (min.equals(0, 0) || max.equals(0, 0)) {
			return 0;
		}

		// If the number of entities in the graph is 1, return
		if (min.equals(max)) {
			return -1;
		}

		// Find the maximizing scaling factor

		float visibleWidth, visibleHeight;
		if (!inline)
		{
			visibleWidth = ((GraphWindow)parent).scrollPanel.getWidth();
			visibleHeight = ((GraphWindow)parent).scrollPanel.getHeight();
		}
		else
		{
			visibleWidth = ((InlineApplet)parent).getCanvas().getWidth();
			visibleHeight = ((InlineApplet)parent).getCanvas().getHeight();
		}

		float graphWidth = max.x - min.x;
		float graphHeight = max.y - min.y;

		float widthRatio = visibleWidth / graphWidth;
		float heightRatio = visibleHeight / graphHeight;

		float ratio = Math.min(widthRatio, heightRatio);
		ratio *= graph.getScale();

		return ratio;
	}
	
	/**
	 * Finds the maximizing scale.
	 */
	public float maximizingScaleForPrinting() {
		Point[] bounds = graph.getBound();
		Point min = bounds[0];
		Point max = bounds[1];

		// If the number of entities in the graph is 0, return
		if (min.equals(0, 0) || max.equals(0, 0)) {
			return 0;
		}

		// If the number of entities in the graph is 1, return
		if (min.equals(max)) {
			return -1;
		}

		// Find the maximizing scaling factor
		float printingWidth = 450;
		float printingHeight = 400;


		float graphWidth = max.x - min.x;
		float graphHeight = max.y - min.y;

		float widthRatio = printingWidth / graphWidth;
		float heightRatio = printingHeight / graphHeight;

		float ratio = Math.min(widthRatio, heightRatio);
		ratio *= graph.getScale();

		return ratio;
	}


	/** Sets the graph scale. */
	public void zoom(float scale) {
		Point[] bounds = graph.getBound();
		Point min = bounds[0];
		Point max = bounds[1];

		// If the number of entities in the graph is 0, return
		if (min.equals(0, 0) || max.equals(0, 0)) {
			return;
		}

		//find middle of graph
		float midX = ((Math.abs(max.x - min.x)) / 2) + min.x;
		float midY = ((Math.abs(max.y - min.y)) / 2) + min.y;
	
		// then scale
		graph.setScale(scale);	
		moveGraphToPoint(new Point(midX, midY));

		repaint();
	}
	
	/**
	 * Center the zoom on a particular point.
	 * @param point - the point to center the zoom on.
	 * @param scale - the zoom scale.
	 * @author Byron Knoll
	 */
	public void zoom(Point point, float scale) {		
		Point[] bounds = graph.getBound();
		Point min = bounds[0];
		Point max = bounds[1];

		// If the number of entities in the graph is 0, return.
		if (min.equals(0, 0) || max.equals(0, 0)) {
			return;
		}
		
		// Calculate the horizontal and vertical ratios of the position of the point relative
		// to the graph bounding box (used to find the same relative position after scaling):
		double hRatio = (point.x - min.x) / Math.abs(max.x - min.x);
		double vRatio = (point.y - min.y) / Math.abs(max.y - min.y);
		
		// Scale:
		graph.setScale(scale);	

		// Reset the bounding box:
		min = graph.getBound()[0];
		max = graph.getBound()[1];

		// Find middle of graph:
		float midX = ((Math.abs(max.x - min.x)) / 2) + min.x;
		float midY = ((Math.abs(max.y - min.y)) / 2) + min.y;
		
		// Calculate the position of the point on the new graph:
		float newX = min.x + (float) ((max.x - min.x)*hRatio);
		float newY = min.y + (float) ((max.y - min.y)*vRatio);
		
		// Adjust the middle of the graph to center on the point:
		midX += (newX-midX);
		midY += (newY-midY);
		
		// Translate the graph:
		graph.translateAll(point.x - midX, point.y - midY);

		repaint();
	}
	
	/**
	 * Sets the graph scale.
	 */
	public void zoomForPrinting() {
		Point min = graph.getBound()[0];
		Point max = graph.getBound()[1];

		// If the number of entities in the graph is 0, return
		if (min.equals(0, 0) || max.equals(0, 0)) {
			return;
		}
	
		// scale
		graph.setScale(maximizingScaleForPrinting());
		moveGraphToPoint(new Point(300, 275));		
	}
	
	public void moveGraphToMiddle() {
		Point min = graph.getBound()[0];
		Point max = graph.getBound()[1];

		// If the number of entities in the graph is 0, return
		if (min.equals(0, 0) || max.equals(0, 0)) {
			return;
		}

		//find middle of graph
		float midX = ((Math.abs(max.x - min.x)) / 2) + min.x;
		float midY = ((Math.abs(max.y - min.y)) / 2) + min.y;
		float middleX, middleY;
		if (!inline)
		{
			middleX = (((GraphWindow)parent).scrollPanel.getHorizontalScrollBar().getValue())
			  + (((GraphWindow)parent).scrollPanel.getWidth()/2);
			middleY = (((GraphWindow)parent).scrollPanel.getVerticalScrollBar().getValue())
			  + (((GraphWindow)parent).scrollPanel.getHeight()/2);
		}
		else
		{
			middleX = ((InlineApplet)parent).getCanvas().getWidth()/2;
			middleY = ((InlineApplet)parent).getCanvas().getHeight()/2;
		}

		//find distance from middle of graph to middle of canvas
		graph.translateAll(middleX - midX, middleY - midY);
	}	

	public void moveToTopMiddle() {
		Point min = graph.getBound()[0];
		Point max = graph.getBound()[1];

		// If the number of entities in the graph is 0, return
		if (min.equals(0, 0) || max.equals(0, 0)) {
			return;
		}

		//find middle of graph
		float midX = ((Math.abs(max.x - min.x)) / 2) + min.x;
		
		float middleX, middleY;
		if (!inline)
		{
			middleX = (((GraphWindow)parent).scrollPanel.getHorizontalScrollBar().getValue())
			  + (((GraphWindow)parent).scrollPanel.getWidth()/2);
			middleY = (((GraphWindow)parent).scrollPanel.getVerticalScrollBar().getValue())
			  + (((GraphWindow)parent).scrollPanel.getHeight()/2);
		}
		else
		{
			middleX = ((InlineApplet)parent).getCanvas().getWidth()/2;
			middleY = ((InlineApplet)parent).getCanvas().getHeight()/2;
		}


		//translate graph so that middle of graph coincides with middle of canvas
		graph.translateAll(middleX - midX, middleY + 40);
	}
	
	/**
	 * Centers the view and maximizes scale.
	 */
	public void autoscale() {
		zoom(maximizingScale());
		zoom(maximizingScale());
		zoom(maximizingScale());

    	moveGraphToMiddle();
	}

	/**
	 * Paints the components.
	 * @param g Graphics for this canvas.
	 */
	@Override
	public void paintComponent(Graphics g) {
		super.paintComponent(g);

		drawPrint((Graphics2D)g);
	}
	
	/** override this method in canvases with windows that don't subclass GraphWindow */
	public boolean isAntiAliasingEnabled() {
		return ((GraphWindow) parent).isAntiAliasingEnabled();
	}

	public void drawPrint(Graphics2D g2) {
		if(isAntiAliasingEnabled()) {
		    g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		}
		g2.setColor(Color.black);
		if (isDragging && !isMoving && mode!=GraphConsts.SOLVE) {
			// draw the selection rectangle
			//parameters are x, y, width, height for drawRect
			g2.drawRect((int) Math.min(mPressedPos.x, mDraggedPos.x), (int) Math.min(mPressedPos.y, mDraggedPos.y), (int) Math.abs(mPressedPos.x - mDraggedPos.x), (int) Math.abs(mPressedPos.y - mDraggedPos.y));
		}
		
		if ((makingEdge == YES) || (makingEdge == DONE)) {
			// draw the partial edge
			g2.drawLine((int) from.pos.x, (int) from.pos.y, (int) mMovedPos.x, (int) mMovedPos.y);
		}
		
		//draw the graph
		graph.draw(g2, isMoving);
	}

	
	/**
	 * Changes the submode of the canvas. Also clears any temp mode stored.
	 */
	public void setSubmode(int newMode) {
		submode = newMode;
		tempMode = 0;
		clearMouse();
		graph.deselectAll();
		repaint();
	}

	/**
	 * checks if right conditions to reset submode, and does if meets conditions
	 * (entClicked != tempEnt) && (tempMode != 0)
	 * should be used at start of mousePressed method, in subclasses
	 * @param ent - the entity just clicked
	 */
	public void checkAndResetSubMode(Entity ent) {
		if((ent != tempEnt) && (tempMode != 0)) {
			submode = tempMode;
		}
	}

	/**
	 * Changes the mode of the canvas.
	 */
	public void setMode(int newMode) {
		mode = newMode;
		clearMouse();
		graph.deselectAll();
	}

	/**
	 * Changes the right-click mode.
	 */
	public void setRMode(int newMode) {
		rmode = newMode;
	}
	
	public int getRMode(){
		return rmode;
	}

	/**
	 * Prevents user interaction until unpause() is called.
	 */
	public void pause() {
		pause = true;
	}

	/**
	 * Allows user interaction again.
	 */
	public void unpause() {
		pause = false;
	}

	public void disposeWindows() {
		if (nDialog != null) {
			nDialog.dispose();
		}
		if (eDialog != null) {
			eDialog.dispose();
		}
		if (md != null) {
			md.dispose();
		}
		if (mf != null) {
			mf.dispose();
		}
	}

	//
	// the following methods implement MouseListener
	//

	/**
	 * Invoked when the mouse has been clicked on a component.
	 */
	public void mouseClicked(MouseEvent e) {
		if (pause) {
			return;
		}
		if (entClicked != null) {
			entClicked.isBold = false;
		}
		Point p = new Point(e.getX(), e.getY());
		mClickedPos = new Point(p);
		entClicked = graph.searchEntities(p);
		
		if (e.getModifiers() == InputEvent.BUTTON3_MASK) {
			isSelected = graph.deselectAll();
			if (hover != null) {
				hover.isBold = false;
			}
			if (makingEdge == YES) {
				makingEdge = NO;
				submode = tmpMode;
				
				return;
			}
			if ((entClicked == null)) { 
				 popupCanvas(e.getX(), e.getY());
				
			} else {
				entClicked.isBold = true;
				isClicked = true;
				isSelected = graph.select(entClicked);
		
				if (entClicked.type == GraphConsts.NODE) {
					popupNode(e.getX(), e.getY());
				} else if (entClicked.type == GraphConsts.EDGE) {
					popupEdge(e.getX(), e.getY());
				} else if (entClicked.type == GraphConsts.EDGE_LABEL) {
					entClicked.isBold = false;
					entClicked = ((EdgeLabel) entClicked).parent;
					isClicked = true;
					isSelected = graph.deselectAll();
					isSelected = graph.select(entClicked);
					popupEdge(e.getX(), e.getY());
				}
			}
		}
		repaint();
	}

	/**
	 * Invoked when a mouse button has been pressed on a component.
	 */
	public void mouseEntered(MouseEvent e) {
		if (pause) {
			return;
		}
		repaint();
	}

	/**
	 * Invoked when the mouse exits a component.
	 */
	public void mouseExited(MouseEvent e) {
		if (pause) {
			return;
		}
		repaint();
	}

	/**
	 * Invoked when a mouse button has been pressed on a component.
	 */
	public void mousePressed(MouseEvent e) {
		if (pause) {
			return;
		}
		requestFocus();

		Point p = new Point(e.getX(), e.getY());
		Entity ent = graph.searchEntities(p);
		mPressedPos.move(p);
		mDraggedPos.move(p);
		mOriginalPos.move(p);

		if (e.getModifiers() == InputEvent.BUTTON3_MASK) {
			return;
		} else if ((ent == null) && !(e.isControlDown()||e.isShiftDown())) {
			graph.deselectAll();
			if(pop.isVisible()) { 
				pop.setVisible(false);
				return;
			}
		}

		if (mode == GraphConsts.CREATE) {
			switch(submode) {
				case GraphConsts.C_CREATE_NODE:
					createNode(ent, p);
					break;
				case GraphConsts.C_CREATE_EDGE:
					createEdge(ent, p);
					break;
				case GraphConsts.C_SELECT:
					mPressedSelect(ent, (e.isControlDown()||e.isShiftDown()));
					break;
				case GraphConsts.C_DELETE:
					if(ent != null) {
						undo();
						deleteEnt(ent);
					}
					break;
				case GraphConsts.C_SET_PROP:
					if(ent != null) {
						setEntProperties(ent);
					}
					break;
				case GraphConsts.C_SELECT_NODE:
					if((tempEnt != null) && (ent == tempEnt)) {
						mPressedSelect(ent, false);
					} else if(tempMode != 0) {
						submode = tempMode;
						tempMode = 0;
						tempEnt = null;
					}
					break;
			}
		} 
		else if (mode == GraphConsts.SOLVE) {
			mPressedSelect(ent, false);
		}
		repaint();
	}

	/**
	 * Invoked when a mouse button has been released on a component.
	 */
	public void mouseReleased(MouseEvent e) {
		if (pause) {
			return;
		}
		Point p = new Point(e.getX(), e.getY());
		Entity ent = graph.searchEntities(p);
		mReleasedPos.move(p);
		if (e.getModifiers() != InputEvent.BUTTON3_MASK) {
			if ((mode == GraphConsts.CREATE)) {
				if ((submode == GraphConsts.C_SELECT) || (submode == GraphConsts.C_SELECT_NODE)) {
					mReleasedSelect(ent, (e.isControlDown()||e.isShiftDown()));
				}
			}
			else if (mode == GraphConsts.SOLVE) {
				graph.deselectAll();
			}
		}
		if((tempMode != 0) && ((submode == GraphConsts.C_SELECT) || (submode == GraphConsts.C_SELECT_NODE))) {
			submode = tempMode;
			tempMode = 0;
			tempEnt = null;
		}
		isDragging = false;
		isMoving = false;
		mDraggedPos.move(0, 0);
		repaint();
	}

	//
	// The end of the interface MouseListener
	//

	//
	// The following methods implement MouseMotionListener
	//

	/**
	 * Invoked when a mouse button is pressed on a component and then dragged. 
	 * Mouse drag events will continue to be delivered to the component where  
	 * the first originated until the mouse button is released (regardless of 
	 * whether the mouse position is within the bounds of the component).
	 */
	public void mouseDragged(MouseEvent e) {
		if (pause) {
			return;
		}
		Point p = new Point(e.getX(), e.getY());
		mDraggedPos.move(p);
		
		if (e.getModifiers() == InputEvent.BUTTON3_MASK) {
			if (!isDragging) {
				// The user must drag for at least 5 pixels first
				if ((Math.abs(mPressedPos.x - p.x) > 5) || (Math.abs(mPressedPos.y - p.y) > 5)) {
					isDragging = true;
					isMoving = true;
				}

				repaint();
				return;
			} else {
				float dx = p.x - mPressedPos.x;
				float dy = p.y - mPressedPos.y;
				if (rmode == GraphConsts.PAN) {
					// Pan:
					graph.translateAll(dx, dy);
				} else if (rmode == GraphConsts.ZOOM) {
					// Zoom:
					float scale = graph.getScale() * (1 - dy/100);
					if (scale < 0.1) {
						scale = 0.1f;
					}
					zoom(mOriginalPos, scale);
				}
				// Change mPressedPos so that zoom/pan are relative to the last position.
				mPressedPos.x = p.x;
				mPressedPos.y = p.y;
			}
		} 
		else if (submode != GraphConsts.C_DELETE)
		{
			mDraggedSelect();
		}
		repaint();		
	}

	/**
	 * Invoked when the mouse cursor has been moved over a component (with no 
	 * buttons down).
	 */
	public void mouseMoved(MouseEvent e) {
		if (pause) {
			return;
		}
	
		Point p = new Point(e.getX(), e.getY());
		Entity ent = graph.searchEntities(p);
		if ((hover != null) && ((ent == null) || (ent != hover))) {
			hover.isBold = false;
			hover = null;
		} else if (hover != ent) {
			hover = ent;
			hover.isBold = true;
		}
		if (submode == GraphConsts.C_CREATE_EDGE) {
			if (makingEdge == YES) {
				mMovedPos.move(p);
			}
		}
		repaint();
	}

	//
	// The end of the interface MouseMotionListener
	//

	/**
	 * Called by the mouse functions.  This method should be overridden in
	 * the subclasses of this class.
	 */
	protected void createNode(Entity ent, Point p) {
		if ((ent == null) || ((ent.type != GraphConsts.NODE) && (ent.type != GraphConsts.EDGE) && (ent.type != GraphConsts.EDGE_LABEL))) {
			pause();
		
			if (nDialog == null) {
				nDialog = new NodeDialog(this);
			}
		
			nDialog.open( p );
		
			unpause();
		}
	}

	/**
	 * Called by the mouse functions.  This method should be overridden in
	 * the subclasses of this class.
	 */
	protected void createEdge(Entity ent, Point p) {
		if (makingEdge == YES) {
			if ((ent == null) || (ent.type != GraphConsts.NODE)) {
				makingEdge = NO;
			} else if (ent.type == GraphConsts.NODE) {
				pause();
				makingEdge = DONE;
				to = (Node) ent;
				if (eDialog == null) {
					eDialog = new EdgeDialog(this);
				}
				eDialog.open(from, to);
				unpause();
				makingEdge = NO;
			}
			if (tmpMode != -1) {
				submode = tmpMode;
			}
			tmpMode = -1;
		} else if ((ent != null) && (ent.type == GraphConsts.NODE)) {
			from = (Node) ent;
			makingEdge = YES;
			mMovedPos.move(p);
		}
	}

	/**
	 * Called by the mouse functions.  This method should be overridden in
	 * the subclasses of this class.
	 */
	protected void setEntProperties(Entity ent) {
		if (ent != null) {
			pause();
			
			switch(ent.type) {
				case GraphConsts.NODE:
					if (nDialog == null) {
						nDialog = new NodeDialog(this);
					}
					nDialog.open((Node) ent);
					graph.updateNodeSize((Node) ent);
					break;
				case GraphConsts.EDGE:
					if (eDialog == null) {
						eDialog = new EdgeDialog(this);
					}
					eDialog.open((Edge) ent);
					break;
				case GraphConsts.EDGE_LABEL:
					if (eDialog == null) {
						eDialog = new EdgeDialog(this);
					}
					eDialog.open(((EdgeLabel) ent).parent);
					break;
			}
			
			unpause();
		}
	}

	protected void undo()
	{
	}

	/**
	 * Called by the mouse functions.
	 */
	protected void deleteEnt(Entity ent) {
		if (ent != null) {
			if (ent.type == GraphConsts.EDGE_LABEL) {
				graph.select(((EdgeLabel) ent).parent);
			} else {
				graph.select(ent);
			}
			graph.deleteSelected();
			isSelected = false;
			isClicked = false;
		}
	}

	/**
	 * Entity selection algorithm for mousePressed.
	 */
	protected void mPressedSelect(Entity ent, boolean shiftOrCtrl) {
		if (ent != null) {
			if (!shiftOrCtrl && !ent.isSelected)
			{
				isSelected = graph.deselectAll();
			}
			isClicked = ent.isSelected;
			isSelected = graph.select(ent);
		
			if (isClicked && (ent.type != GraphConsts.EDGE)) {
				isMoving = true;
			}
		}
		repaint();
	}

	/**
	 * Entity selection algorithm for mouseReleased.
	 */
	protected void mReleasedSelect(Entity ent, boolean shiftOrCtrl) {
		if (isDragging) {
			if (!isMoving) {
				float x1 = Math.min(mPressedPos.x, mReleasedPos.x);
				float y1 = Math.min(mPressedPos.y, mReleasedPos.y);
				float w = Math.abs(mPressedPos.x - mReleasedPos.x);
				float h = Math.abs(mPressedPos.y - mReleasedPos.y);
				Point a = new Point(x1, y1);
				Point b = new Point(x1 + w, y1 + h);
				ArrayList<Entity> selectedEntities = graph.searchEntities(a, b);
				if (!shiftOrCtrl) 
				{
					isSelected = graph.deselectAll();
				}
				for (int i = 0; i < selectedEntities.size(); i++) {
					isSelected = graph.select(selectedEntities.get(i));
				}
			}
		} else if (ent != null) {
			if (shiftOrCtrl) {
				if (isClicked) {
					isSelected = graph.deselect(ent);
				}
			} else {
				isSelected = graph.deselectAllButThis(ent);
			}
		}
		repaint();
	}

	/**
	 * Entity selection algorithm for mouseDragged.
	 */
	protected void mDraggedSelect() {
		// The user must drag for at least 5 pixels first
		if (!isDragging) {
			if ((Math.abs(mPressedPos.x - mDraggedPos.x) > 5) || (Math.abs(mPressedPos.y - mDraggedPos.y) > 5)) {
				isDragging = true;
				if ((hover != null) && (hover.type == GraphConsts.NODE))
				{
					// Move the node so that it appears on top.
					graph.moveNodeToTop((Node)hover);
					// Keep track of the difference between the center of the node and the
					// position of the cursor.
					draggingDifX = (((Node)hover).getPos().x-mPressedPos.x);
					draggingDifY = (((Node)hover).getPos().y-mPressedPos.y);
				}
			}
			return;
		} else if ((hover != null) && (hover.type != GraphConsts.EDGE)) {
			if (isMoving) {
				switch(hover.type) {
					case GraphConsts.NODE:
						float offsetX =  mDraggedPos.x + draggingDifX;
						float offsetY =  mDraggedPos.y + draggingDifY;						
						Point point = new Point (offsetX, offsetY);
						graph.moveEntities(point, (Node) hover); 
						break;
					case GraphConsts.EDGE_LABEL:
						graph.moveLabel(mDraggedPos, (EdgeLabel) hover); break;
				}
			} else if (isSelected) {
				isMoving = isDragging = true;
			}
		}
		repaint();
	}

	//
	// The following methods implement KeyListener
	//
	/**
	 * Invoked when a key has been pressed.
	 */
	public void keyPressed(KeyEvent e) {
		if (pause || (mode == GraphConsts.SOLVE)) {
			return;
		}
		int arg = e.getKeyCode();
		if (arg == KeyEvent.VK_DELETE) {
			graph.deleteSelected();
		} else if (arg == KeyEvent.VK_ESCAPE) {
			makingEdge = NO;
			submode = tmpMode;
		}
		repaint();
	}

	public void keyReleased(KeyEvent e) {}
	public void keyTyped(KeyEvent e) {}

	//
	// The end of the interface KeyListener
	//

	//
	// The following implements ActionListener
	//
	public void actionPerformed(ActionEvent e) {
		aPerformed(e);
		repaint();
	}

	//
	// The end of the interface ActionListener
	//
	protected void aPerformed(ActionEvent e) {
		String arg = e.getActionCommand();
		if (arg.equals("msgBoxClosed")) {
			unpause();
		} else if (pause) {
			return;
		}

		if (arg.equals("Create Node")) {
			createNode(entClicked, mClickedPos);
		} else if (arg.equals("Create Edge")) {
			createEdge(entClicked, mClickedPos);
			if(tmpMode <= 0) {
				tmpMode = submode;
			}
			submode = GraphConsts.C_CREATE_EDGE;
		} else if (arg.startsWith("Set Properties")) {
			if(tempMode == 0) {
				tempMode = submode;
			}
			submode = GraphConsts.C_SET_PROP;
			setEntProperties(entClicked);
			submode = tempMode;
		} else if (arg.startsWith("Delete")) {
			if(tempMode == 0) {
				tempMode = submode;
			}
			submode = GraphConsts.C_DELETE;
			deleteEnt(entClicked);
			submode = tempMode;
		} else if(arg.startsWith("Select")){
			tempEnt = entClicked;
			if(tempMode == 0) {
				tempMode = submode;
			}
			submode = GraphConsts.C_SELECT;
			return;
		} else if (arg.equals("Autoscale")) {
		 	autoscale();
		} else if (arg.equals("Reset Edge Labels")) {
			resetLabels();
		} else if (arg.equals("Reset Edge Label")) {
			graph.resetLabel((Edge) entClicked);
		} else if (arg.equals("Pan")) {
			setRMode(GraphConsts.PAN);
			if (!inline)
			{
				((GraphWindow) parent).setPanZoom(1);
				((GraphWindow) parent).setPromptLabel("Drag the right mouse button to pan the graph.");
			}
		} else if (arg.equals("Zoom")) {
			setRMode(GraphConsts.ZOOM);
			if (!inline)
			{
				((GraphWindow) parent).setPanZoom(0);
				((GraphWindow) parent).setPromptLabel("Drag the right mouse button up to zoom in and down to zoom out.");
			}
		} 
		graph.deselectAll();
	}

	public int print(Graphics g, PageFormat pf, int pi) throws PrinterException{
		if(pi >= 1){
			return Printable.NO_SUCH_PAGE;	
		}
	
		zoomForPrinting();
		drawPrint((Graphics2D) g);
	
		return Printable.PAGE_EXISTS;
	}

	public void print() {
		PrinterJob printJob = PrinterJob.getPrinterJob();
		printJob.setPrintable(this);
		if(printJob.printDialog()){
			try{
				printJob.print();
			} catch (Exception ex){
				ex.printStackTrace();
			}
		}
	}
}