/*
 * File: InlineApplet.java
 * June 7, 2007
 */

package AIspace.graphToolKit;

import java.awt.Color;
import java.awt.Container;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;

import javax.swing.AbstractButton;
import javax.swing.ImageIcon;
import javax.swing.JApplet;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JToggleButton;

/**
 * This class provides common functionality for the inline applets. An inline applet is
 * embedded in a webpage and receives parameters to determine the applet features.
 * 
 * @author Byron Knoll
 */
public abstract class InlineApplet extends JApplet implements ActionListener, 
		ComponentListener{

	private static final long serialVersionUID = 1L;

	/**
	 * The inline applet canvas.
	 */
	protected GraphCanvas canvas;

	/**
	 * The inline applet buttons. A button's visibility is based on the corresponding
	 * button parameter.
	 */
	protected JToggleButton[] buttons;

	/**
	 * Used to display messages.
	 */
	protected LabelCanvas messageCanvas;

	/**
	 * The panel for the canvas and the message canvas.
	 */
	protected JPanel canvasPanel;

	/**
	 * Used to specify constraints for added components.
	 */
	protected GridBagConstraints gbc;

	/**
	 * Used for the canvas layout.
	 */
	protected GridBagLayout gbl;

	/**
	 * The main panel for the inline applet.
	 */
	protected JPanel mainPanel;

	/**
	 * The panel for the button tool bar. If no buttons are enabled this should not be
	 * visible.
	 */
	protected JPanel toolBarPanel;

	/**
	 * Set to true if the "messagePanel" parameter is "on".
	 */
	protected boolean showMessagePanel = false;

	/**
	 * Set to false if the "popMenu" parameter is "off".
	 */
	protected boolean popupMenuOn = true;
	
	/**
	 * Set to false if the "autoscale" parameter is "off".
	 */
	protected boolean autoscale = true;
	
	/**
	 * Set to true if the "border" parameter is "on".
	 */
	protected boolean showBorder = false;

	/**
	 * Set to true if any buttons are enabled.
	 */
	protected boolean showButtons = false;
	
	/**
	 * Set to true if the popup menu should show text representation items.
	 */
	public boolean textRepresentationPopupOn = false;
	
	/**
	 * Set to true if an incorrect parameter value has been used.
	 */
	protected boolean incorrectParameter = false;

	/**
	 * Constructor.
	 */
	public void init() {
		setBackground(GraphConsts.bg);
		setForeground(GraphConsts.fg);

		gbl = new GridBagLayout();
		gbc = new GridBagConstraints();
		gbc.anchor = GridBagConstraints.NORTHWEST;
		gbc.fill = GridBagConstraints.BOTH;

		mainPanel = new JPanel();

		messageCanvas = new LabelCanvas();
		canvasPanel = new JPanel();
		canvasPanel.setBackground(Color.white);
		canvasPanel.setLayout(gbl);
	}

	/**
	 * @return the inline applet canvas.
	 */
	public GraphCanvas getCanvas() {
		return canvas;
	}

	/**
	 * @return true if the popup menu is enabled, false otherwise.
	 */
	public boolean isPopupMenuOn() {
		return popupMenuOn;
	}

	/**
	 * Loads parameter values to determine applet features.
	 */
	protected void loadParams() {
		// This method loads parameters which are common to all inline applets.

		String param = getParameter("autoscale");
		if (param != null)
		{
			if (param.equals("off"))
				autoscale = false;
			else if (!param.equals("on"))
				incorrectParameter("autoscale", param);
		}
		
		param = getParameter("filename");
		if (param != null) {
			load(param);
		}

		param = getParameter("messagePanel");
		if (param != null)
		{
			if (param.equals("on"))
				showMessagePanel = true;
			else if (!param.equals("off"))
				incorrectParameter("messagePanel", param);
		}
		
		param = getParameter("popMenu");
		if (param != null)
		{
			if (param.equals("off"))
				popupMenuOn = false;
			else if (!param.equals("on"))
				incorrectParameter("popMenu", param);
		}
		
		param = getParameter("border");
		if (param != null)
		{
			if (param.equals("on"))
				showBorder = true;
			else if (!param.equals("off"))
				incorrectParameter("border", param);
		}
		
		param = getParameter("textRepresentationPopup");
		if (param != null)
		{
			if (param.equals("on")) {
				textRepresentationPopupOn = true;
			}
			else if (!param.equals("off"))
				incorrectParameter("textRepresentationPopup",param);
		}
	}
	
	/**
	 * Displays an error message about incorrect usage of a parameter value. Stops the
	 * applet from running by removing all of its components.
	 * 
	 * @param param the parameter name.
	 * @param value the parameter value.
	 */
	protected void incorrectParameter (String param, String value)
	{
		getContentPane().removeAll();
		getContentPane().setLayout(new FlowLayout());
		getContentPane().add(new JLabel("<html><font color=red>\""+value+"\" is not a valid value for the " +
				"\""+param+"\" parameter.</font></html>"));
		incorrectParameter=true;
	}

	/**
	 * Loads a preset graph.
	 */
	public abstract void load(String filename);

	/** Returns an ImageIcon, or null if the path was invalid. */
	protected static ImageIcon createImageIcon(String path) {
		java.net.URL imgURL = GraphWindow.class.getResource(path);
		if (imgURL != null) {
			return new ImageIcon(imgURL);
		} else {
			System.err.println("Couldn't find file: " + path);
			return null;
		}
	}

	/**
	 * Adds a component to the gridbag.
	 */
	protected void addComponent(JComponent c, Container p, int row, int column,
			int width, int height, double wx, double wy) {
		// set gridx and gridy
		gbc.gridx = column;
		gbc.gridy = row;

		// set gridwidth and gridheight
		gbc.gridwidth = width;
		gbc.gridheight = height;

		// set weights
		gbc.weightx = wx;
		gbc.weighty = wy;

		gbl.setConstraints(c, gbc);
		p.add(c);
	}

	/**
	 * Creates a new button for the solve panel.
	 * 
	 * @param num
	 *            the index of the button.
	 * @param buttonName
	 *            the button name.
	 * @param icon
	 *            an icon for the button.
	 * @param toolTip
	 *            the button tool tip.
	 */
	protected void setSolveButtonsWithText(int num, String buttonName, ImageIcon icon,
			String toolTip) {
		buttons[num] = new JToggleButton(buttonName, icon);
		buttons[num].addActionListener(this);
		buttons[num].setActionCommand(toolTip);
		buttons[num].setToolTipText(toolTip);

		buttons[num].setVerticalTextPosition(AbstractButton.BOTTOM);
		buttons[num].setHorizontalTextPosition(AbstractButton.CENTER);
	}

	/**
	 * Sets the label for the message canvas.
	 * 
	 * @param str
	 *            the string to set the label.
	 */
	public void setPromptLabel(String str) {
		messageCanvas.setText(str);
	}
	
	/**
	 * @return
	 *            the string currently set as the prompt label.
	 */
	public String getPromptLabel() {
		return messageCanvas.getText();
	}

	public void actionPerformed(ActionEvent e) {
		// Send action events to the canvas.
		canvas.actionPerformed(e);
	}
	
	/*
	 * COMPONENT LISTENER
	 */
	public void componentResized(ComponentEvent e) {
		if (autoscale)
			canvas.autoscale();
	}

	public void componentHidden(ComponentEvent e) {
	}

	public void componentMoved(ComponentEvent e) {
	}

	public void componentShown(ComponentEvent e) {
	}
}
