package AIspace.graphToolKit.dialogs;

import java.awt.Container;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JComponent;
import javax.swing.JFrame;

public abstract class BasicFrame extends JFrame implements ActionListener {
	protected GridBagConstraints gbc = null;
	protected GridBagLayout gbl = null;
	
	/** represents size of insets */
	private int insetSize;
	/** string for ok action, ignores case (defaults to "OK") */
	private String okString = "OK";
	/** string for cancel action, ignores case (defaults to "Cancel") */
	private String cancelString = "Cancel";

	/** constructor */
	public BasicFrame() {
		super();
	}
	
	/**
	 * constructor
	 * @param insetSize - size of the insets for this frame
	 */
	public BasicFrame(int insetSize) {
		super();
		setInsetSize(insetSize);
	}
	
	/**
	 * constructor
	 * @param title - title for this frame
	 */
	public BasicFrame(String title) {
		super(title);
	}
	
	/**
	 * constructor
	 * @param title - title for this frame
	 * @param insetSize - size of the insets for this frame
	 */
	public BasicFrame(String title, int insetSize) {
		super(title);
		setInsetSize(insetSize);
	}
	
	/** Adds a component to the gridbag. */
	protected final void addComponent(JComponent c, Container p, int row, int column, int width, int height, double wx, double wy) {
		if(( gbc == null) || (gbl == null))
			return;
		// set gridx and gridy
		gbc.gridx = column;
		gbc.gridy = row;

		// set gridwidth and gridheight
		gbc.gridwidth = width;
		gbc.gridheight = height;

		// set weights
		gbc.weightx = wx;
		gbc.weighty = wy;

		gbl.setConstraints(c, gbc);
		p.add(c);
	}
	
	/** 
	 * sets the inset size to number >= 0
	 * @param size - size of inset
	 */
	private void setInsetSize(int size) {
		if(insetSize < 0)
			this.insetSize = 0;
		else
			this.insetSize = size;
	}
	
	/** this method packs, centers and then sets the dialog to visible */
	protected final void packCenterOpen() {
		pack();
		centerWindow();
		setVisible(true);
	}
	
	/** centers the window in the screen */
	protected final void centerWindow() {
		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
		Point centerPosition = new Point(screenSize.width / 2, screenSize.height / 2);
		Dimension windowSize = getSize();
		setLocation(centerPosition.x - windowSize.width / 2, centerPosition.y - windowSize.height / 2);
	}
	
	/** adjusts frame to make sure that it stays on the screen */
	protected final void ensureOnScreen(AIspace.graphToolKit.elements.Point loc) {
		if(loc != null){
			int x =(int) loc.x; 
			int y =(int) loc.y;
			//if dialog will go off the screen, adjust the position so it remains
			//onscreen with a small padding near the edge
			Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
			if(x + getSize().width > screenSize.width){
				x = screenSize.width - getSize().width;
			}
			if(y + getSize().height > screenSize.height){
				y = screenSize.height - getSize().height - 60;
			}

			setLocation(x, y);
		} else { 
			centerWindow();
		}
	}
	
	/** 
	 * sets the string to use for the "OK" command (defaults to OK)
	 * @param okString - the string to use for the actionCommand
	 */
	protected final void setOKString(String okString) {
		this.okString = okString;
	}
	
	/** 
	 * sets the string to use for the "Cancel" command (defaults to Cancel)
	 * @param cancelString - the string to use for the actionCommand
	 */
	protected final void setCancelString(String cancelString) {
		this.cancelString = cancelString;
	}
	
	/** allow for spacing around edge of frame, overrides a super method */
	public final Insets getInsets() {
		Insets ins = (Insets) super.getInsets().clone();
		ins.left += insetSize;
		ins.right += insetSize;
		ins.bottom += insetSize;
		ins.top += insetSize;
		return ins;
	}
	
	/** basic actionPerformed, makes dialog disappear when OK or Cancel pressed */
	public void actionPerformed(ActionEvent e){
		String arg = e.getActionCommand().trim();
		if (arg.equalsIgnoreCase(okString)) {
			if(actionOK()) {
				setVisible(false);
				dispose();
			}
		} else if (arg.equalsIgnoreCase(cancelString)) {
			if(actionCancel()) {
				setVisible(false);
				dispose();
			}
		}
	}
	
	/** method to be performed when OK is triggered by actionEvent */
	protected abstract boolean actionOK();
	/** method to be performed when Cancel is triggered by actionEvent */
	protected abstract boolean actionCancel();
}
