/**
 * File: TextFrame.java
 * Date: July 20, 2006
 * Author: Kyle Porter
 */

package AIspace.graphToolKit.dialogs;

import java.awt.BorderLayout;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.SwingUtilities;
import javax.swing.event.CaretEvent;
import javax.swing.event.CaretListener;
import javax.swing.text.BadLocationException;

import AIspace.graphToolKit.GraphCanvas;

/**
 * TextFrame is an abstract class to produce a frame to show the text representation of the
 * graph, and allow the user to edit the text representation. 
 *
 * If the user clicks "Update" the updateAction method is called, which
 * is implemented in the subclases.
 */
public abstract class TextFrame extends BasicFrame {
	private JPopupMenu pop;
	protected JTextArea display;
	protected GraphCanvas canvas;
	
	protected JButton updateButton;

	//
	// Methods
	//

	/**
	 * Builds the frame
	 */
	public TextFrame (GraphCanvas canvas, String textRep, String title) {
		super ( title );
		super.setOKString("Update");
		
		this.setLocation(50,50);
		this.canvas = canvas;

		JPanel p = new JPanel();
		updateButton = new JButton("Update");
		updateButton.addActionListener(this);
		updateButton.setActionCommand(updateButton.getText());
		this.getRootPane().setDefaultButton(updateButton);
		p.add(updateButton);
		
		JButton button = new JButton("Cancel");
		button.addActionListener(this);
		button.setActionCommand(button.getText());
		p.add(button);

		display = new JTextArea(40, 80);
		display.setFont( new Font( "Monospaced", Font.PLAIN, 12 ) );	
		display.setText(textRep);
		display.addMouseListener(new MouseAdapter() {
			public void mouseClicked(MouseEvent e) {
				if (e.getModifiers() == MouseEvent.BUTTON3_MASK) {
					//finds point clicked on screen
					
					java.awt.Point screenp = new java.awt.Point(e.getX()-((JTextArea) e.getSource()).getTopLevelAncestor().getLocation().x,e.getY()-((JTextArea) e.getSource()).getTopLevelAncestor().getLocation().y);
					SwingUtilities.convertPointToScreen(screenp, display);
				
					popupEdit(screenp.x, screenp.y);
				}
			}
		});
		JScrollPane displaypane = new JScrollPane(display, JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		
		pop = new JPopupMenu();
		
        //Create the status area to display the caret position.
        JPanel statusPane = new JPanel(new BorderLayout());
        CaretListenerLabel caretListenerLabel = new CaretListenerLabel("1 : 1");
        statusPane.add(caretListenerLabel, BorderLayout.EAST);
        display.addCaretListener(caretListenerLabel);
        JPanel bottomPanel = new JPanel(new BorderLayout());
        bottomPanel.add(p, BorderLayout.CENTER);
        
        bottomPanel.add(statusPane, BorderLayout.SOUTH);
		
		getContentPane().add("Center", displaypane );
		getContentPane().add("South", bottomPanel );
	}
	
	/** called to open the frame, used so you can set other items in subclassed constructors */
	protected void open() {
		//move the caret to the beginning of the display
		display.setCaretPosition(0);
		packCenterOpen();
	}

	protected final void popupEdit(int x, int y){
		 pop.removeAll();

		 JMenuItem mItem;
		 mItem = new JMenuItem("Cut");
		 mItem.setActionCommand(mItem.getText());
		 mItem.addActionListener(this);
		 pop.add(mItem);
		 mItem = new JMenuItem("Copy");
		 mItem.setActionCommand(mItem.getText());
		 mItem.addActionListener(this);
		 pop.add(mItem);
		 mItem = new JMenuItem("Paste");
		 mItem.setActionCommand(mItem.getText());
		 mItem.addActionListener(this);
		 pop.add(mItem);
		 pop.addSeparator();
		 mItem = new JMenuItem("Select All");
		 mItem.setActionCommand(mItem.getText());
		 mItem.addActionListener(this);
		 pop.add(mItem);
				 
		 pop.show(this, x, y);
	}
	
	protected final boolean actionOK() {
		return updateAction();
	}
	
	protected final boolean actionCancel() {
		return true;
	}
	
	/** implemented in subclasses */
	protected abstract boolean updateAction();
	
	public final void actionPerformed (ActionEvent e) {
		super.actionPerformed(e);
		String arg = e.getActionCommand();
		if (arg.equals("Cut")) {
			display.cut();
		} else if (arg.equals("Copy")) {
			display.copy();
		} else if (arg.equals("Paste")) {
			display.paste();
		} else if (arg.equals("Select All")) {
			display.selectAll();
		}
	}
	
    //This listens for and reports caret movements.
    protected class CaretListenerLabel extends JLabel
                                       implements CaretListener {
		private static final long serialVersionUID = 1L;

		public CaretListenerLabel(String label) {
            super(label);
        }

        //Might not be invoked from the event dispatch thread.
        public void caretUpdate(CaretEvent e) {
            displaySelectionInfo(e.getDot(), e.getMark());
        }

        //This method can be invoked from any thread.  It 
        //invokes the setText and modelToView methods, which 
        //must run on the event dispatch thread. We use
        //invokeLater to schedule the code for execution
        //on the event dispatch thread.
        protected void displaySelectionInfo(final int dot,
                                            final int mark) {
            SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                	try {
                		setText((display.getLineOfOffset(dot)+1) + " : "
                				+ (1 + dot - display.getLineStartOffset
                						(display.getLineOfOffset(dot))) + "\n");
                        } catch (BadLocationException ble) {
                            setText("text position: " + dot + "\n");
                        }
                }
            });
        }
    }
}