/**
 * File: Edge.java
 * Jul 5, 2000
 */


package AIspace.graphToolKit.elements;

import java.awt.Color;
import java.awt.Graphics;

import AIspace.graphToolKit.Graph;
import AIspace.graphToolKit.GraphConsts;


/**
 * A general purpose edge class.  It is a subclass of Entity
 * @author Shinjiro Sueda
 */
public class Edge extends Entity {
	//
	// Fields
	//

	public Node start, end;					// starting and ending nodes for this edge
	protected float length;					// length of this edge
	public int edgeType;					// directional, bi-directional, non-directional
	protected Point[] b;					// border points
	protected Point[] p;					// ?? points
	public EdgeLabel eLabel;				// label class associated with this edge

	
	//
	// Methods
	//

	/**
	 * Default Constructor.
	 */
	public Edge( Graph graph ) {
		super( graph );
		label[0] = "";
		start = end = null;
		length = 0;
		edgeType = GraphConsts.NON_DIRECTIONAL;
		type = GraphConsts.EDGE;
		b = new Point[2];
		b[0] = new Point( 0, 0 );
		b[1] = new Point( 0, 0 );
		p = new Point[2] ;
		p[0] = new Point( 0, 0 );
		p[1] = new Point( 0, 0 );
		eLabel = new EdgeLabel( graph, this );
	}


	/**
	 * Another constructor
	 * @param graph - the graph the node is in
	 * @param label - the label for the node
	 * @param start - the starting node for the edge
	 * @param end - the ending node for the edge
	 * @param color - the color of the edge
	 * @param edgeType - the type of edge (DIRECTIONAL, etc)
	 */
	public Edge( Graph graph, String label, Node start, Node end, Color color, int edgeType ) {
		super( graph );
		this.label[0] = label;
		this.start = start;
		this.end = end;
		this.color = color;
		start.addEdgesOut( this );
		end.addEdgesIn( this );
		type = GraphConsts.EDGE;
		this.edgeType = edgeType;
		b = new Point[2];
		p = new Point[2] ;
		
		updateLength();
		eLabel = new EdgeLabel( graph, this );
	}


	/**
	 * Another constructor
	 * @param graph - the graph the node is in
	 * @param start - the starting node for the edge
	 * @param end - the ending node for the edge
	 */
	public Edge( Graph graph, Node start, Node end ) {
		super( graph );
		this.start = start;
		this.end = end;
		start.addEdgesOut( this );
		end.addEdgesIn( this );
		type = GraphConsts.EDGE;
		edgeType = GraphConsts.NON_DIRECTIONAL;
		b = new Point[2];
		p = new Point[2] ;

		updateLength();
		eLabel = new EdgeLabel( graph, this );
	}


	/**
	 * Sets the label.
	 */
	public void setLabel( String str ) {
		label[0] = str;
	}


	/**
	 * Sets its fields.
	 */
	public void set( String edgeLabel, Node start, Node end, Color color, int edgeType ) {
		label[0] = edgeLabel;
		this.start = start;
		this.end = end;
		this.color = color;
		this.edgeType = edgeType;
	}


	/**
	 * Draws itself onto the offscreen canvas.
	 */
	public void draw( Graphics g, boolean moving ) {
		// draw a thicker line
		drawLine( g );
		
		// draw the arrows
		if( edgeType != GraphConsts.NON_DIRECTIONAL ) {
			drawArrow( g, false, b[0], b[1] );
			if( edgeType == GraphConsts.BI_DIRECTIONAL ) {
				drawArrow( g, false, b[1], b[0] );
			}
		}
		
		// Draw the selection handles
		if( !moving && isSelected && graph.canvas.getMode() != GraphConsts.SOLVE  ) {
			drawHandle( g );
		}
	}
	
	/**
	* Draws itself onto the offscreen canvas.
	*/
	public void drawPrint( Graphics g, boolean moving ) {
		draw(g, moving);
	}
	
	public void drawStructure( Graphics g, boolean moving ) {
		if( start.equals( end ) ) {
			// draw a self-looping edge and return
			drawLine2( g);
			return;
		}

		// draw a thin line
		g.setColor( color );
		g.drawLine( (int)start.pos.x, (int)start.pos.y, (int)end.pos.x, (int)end.pos.y );
	 
		// draw the arrows
		if( edgeType != GraphConsts.NON_DIRECTIONAL ) {	
			drawArrowStructure( g, b[0], b[1] );
			if( edgeType == GraphConsts.BI_DIRECTIONAL ) {
				drawArrowStructure( g, b[1], b[0] );
			}
		}
	}

	protected void drawArrowStructure( Graphics g, Point tail, Point head ) {
		float dx = (head.x - tail.x)/2;
		float dy = (head.y - tail.y)/2;
	
		float lineWidth = graph.getLineWidth()/2 + xw;
		float arrowWidth = graph.getArrowWidth()/2 + xw;
		float arrowHeight = graph.getArrowHeight()/2 + xw;
		if( isBold ) {
			lineWidth = (float)Math.ceil(lineWidth*1.4);
			arrowWidth = (float)Math.ceil(arrowWidth*1.4);
			arrowHeight = (float)Math.ceil(arrowHeight*1.4);
		}
		
		if( arrowHeight > length )
			return;
	
		Point p1 = new Point( head );
		if (length!=0&&!Double.isNaN(length)) // Make sure that length is a valid for division.
		{
			p1.translate( (-arrowHeight/length)*dx, (-arrowHeight/length)*dy );
		}
		
		Point p2 = new Point( p1 );
		if (length!=0&&!Double.isNaN(length)) // Make sure that length is a valid for division.
		{ 
			p1.translate( (-arrowWidth/length)*dy, (arrowWidth/length)*dx );
			p2.translate( (arrowWidth/length)*dy, (-arrowWidth/length)*dx );
		}
		
		// Arrow head polygon
		int[] x = { (int)head.x, (int)p1.x, (int)p2.x };
		int[] y = { (int)head.y, (int)p1.y, (int)p2.y };

		//if bold, draw a solid triangle, otherwise white to hide
		//any lines underneath the arrow
		if( isBold ) {
			g.setColor( color );
		} else {
			g.setColor( Color.white );
		}
		g.fillPolygon( x, y, 3 );
		
		g.setColor( color );
		//draw the outline of the polygon
		g.drawPolygon( x, y, 3 );
	}

	/**
	 * Draws a thicker line from b1 to b2.
	 */
	protected void drawLine( Graphics g ) {
		if( start.equals( end ) ) {
			// draw a self-looping edge and return
			drawLine2( g);
			return;
		}

		g.setColor( color );
		float lineWidth = graph.getLineWidth() + xw;
		if( lineWidth == 1 && !isBold ) {
			//draw the thin line
			g.drawLine( (int)start.pos.x, (int)start.pos.y, (int)end.pos.x, (int)end.pos.y );
			return;
		}
		
		if( isBold ) 
			lineWidth = (float)Math.ceil(lineWidth*1.4);

		float moveX = (lineWidth / length) * (b[1].x - b[0].x);
		float moveY = (lineWidth / length) * (b[1].y - b[0].y);

		Point p1 = new Point( b[0] );
		Point p2 = new Point( p1 );
		Point p3 = new Point( b[1] );
		Point p4 = new Point( p3 );
		
		//increase the size of the line (for lineWidth);
		p1.translate( -moveY, moveX );
		p2.translate( moveY, -moveX );
		p3.translate( moveY, -moveX );
		p4.translate( -moveY, moveX );
		
		//shorten the edges so they don't show through the arrows
		if(this.edgeType == GraphConsts.DIRECTIONAL) {
			p3.translate(-3*moveX, -3*moveY);
			p4.translate(-3*moveX, -3*moveY);
		} else if (this.edgeType == GraphConsts.BI_DIRECTIONAL) {
			p1.translate(3*moveX, 3*moveY);
			p2.translate(3*moveX, 3*moveY);
			p3.translate(-3*moveX, -3*moveY);
			p4.translate(-3*moveX, -3*moveY);
		}
		

		int[] x = {(int) p1.x, (int) p2.x, (int) p3.x, (int) p4.x};
		int[] y = {(int) p1.y, (int) p2.y, (int) p3.y, (int) p4.y};
		
		g.setColor( color );
		g.fillPolygon( x, y, 4 );
	}

	protected void drawLine2( Graphics g ) {
		float dx = GraphConsts.selfDx;
		float dy = start.height + GraphConsts.selfDy;

		Point ne = new Point( start.pos.x + dx, start.pos.y - dy);
		Point nw = new Point( start.pos.x - dx, start.pos.y - dy );
		
		int[] x = new int[4];
		int[] y = new int[4];

		x[0] = (int)start.pos.x;
		y[0] = (int)start.pos.y;
		x[1] = (int)ne.x;
		y[1] = (int)ne.y;
		x[2] = (int)nw.x;
		y[2] = (int)nw.y;
		x[3] = x[0];
		y[3] = y[0];

		// draw a thin line
		g.setColor( color );
		g.drawPolyline( x, y, 4 );

		float lineWidth = graph.getLineWidth() + xw;
		if( lineWidth == 1 && !isBold ) return;

		if( isBold ) lineWidth = (float)Math.ceil(lineWidth*1.4);
		
		x = new int[8];
		y = new int[8];

		x[0] = (int)(start.pos.x + lineWidth);
		y[0] = (int)(start.pos.y + lineWidth);
		x[1] = (int)(ne.x + lineWidth*1.5);
		y[1] = (int)(ne.y - lineWidth);
		x[2] = (int)(nw.x - lineWidth*1.5);
		y[2] = (int)(nw.y - lineWidth);
		x[3] = (int)(start.pos.x - lineWidth);
		y[3] = y[0];
		x[4] = x[0];
		y[4] = y[0];
		x[5] = (int)nw.x;
		y[5] = (int)nw.y;
		x[6] = (int)ne.x;
		y[6] = (int)ne.y;
		x[7] = x[3];
		y[7] = y[3];

		g.fillPolygon( x, y, 8 );
	}


	/**
	 * Draws an arrow head.
	 */
	protected void drawArrow( Graphics g, boolean drawAnyways, Point tail, Point head ) {
		if( start.equals( end ) && !drawAnyways ) {
			drawArrow2( g );
			return;
		}
		arrow(g, tail, head);
	}


	public void arrow(Graphics g, Point tail, Point head) {
		float dx = head.x - tail.x;
		float dy = head.y - tail.y;
	
		float lineWidth = graph.getLineWidth() + xw;
		float arrowWidth = graph.getArrowWidth() + xw;
		float arrowHeight = graph.getArrowHeight() + xw;
		if( isBold ) {
			lineWidth = (float)Math.ceil(lineWidth*1.4);
			arrowWidth = (float)Math.ceil(arrowWidth*1.4);
			arrowHeight = (float)Math.ceil(arrowHeight*1.4);
		}
	
		Point p1 = new Point( head );
		if (length!=0&&!Double.isNaN(length)) // Make sure that length is a valid for division.
		{
			p1.translate( (-arrowHeight/length)*dx, (-arrowHeight/length)*dy );
		}
		
		Point p2 = new Point( p1 );
		if (length!=0&&!Double.isNaN(length)) // Make sure that length is a valid for division.
		{ 
			p1.translate( (-arrowWidth/length)*dy, (arrowWidth/length)*dx );
			p2.translate( (arrowWidth/length)*dy, (-arrowWidth/length)*dx );
		}
		
		// Arrow head polygon
		int[] x = { (int)head.x, (int)p1.x, (int)p2.x };
		int[] y = { (int)head.y, (int)p1.y, (int)p2.y };

		//if bold, draw a solid triangle, otherwise white to hide
		//any lines underneath the arrow
		if( isBold ) {
			g.setColor( color );
		} else {
			g.setColor( Color.white );
		}
		g.fillPolygon( x, y, 3 );
		
		g.setColor( color );
		//draw the outline of the polygon
		g.drawPolygon( x, y, 3 );
	}

	protected void drawArrow2( Graphics g ) {
		float dx = GraphConsts.selfDx;
		float dy = start.height + GraphConsts.selfDy;

		Point nw = new Point( start.pos.x - dx, start.pos.y - dy );
		
		// Make a temporary node at nw
		Node tmp = new Node( graph, "", nw, color, GraphConsts.OVAL );
		Point b = start.findBorder( tmp );

		length = distance( nw, b );
		drawArrow( g, true, nw, b );
	}

	/**
	 * Draw the selection handle for an edge.
	 */
	public void drawHandle( Graphics g) {
		// Check if the edge is self-looping.
		if( start.equals( end ) ) {
			drawHandle2( g );
			return;
		}
		Point first = new Point ((5*b[0].x + b[1].x)/6.0f,(5*b[0].y + b[1].y)/6.0f);
		Point mid = middle();
		Point last = new Point ((b[0].x + 5*b[1].x)/6.0f,(b[0].y + 5*b[1].y)/6.0f);
		g.setColor( Color.magenta );
		g.fillRect( (int)first.x - 2, (int)first.y - 2, 5, 5 );
		g.fillRect( (int)last.x - 2, (int)last.y - 2, 5, 5 );
		g.fillRect( (int)mid.x - 2, (int)mid.y - 2, 5, 5 );
	}

	/**
	 * Draw the selection handle for a self-looping edge.
	 */
	public void drawHandle2( Graphics g ) {
		float dx = GraphConsts.selfDx;
		float dy = start.height + GraphConsts.selfDy;

		Point ne = new Point( start.pos.x + dx, start.pos.y - dy);
		Point nw = new Point( start.pos.x - dx, start.pos.y - dy );
		Point n = new Point( start.pos.x, start.pos.y - dy );
		
		g.setColor( Color.magenta );
		g.fillRect( (int)n.x - 2, (int)n.y - 2, 5, 5 );
		g.fillRect( (int)ne.x - 2, (int)ne.y - 2, 5, 5 );
		g.fillRect( (int)nw.x - 2, (int)nw.y - 2, 5, 5 );
	}

	/**
	 * Calculates the distance between two points
	 */
	public float distance( Point a, Point b ) {
		float dx = b.x - a.x;
		float dy = b.y - a.y;
		return (float)Math.sqrt( dx*dx + dy*dy );
	}


	/**
	 * Updates the length of this edge.
	 */
	public void updateLength() {
		// Find the border points of this edge
		b[0] = start.findBorder( end );
		b[1] = end.findBorder( start );

		length = distance( b[0], b[1] );
	}
	
	/**
	 * Updates the length of this edge.
	 */
	public void updateLengthStructure() {
		updateLength();
		length /= 2;
	}

	/**
	 * Returns whether the point p is touching this edge.
	 */
	public boolean contains( Point p ) {
		return contains1( p, false, start.pos, end.pos );
	}


	protected boolean contains1( Point p, boolean checkAnyways, Point a, Point b ) {
		if( start.equals( end ) && !checkAnyways ) {
			return contains2( p );
		}

		float dx = b.x - a.x;
		float dy = b.y - a.y;

		Point c = new Point( a.x - dy, a.y + dx );
		Point d = new Point( b.x + dy, b.y - dx );

		// Treat x and y as vectors to calculate projections.
		Point x = new Point( p.x - a.x, p.y - a.y );
		Point y = new Point( b.x - a.x, b.y - a.y );

		float projX = x.x - (x.x*y.x + x.y*y.y)/(y.x*y.x + y.y*y.y)*y.x;
		float projY = x.y - (x.x*y.x + x.y*y.y)/(y.x*y.x + y.y*y.y)*y.y;

		float distSqrd = projX*projX + projY*projY;
		return ( (p.leftOf( a, c ) < 0) && (p.leftOf( b, d ) < 0)  && (distSqrd < 100) );
	}


	protected boolean contains2( Point p ) {
		float dx = GraphConsts.selfDx;
		float dy = start.height + GraphConsts.selfDy;

		Point ne = new Point( start.pos.x + dx, start.pos.y - dy);
		Point nw = new Point( start.pos.x - dx, start.pos.y - dy );
		
		return ( contains1( p, true, end.pos, ne ) || contains1( p, true, end.pos, nw ) || contains1( p, true, ne, nw ) );
	}


	/**
	 * Returns whether the node is inside the rectangle given by 
	 * the points a and b, where a is the upper left hand point,
	 * and b is the lower right hand point of the rectangle.
	 */
	public boolean isInRect( Point a, Point b ) {
		Point c = start.pos;
		Point d = end.pos;
		return ( c.x > a.x && c.x < b.x && c.y > a.y && c.y < b.y &&
					d.x > a.x && d.x < b.x && d.y > a.y && d.y < b.y );
	}


	/**
	 * Doesn't do anything.  This function is needed because "translate" is
	 * called on every entity in the "selected" list.
	 */
	public void translate( float dx, float dy ) {
		// Do nothing
	}


	/**
	 * Doesn't do anything.  This function is needed because "move" is
	 * called on every entity in the "selected" list.
	 */
	public void move( Point p ) {
		// Do nothing
	}


	/**
	 * Removes itself from starting and ending nodes' array of edges.
	 */
	public void removeFromNodes() {
		start.removeEdgesOut( this );
		end.removeEdgesIn( this );
	}


	public boolean equals( Entity ent ) {
		if( ent.type == GraphConsts.EDGE ) {
			Edge edge = (Edge)ent;
			return ( start.equals( edge.start ) && end.equals( edge.end ) );
		}
		return false;
	}


	public Point middle() {
		float mx = (b[0].x + b[1].x)/2.0f;
		float my = (b[0].y + b[1].y)/2.0f;
		return new Point( mx, my );
	}


	public float getDx() {
		return b[1].x - b[0].x;
	}

	public float getDy() {
		return b[1].y - b[0].y;
	}
	
	public float getPx() {
		return p[1].y - p[0].y;
	}

	public float getPy() {
		return p[1].y - p[0].y;
	}

	public Point[] getB() {
		return b;
	}
 
	public Point[] getP() {
		return p;
	}

	public float getLength() {
		return length;
	}

	/**
	 * Gets the other node connected to this edge.
	 */
	public Node otherNode( Node n1 ) {
		if( start.equals( n1 ) ) 
			return end;
		return start;
	}
	
	/** returns the start node of the edge */
	public Node getStartNode() {
		return start;
	}
	
	/** returns the end node of the edge */
	public Node getEndNode() {
		return end;
	}
	
	/** returns the integer edge type */
	public int getEdgeType() {
		return edgeType;
	}

	public String getDescription(){
		return start.getLabel() + "->" + end.getLabel();
	}

	/**
	 * Returns a text representation of this edge.
	 */
	public String toString() {
		return start + " -> " + end;
	}
}