/**
 * File: EdgeLabel.java
 * Jul 5, 2000
 */


package AIspace.graphToolKit.elements;

import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Graphics;

import AIspace.graphToolKit.Graph;
import AIspace.graphToolKit.GraphConsts;


/**
 * Draws the edge labels.  The position of the label is stored as a 2D offsetPoint
 * (x, y), where x is the % length along the edge from the start node of the
 * edge, and y is the absolute length perpendicular to the edge from the start
 * node of the edge.
 */
public class EdgeLabel extends Entity {

	//
	// Fields
	//

	public  Point offsetPoint;	// offset point in the parent basis
	public  Edge parent;	 // parent edge of this label
	private FontMetrics fm; // FontMetrics used by the parent canvas


	//
	// Methods
	//

	public EdgeLabel( Graph graph, Edge parent ) {
		super( graph );
		this.parent = parent;
		offsetPoint = new Point( 0.5f, 0.0f );
		type = GraphConsts.EDGE_LABEL;
	}

	public void reset() {
		offsetPoint.move( 0.5f, 0.0f );
	}

	public void draw( Graphics g, boolean moving ) {
		fm = graph.canvas.getFontMetrics();
		Point center = new Point( parent.getB()[0] );
		center.translate( changeVecBasis() );

		drawLabel( g, moving, center, false );
	}
	
	public void drawForPrinting( Graphics g, boolean moving ) {
		fm = graph.canvas.getFontMetrics();
		Point center = new Point( parent.getP()[0] );
		center.translate( changeVecBasisForPrinting() );

		drawLabel( g, moving, center, false );
	}
	
	/** does nothing */
	public void drawStructure( Graphics g, boolean moving ) {}
  
	public void drawLabel( Graphics g, boolean moving, Point center, boolean drawAnyways ) {
		label[0] = parent.label[0];

		if( label[0].equals( "" ) ) 
			return;
		if( parent.start.equals( parent.end ) && !drawAnyways ) {
			drawLabel2( g, moving );
			return;
		}

		int w = fm.stringWidth( label[0] );
		int h = fm.getHeight();
		int l = fm.getLeading();
		int a = fm.getAscent();
		
		if( isBold )
			g.setColor( Color.black );
		else
			g.setColor( Color.white );
		
		g.fillRect( (int)(center.x - w/2 - 2), (int)(center.y - h/2 - 2), w + 4, h + 4 );
		if( isBold ) {
			g.setColor( Color.white );
		} else {
			g.setColor( Color.black );
		}

		g.drawString( label[0], (int)center.x - w/2, (int)center.y - h/2 + l + a );
	}

	private void drawLabel2( Graphics g, boolean moving ) {
		float dy = parent.start.height + GraphConsts.selfDy;
		Point n = new Point( parent.start.pos.x, parent.start.pos.y - dy );
		n.translate( changeVecBasis() );
		drawLabel( g, moving, n, true );
	}
 
	/**
	 * Returns the offsetPoint using the elementary basis instead of the edge basis.
	 */
	public Point changeVecBasis() {
		if( parent.start.equals( parent.end ) ) {
			return new Point( (float)(offsetPoint.x - 0.5), offsetPoint.y );
		}
		float dx = parent.getDx();
		float dy = parent.getDy();
		float vx = dx*offsetPoint.x - dy*offsetPoint.y/parent.getLength();
		float vy = dy*offsetPoint.x + dx*offsetPoint.y/parent.getLength();
		return new Point( vx, vy );
	}
	/**
	 * Returns the offsetPoint using the elementary basis instead of the edge basis.
	 */
	public Point changeVecBasisForPrinting() 	{
		if( parent.start.equals( parent.end ) ) {
			return new Point( (float)(offsetPoint.x - 0.5), offsetPoint.y );
		}
		float dx = parent.getPx();
		float dy = parent.getPy();
		float vx = dx*offsetPoint.x - dy*offsetPoint.y/parent.getLength();
		float vy = dy*offsetPoint.x + dx*offsetPoint.y/parent.getLength();
		return new Point( vx, vy );
	}

	public Point[] getBounds() {
		fm = graph.canvas.getFontMetrics();
		label[0] = parent.label[0];
		if( label[0].equals( "" ) ) 
			return null;

		Point center = new Point( parent.getB()[0] );
		center.translate( changeVecBasis() );

		int w = fm.stringWidth( label[0] );
		int h = fm.getHeight();
		Point min = new Point( center );
		min.translate( -w/2 - 2, -h/2 - 2 );

		Point max = new Point( min );
		max.translate( w + 4, h + 4 );

		Point[] tmp = { min, max };
		return tmp;
	}

	public boolean contains( Point p ) {
		if( label[0].equals( "" ) ) 
			return false;
		Point center = new Point( parent.getB()[0] );
		center.translate( changeVecBasis() );
		return contains1( p, center, false );
	}

	public boolean contains1( Point p, Point center, boolean checkAnyways ) {
		if( parent.start.equals( parent.end ) && !checkAnyways ) {
			return contains2( p );
		}
		int w = fm.stringWidth( label[0] );
		int h = fm.getHeight();

		Point nw = new Point( center );
		nw.translate( -w/2 - 1, -h/2 - 1 );
		Point se = new Point( nw );
		se.translate( w + 2, h + 2 );

		return ( nw.x < p.x && p.x < se.x && nw.y < p.y && p.y < se.y );
	}

	public boolean contains2( Point p ) {
		float dy = parent.start.height + GraphConsts.selfDy;
		Point n = new Point( parent.start.pos.x, parent.start.pos.y - dy );
		n.translate( changeVecBasis() );
		return contains1( p, n, true );
	}


	/** returns false */
	public boolean isInRect( Point a, Point b ) {
		return false;
	}

	/** does nothing */
	public void translate( float dx, float dy ) {}

	/**
	 * This performs a change of basis from the elementary basis to edge basis.
	 */
	public void move( Point p ) {
		float moveX;
		if( parent.start.equals( parent.end ) ) {
			moveX = p.x - parent.start.pos.x;
			
			// move label along edge only
			if (moveX > parent.getLength()) 
				moveX = parent.getLength();
			else if (moveX < -parent.getLength()) 
				moveX = -parent.getLength();
		} else {
			float cx = p.x - parent.getB()[0].x; 
			float cy = p.y - parent.getB()[0].y;
			float x = parent.getDx(); 
			float y = parent.getDy();
			moveX = (x*cx + y*cy)/(x*x + y*y);
			
			// move label along edge only
			if(moveX > 1) 
				moveX = 1;
			else if(moveX < 0) 
				moveX = 0;
		}
		offsetPoint.move( moveX, 0 );
	}

	public boolean equals( Entity ent ) {
		if( ent.type == GraphConsts.EDGE_LABEL ) {
			return parent.equals( ((EdgeLabel)ent).parent );
		}
		return false;
	}

	public String toString() {
		return label[0] + " for edge:( " + parent + " )";
	}
}