

/**
 * This class extends the Node class to provide the capability to
 * display a histogram on the node.  
 * 
 * To use this class, create an object, call setHistogramCaption to
 * set the caption that appears at the top, then call addHistogramLabel
 * and addHistogramValue for each label/value pair to be displayed.
 * 
 * Note 1: The histogram will not display if there is insufficient data.
 * 
 * Note 2: Values are passed in as integers but displayed as decimals
 * 
 * Note 3: The constants for histogramNodeShape are in an updated GraphConsts class.  Valid shapes
 * are RECT and ROUNDED_RECT.  Rounded rectangle nodes will appear somewhat larger than the
 * rectangle nodes so that the histogram still fits.  RECT is the default value.
 * 
 * @author Wesley Coelho
 *
 */

package AIspace.graphToolKit.elements;

import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.text.DecimalFormat;
import java.util.ArrayList;

import AIspace.graphToolKit.Graph;
import AIspace.graphToolKit.GraphConsts;

public class HistogramNode extends Node {

	protected char[] data;
	protected int numdecimalplaces;  				//The number of decimal places to display
	protected boolean showHistogram;				//True if the histogram is to be displayed
	protected String caption;	 					//The caption to display above the histogram.
	protected ArrayList<String> labels; 			//A offsetPoint of labels for the histogram bars	
	protected ArrayList<Number> values;				//A offsetPoint of integer values for the histogram.
	protected int totalValue; 						//The total of all values to display
	protected FontMetrics fontMetrics; 				//Object used to get information about font measurements
	protected DecimalFormat numFormat;				//Used to format the values
	protected int maxLabelWidth;					//The screen width of the largest label
	protected Point nw;								//The North West point of the histogram node
	protected int lineHeight;						//The height of each row of the histogram node
	protected int histogramNodeShape;				//Indicates whether the histogram node is to be RECT or ROUNDED_RECT
	
	//CONSTANTS
	protected final int LINE_PADDING = 3;			//Pixels of padding between rows
	protected final int HORIZ_PAD = 3;				//Pixels of horizontal padding
	protected final int VERT_PAD = 3;				//Pixels of vertical padding
	protected final int MAX_BAR_WIDTH = 50;			//Maximum with of the histogram bars
	protected final int NUM_DASHED_LINES = 2;		//The number of vertical dashed lines to draw under the bars
	protected final int DASH_SIZE = 4;				//Length of the dash and space for the vertical dashed lines
	protected final int MINIMUM_LABEL_VAL_SPACE = 4;//Width of the minimum space between a label and its value

	/**
	 * A constructor that accepts the graph as a parameter
	 */
	public HistogramNode(Graph graph) {
		super(graph);
		construct();
	}

	/**
	 * Another constructor with more parameters.  The 'shape' parameter here refers
	 * to the node's shape when it is not being displayed as a histogram.  Use 
	 * setHistogramNodeShape to set the appearance for the histogram view.
	 */
	public HistogramNode(Graph graph, String label, Point pos, Color color, int shape) {
		super(graph);
		this.label[0] = label;
		this.pos = pos;
		this.color = color;
		this.shape = shape;
		construct();
	}

	/**
	 * This is called by the two constructors to initialize variables.
	 * 
	 */
	protected void construct()
	{
		showHistogram = false;
		caption = new String("");
		totalValue = 0;
		fontMetrics = graph.canvas.getFontMetrics();
		maxLabelWidth = 0;
	
		numFormat = new DecimalFormat("0.0####;0.0####");
		data = new char[7];
		data[0]= '0';
		data[1]= '.';
		for(int i = 2; i < 7; i ++ ) {
			data[i] = '0';
		}
		numdecimalplaces = 2;
		histogramNodeShape = GraphConsts.RECT;
	}

	/**
	 * Adds a new label to the histogram.  A corresponding value
	 * must also be added to this node.
	 */
	public void addHistogramLabel(String newLabel) {
		int newLabelWidth = fontMetrics.stringWidth(newLabel) + MINIMUM_LABEL_VAL_SPACE;
		if (newLabelWidth > maxLabelWidth) {
			maxLabelWidth = newLabelWidth;	
		}
		
		labels.add(newLabel);
	}

	/**
	 * Adds a new value to the histogram.
	 */
	public void addHistogramValue(int newValue) {
		values.add(new Integer(newValue));
		totalValue += newValue;
	}

	/**
	 * Removes the labels and values from the histogram
	 */
	public void clearHistogram() {
		totalValue = 0;
		labels = new ArrayList<String>();
		values = new ArrayList<Number>();
	}

	/**
	 * Sets the caption that appears above the histogram.
	 * 
	 */
	public void setHistogramCaption(String caption) {
		this.caption = caption;
	}

	/**
	 * Turns the histogram view on or off.
	 */
	public void showHistogramView(boolean showHistogram) {
		this.showHistogram = showHistogram;
		
		if (!showHistogram) {
			this.updateSize();			
			updatePosition();
		}
		else
		{
			updateBounds();
		}
	}

	/**
	 * Turns on histogram view if it's off, or turns it off
	 * if it's on.
	 */
	public void toggleHistogramView() {
		showHistogram = !showHistogram;
		
		if (!showHistogram) {
			this.updateSize();
			updatePosition();
		}
		else
		{
			updateBounds();
		}
	}
	
	/**
	 * Override the move method in order to update the nw point.
	 */
	public void move(Point p) {
		super.move(p);
		updatePosition();
	}
	
	/**
	 * Update the width and height of the node.
	 */
	public void updateBounds ()
	{
		if (!showHistogram)
			return; // Do not update the bounds if not showing histogram.
		// Update the node width and height.
		lineHeight = fontMetrics.getHeight();
		if (labels != null)
		{
			// The height of the histogram node
			this.height = (lineHeight + LINE_PADDING) * ( labels.size() + 1) + (2 * VERT_PAD);					
			// The width of the histogram node
			this.width = maxLabelWidth + MAX_BAR_WIDTH + fontMetrics.charsWidth(data, 0, numdecimalplaces + 2) + (4 * HORIZ_PAD);
		}
		else
		{
			// The height of the histogram node
			this.height = (lineHeight + LINE_PADDING) + (2 * VERT_PAD);					
			// The width of the histogram node
			this.width = maxLabelWidth + MAX_BAR_WIDTH + fontMetrics.charsWidth(data, 0, numdecimalplaces + 2) + (4 * HORIZ_PAD);
		}
		//Add extra space to accommodate the rounded rectangle
		if (histogramNodeShape == GraphConsts.ROUNDED_RECT) {
			this.height += lineHeight/2;
		}
	}
	
	/**
	 * Updates the NW point of the node.
	 */
	public void updatePosition()
	{
		nw = new Point (pos);
		nw.translate(-width / 2, -height / 2);
	}
	
	/**
	 * True if the histogram is currently on (it is visible)
	 */
	public boolean isHistogramViewOn() {
		return showHistogram;	
	}
	
	/**
	 * Recalculates the maximum label string width (ie. after a font size change)
	 * 
	 */
	protected void resetMaxLabelWidth() {
		maxLabelWidth = 0;
		if(labels != null){
		
			for (int i = 0; i < labels.size(); i++) {
				int newLabelWidth = fontMetrics.stringWidth(labels.get(i))  + MINIMUM_LABEL_VAL_SPACE;
				if (newLabelWidth > maxLabelWidth) {
					maxLabelWidth = newLabelWidth;	
				}	
			}
		}
	}
	

	/**
	* This is called when this node is to be drawn on screen.
	* 
	*/
	public void draw(Graphics offscreen, boolean moving)
	{

		if (fontMetrics != graph.canvas.getFontMetrics()) { 
			fontMetrics = graph.canvas.getFontMetrics();
			resetMaxLabelWidth(); 
		}
		
		if (showHistogram && values.size() > 0 && totalValue > 0.1) {
			drawNode(offscreen, moving);
		} else {
			super.draw(offscreen, moving);
		}
	}

	/**
	 * Draws the histogram view of the node
	 * 
	 */
	private void drawNode(Graphics offscreen, boolean moving) {
		lineHeight = fontMetrics.getHeight();																//The height of each row of the histogram node
		this.height = (lineHeight + LINE_PADDING) * ( labels.size() + 1) + (2 * VERT_PAD);					//The height of the histogram node
		this.width = maxLabelWidth + MAX_BAR_WIDTH + fontMetrics.charsWidth(data, 0, numdecimalplaces + 2) + (4 * HORIZ_PAD);		//The width of the histogram node
		
		//Add extra space to accommodate the rounded rectangle
		if (histogramNodeShape == GraphConsts.ROUNDED_RECT) {
			this.height += lineHeight/2;
		}
		
		if (this.width < fontMetrics.stringWidth(caption)) {
			width = fontMetrics.stringWidth(caption) + (2 * HORIZ_PAD);	
			
			if (histogramNodeShape == GraphConsts.ROUNDED_RECT) {
				width += 10 * HORIZ_PAD;	
			}
		}

		nw = new Point(pos);																				//The top left corner of the histogram node
		nw.translate(-width / 2, -height / 2);																//Actually sets the correct position
		drawOutline(offscreen);

		drawHistogramNode(offscreen);
	}


	/**
	 * Actually draws the histogram of a stationary node
	 * 
	 */
	private void drawHistogramNode(Graphics g) {
		drawOutline(g);
		drawCaption(g);
		drawRows(g);
	}

	/**
	 * Draws the actual rows of the histogram and the vertical line
	 * separating the bars from the rest of the node body.
	 */
	protected void drawRows(Graphics g) {
		int currentRowY =  (int) nw.y + (2 * lineHeight) + (2 * LINE_PADDING);
		//int valueX = (int) nw.x + width - HORIZ_PAD - MAX_BAR_WIDTH - HORIZ_PAD - fontMetrics.stringWidth("0.00");
		int valueX = (int) nw.x + width - HORIZ_PAD - MAX_BAR_WIDTH - HORIZ_PAD - fontMetrics.charsWidth(data, 0, numdecimalplaces +2);
		int barX = (int) nw.x + width - MAX_BAR_WIDTH;
		
		//Draw the vertical line before the bars		   
		g.drawLine(barX - 1,
		   (int) nw.y + lineHeight + LINE_PADDING, 
		   barX - 1,
		   (int) nw.y + height);
		   
		drawDashedLines(g, barX, (int) nw.y + lineHeight + LINE_PADDING, (int) nw.x + width, (int) nw.y + height);
		
		//Each iteration of this for loop is a horizontal line of the histogram
		for (int i = 0; i < labels.size(); i++) {
			int currentValue = values.get(i).intValue();
			
			//Draw Label
			g.setColor(Color.black);
			if (histogramNodeShape == GraphConsts.ROUNDED_RECT) {
				g.drawString(labels.get(i), (int) ( nw.x + HORIZ_PAD) + (2 * HORIZ_PAD), currentRowY);
			} else if (histogramNodeShape == GraphConsts.RECT) {
				g.drawString(labels.get(i), (int) ( nw.x + HORIZ_PAD), currentRowY);
			}

			//Draw Value
			g.drawString( numFormat.format( currentValue / Math.pow(10,numdecimalplaces) ), valueX, currentRowY);
			
			//Draw the bar
			int barWidth = (int) (MAX_BAR_WIDTH * ((double)currentValue / (double)totalValue));
			g.fillRect(barX, currentRowY - lineHeight + 5, barWidth, lineHeight - 3);
			
			//Increment the Y position for the next line
			currentRowY += lineHeight + LINE_PADDING;
		}
	}

	/**
	 * 
	 *Draws the vertical dashed lines under the horizontal histogram bars
	 * 
	 */
	protected void drawDashedLines(Graphics g, int startX, int startY, int endX, int endY) {
		int currentX = startX;
		int increment = (endX - startX) / (NUM_DASHED_LINES + 1);
		currentX += increment;
		
		for (int i = 0; i < NUM_DASHED_LINES; i++) {
			int currentY = startY;
			
			while (currentY < endY - DASH_SIZE) {
				g.drawLine(currentX, currentY, currentX, currentY + DASH_SIZE);
				currentY += (2 * DASH_SIZE);
			}	
			
			//Special case for last dashed line
			if (currentY < endY) {
				g.drawLine(currentX, currentY, currentX, endY);
			}
			
			currentX += increment;	
		}
	}
	
	/**
	 * Returns whether the point (x2, y2) is in this node.
	 */
	public boolean contains(Point p) {		
		// Return false if not initialized yet.
		if (width == 0 || height == 0)
			return false;

		// Use the parent method if not showing the histogram (or if the node is a diamond shape).
		if (!showHistogram || shape==GraphConsts.DIAMOND) return super.contains(p);
		
		// Otherwise treat the node as a rectangle:
		int dx = (int) Math.abs(pos.x - p.x);
		int dy = (int) Math.abs(pos.y - p.y);
		return ((4 * dx * dx / (width * width) + 4 * dy * dy / (height * height)) <= (1 / 2.0));
	}


	/**
	 * Draws the histogram node caption at the top and the 
	 * line between the caption and the histogram.
	 */
	private void drawCaption(Graphics g) {		
		//Draw the Caption
		int captionWidth = fontMetrics.stringWidth(caption);
		int captionHeight = fontMetrics.getHeight();
		g.setColor(Color.black);
		g.drawString(caption, (int) pos.x - captionWidth / 2, (int) nw.y + captionHeight);
		
		//Draw the line under the Caption
		g.setColor(color);
		g.drawLine((int) nw.x,
				   (int) nw.y + captionHeight + LINE_PADDING, 
				   (int) nw.x + width,
				   (int) nw.y + captionHeight + LINE_PADDING );
	}


	/**
	 * Draws the border of the histogram node
	 */
	private void drawOutline(Graphics offscreen) {
		float lineWidth = graph.getLineWidth() + xw;
		
		offscreen.setColor(color);
		if (lineWidth == 1) {
			if (isBold) {
				offscreen.setColor(color);
			} else {
				offscreen.setColor(Color.white);
			}
			if (histogramNodeShape == GraphConsts.ROUNDED_RECT) {
				offscreen.fillRoundRect((int) nw.x, (int) nw.y, width, height, height/2, height/2);
				offscreen.setColor(color);
				offscreen.drawRoundRect((int) nw.x, (int) nw.y, width, height, height/2, height/2);
			} else if (histogramNodeShape  == GraphConsts.RECT) {
				offscreen.fillRect((int) nw.x, (int) nw.y, width, height);
				offscreen.setColor(color);
				offscreen.drawRect((int) nw.x, (int) nw.y, width, height);
			}
		} else {
			if (histogramNodeShape == GraphConsts.ROUNDED_RECT) {
				offscreen.fillRoundRect((int) (nw.x - lineWidth), (int) (nw.y - lineWidth), (int) (width + lineWidth * 2), (int) (height + lineWidth * 2), height/2, height/2);
			} else if (histogramNodeShape  == GraphConsts.RECT) {
				offscreen.fillRect((int) (nw.x - lineWidth), (int) (nw.y - lineWidth), (int) (width + lineWidth * 2), (int) (height + lineWidth * 2));
			}
			
			if (!isBold) {
				offscreen.setColor(Color.white);
			}
			
			if (histogramNodeShape == GraphConsts.ROUNDED_RECT) {
				offscreen.fillRoundRect((int) nw.x, (int) nw.y, width, height, height/2, height/2);
			} else if (histogramNodeShape  == GraphConsts.RECT) {
				offscreen.fillRect((int) nw.x, (int) nw.y, width, height);
			}
		}
	}
	
	/**
	 * Sets the histogram node shape (either GraphConsts.RECT or GraphConsts.ROUNDED_RECT)
	 */
	public void setHistogramNodeShape(int newShape) {
		histogramNodeShape = newShape;	
	}
}