/**
 * File: Node.java
 * Jul 5, 2000
 */

package AIspace.graphToolKit.elements;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.util.ArrayList;

import AIspace.graphToolKit.Graph;
import AIspace.graphToolKit.GraphConsts;

/**
 * A general-purpose graph node.  It is a subclass of Entity
 * @author Shinjiro Sueda
 */
public class Node extends Entity {

	//
	// Fields
	//

	/** position in campus */
	public Point pos;
	/** width in canvas */
	public int width;
	/** height in canvas */
	public int height;
	/** shape of node (GraphConsts.OVAL,GraphConsts.RECT, GraphConsts.DIAMOND) */ 
	public int shape;
	/** edges going into this node */
	protected ArrayList<Edge> edgesIn;
	/** edges going out of this node */
	protected ArrayList<Edge> edgesOut;
	/** unique id for this node */
	public int index;
	/** internal width padding */
	protected int wPad;
	/** internal height padding */
	protected int hPad;
  	protected int printWidth, printHeight;		// ???
  	/** true if the outline of the node is thicker */
  	protected boolean isHeavy;
  	/** the amount added to line width when isHeavy is true */
  	protected int heavyWidth = 2;
  	/** color of the inside of the node when not bold */
  	protected Color internalColor;
	//
	// Methods
	//

	/**
	 * Default Constructor
	 */
	public Node(Graph graph) {
		super(graph);
		internalColor = Color.WHITE;
		pos = new Point();
		shape = 0;
		edgesIn = new ArrayList<Edge>();
		edgesOut = new ArrayList<Edge>();
		index = -1;
		updateSize();
		wPad = 20;
		hPad = 20;
	}

	/**
	 * Another Constructor
	 * @param graph - graph to add node to
	 * @param label - label to display in node
	 * @param pos - position of node
	 * @param color - color of node
	 * @param shape - shape of node (GraphConsts.DIAMOND, GraphConsts.OVAL, GraphConsts.RECTANGLE)
	 */
	public Node(Graph graph, String label, Point pos, Color color, int shape) {
		super(graph);
		internalColor = Color.WHITE;
		this.label[0] = label;
		this.pos = pos;
		isSelected = false;
		isBold = false;
		this.color = color;
		this.shape = shape;
		edgesIn = new ArrayList<Edge>();
		edgesOut = new ArrayList<Edge>();
		type = GraphConsts.NODE;
		wPad = 20;
		hPad = 20;
		updateSize();
	}

	/** updates the size of the node using getWidth, getHeight */
	public void updateSize() {
		width = getWidth();
		height = getHeight();
	}
	
	public void updateSize(int printWidth, int printHeight) {
		width = printWidth;
		height = printHeight;
	}

	/**
	 * Gets the width of this node.
	 */
	protected int getWidth() {
		return graph.canvas.getFontMetrics().stringWidth(label[0]) + wPad;
	}

	/**
	 * Gets the height of this node.
	 */
	protected int getHeight() {
		return graph.canvas.getFontMetrics().getHeight() + hPad;
	}

	public final Point getPos() {
		return pos;
	}
	
	public final void setPos(Point pos) {
		this.pos = pos;
	}
	
	public final float getXPos() {
		return pos.x;
	}
	
	public final float getYPos() {
		return pos.y;
	}

	/**
	 * Sets its fields.
	 */
	public void set(String l, Point p, Color c, int s) {
		label[0] = l;
		pos = p;
		color = c;
		shape = s;
	}
	
	/** set the node shape to have HEAVY outline or REGULAR outline */
	public final void setHeavy(boolean heavy)  {
		this.isHeavy = heavy;
	}
	
	public final boolean isHeavy(){
		return isHeavy;
	}
	
	public final void setShape(int shape){
		if((shape != GraphConsts.DIAMOND) && (shape != GraphConsts.RECT) 
				&& (shape != GraphConsts.OVAL) && (shape != GraphConsts.ROUNDED_RECT))
			this.shape = GraphConsts.RECT;
		else
			this.shape = shape;
	}
	
	public final int getShape(){
		return shape;
	}

	/**
	 * Draws itself onto the offscreen canvas.
	 */
	public void draw(Graphics graphics, boolean moving) {
		//get the width of the lines
		float lineWidth = graph.getLineWidth() + xw;
		if(isHeavy)
			lineWidth += heavyWidth;
		//get the point for the drawing of the object, translated for size of the node
		Point nw = new Point(pos);
		nw.translate(-width / 2, -height / 2);

		graphics.setColor(color);
		switch(shape) {
			case GraphConsts.RECT:
				//draw the coloured part of the rectangle
				graphics.fillRect((int) (nw.x - lineWidth), (int) (nw.y - lineWidth), (int) (width + lineWidth * 2), (int) (height + lineWidth * 2));
				if (!isBold) {
					//if the node isn't bold, draw a smaller white rect inside to 
					//make it look like an outline
					graphics.setColor(internalColor);
					graphics.fillRect((int) nw.x, (int) nw.y, width, height);
				}
				break;
			case GraphConsts.ROUNDED_RECT:
				// Making the minimum width of the rounded rectangle 20:
				if (width < 20) 
				{
					width = 20;
					nw = new Point(pos);
					nw.translate(-width / 2, -height / 2);
				}
				//draw the coloured part of the rectangle
				graphics.fillRoundRect((int) (nw.x - lineWidth), (int) (nw.y - lineWidth), (int) (width + lineWidth * 2), (int) (height + lineWidth * 2), 30, 30);
				if (!isBold) {
					//if the node isn't bold, draw a smaller white rect inside to 
					//make it look like an outline
					graphics.setColor(internalColor);
					graphics.fillRoundRect((int) nw.x, (int) nw.y, width, height, 30, 30);
				}
				break;
			case GraphConsts.DIAMOND:
				//generate the coords for the diamond
				int[] xVals = new int[4];
				xVals[0] = xVals[2] = (int) pos.x;
				xVals[1] = (int) (pos.x + (width / 1.5) + lineWidth);
				xVals[3] = (int) (pos.x - (width / 1.5) - lineWidth);
				int[] yVals = new int[4];
				yVals[0] = (int) (pos.y - (height / 1.5) - lineWidth);
				yVals[1] = yVals[3] = (int) pos.y;
				yVals[2] = (int) (pos.y + (height / 1.5) + lineWidth);

				//draw the coloured diamond
				graphics.fillPolygon(xVals, yVals, 4);
				if (!isBold) {
					//if it isn't bold, draw a smaller white diamond inside it to
					//make it look like an outline
					if(((Graphics2D) graphics).getRenderingHint(RenderingHints.KEY_ANTIALIASING) == RenderingHints.VALUE_ANTIALIAS_ON) {
						//antialiasing enabled
						xVals[1] -= 2*lineWidth;
						xVals[3] += 2*lineWidth;
					} else if(width < 55) {
						//for some reason the diamond looks lopsided if
						//the ratio isn't like this
						xVals[1] -= 1.6*lineWidth;
						xVals[3] += 2*lineWidth;
					} else {
						//for longer diamonds, use a ratio of the width to extend
						//the left/right widths so it doesn't look too narrow
						xVals[1] -= (width/25)*lineWidth;
						xVals[3] += (width/25)*lineWidth;
					}
					yVals[0] += lineWidth;
					yVals[2] -= lineWidth;
					graphics.setColor(internalColor);
					graphics.fillPolygon(xVals, yVals, 4);
				}
				break;
			case GraphConsts.OVAL:
			default:
				//default case and OVAL case are the same
				//draw the coloured part of the oval
				graphics.fillOval((int) (nw.x - lineWidth), (int) (nw.y - lineWidth), (int) (width + lineWidth * 2), (int) (height + lineWidth * 2));
				if (!isBold) {
					//if the node isn't bold, draw a smaller white oval inside to 
					//make it look like an outline
					graphics.setColor(internalColor);
					graphics.fillOval((int) nw.x, (int) nw.y, width, height);
				}
				break;
		}

		// Draw the label.
		drawLabel(graphics);
		
		if (!moving || graph.canvas.getSubmode() == GraphConsts.C_SELECT_NODE) {
			//draw the selection handles for the node
			if (isSelected && graph.canvas.getMode() != GraphConsts.SOLVE) {
				graphics.setColor(Color.magenta);
				graphics.fillRect((int) (pos.x - width / 1.5) - 2, (int) (pos.y - height / 1.5) - 2, 4, 4);
				graphics.fillRect((int) (pos.x - width / 1.5) - 2, (int) (pos.y + height / 1.5) - 2, 4, 4);
				graphics.fillRect((int) (pos.x + width / 1.5) - 2, (int) (pos.y + height / 1.5) - 2, 4, 4);
				graphics.fillRect((int) (pos.x + width / 1.5) - 2, (int) (pos.y - height / 1.5) - 2, 4, 4);
			}
		}
	}

	/**
	 * Draws the label.
	 */
	protected void drawLabel(Graphics offscreen) {
		if (isBold && (color == Color.blue || color == Color.black)) {
			offscreen.setColor(Color.white);
		} else
			offscreen.setColor(Color.black);
		offscreen.drawString(label[0], (int) pos.x - (graph.canvas.getFontMetrics().stringWidth(label[0])/2), (int) pos.y + (graph.canvas.getFontMetrics().getAscent()/2));
	}
	
	public void drawStructure(Graphics g, boolean moving) {
		width /= 2;
		height /= 2;

		draw(g, moving);

		width *= 2;
		height *= 2;
	}

	/**
	 * Adds a new edge into the offsetPoint "edgesIn".
	 */
	public void addEdgesIn(Edge edge) {
		edgesIn.add(edge);
	}

	/**
	 * Adds a new edge into the offsetPoint "edgesOut".
	 */
	public void addEdgesOut(Edge edge) {
		edgesOut.add(edge);
	}

	/**
	 * Removes the edge e from the offsetPoint "edgesIn".
	 */
	public void removeEdgesIn(Edge e) {
		edgesIn.remove(e);
	}

	/**
	 * Removes the edge e from the offsetPoint "edgesOut".
	 */
	public void removeEdgesOut(Edge e) {
		edgesOut.remove(e);
	}

	/**
	 * Returns "edgesIn" & "edgesOut".
	 */
	public ArrayList<Edge>[] getAllEdges() {
		ArrayList[] tmp = {edgesIn, edgesOut};
		return tmp;
	}

	/**
	 * Sets the index.
	 */
	public void setIndex(int i) {
		index = i;
	}

	/**
	 * Gets the index.
	 */
	public int getIndex() {
		return index;
	}
	
	/** set the name of this node */
	public void setLabel(String name) {
		if (name.equals(""))
			label[0] = "Node " + getIndex();
		else
			label[0] = name;
	}

	/**
	 * Returns whether the point (x2, y2) is in this node.
	 */
	public boolean contains(Point p) {
		// Return false if not initialized yet.
		if (width == 0 || height == 0)
			return false;

		int dx = (int) Math.abs(pos.x - p.x);
		int dy = (int) Math.abs(pos.y - p.y);
		if(shape == GraphConsts.OVAL) {
			//get the radiuses squared of the oval
			double radX2 = (width/2.0)*(width/2.0);
			double radY2 = (height/2.0)*(height/2.0);
			//check if it's inside the oval
			//formula for oval is (x^2/a^2) + (y^2/b^2) = 1
			//test inside just test for <= 1
			return ((p.x-pos.x)*(p.x-pos.x)/radX2 + (p.y-pos.y)*(p.y-pos.y)/radY2 <= 1);
		} else if (shape == GraphConsts.RECT || shape == GraphConsts.ROUNDED_RECT) {
			return ((4 * dx * dx / (width * width) + 4 * dy * dy / (height * height)) <= (1 / 2.0));
		} else {
			Point p1 = new Point(pos.x, pos.y - height / (float) 1.5);
			Point p2 = new Point(pos.x + width / (float) 1.5, pos.y);
			Point p3 = new Point(pos.x, pos.y + height / (float) 1.5);
			Point p4 = new Point(pos.x - width / (float) 1.5, pos.y);
			return ((p.leftOf(p1, p4) < 0) && (p.leftOf(p4, p3) < 0) && (p.leftOf(p3, p2) < 0) && (p.leftOf(p2, p1) < 0));
		}
	}

	/**
	 * Returns whether the node is inside the rectangle given by 
	 * the points a and b, where a is the upper left hand point,
	 * and b is the lower right hand point of the rectangle.
	 */
	public boolean isInRect(Point a, Point b) {
		// Find the bounding box of this node
		Point upperLeft = new Point(pos.x - width / 2, pos.y - height / 2);
		Point lowerRight = new Point(pos.x + width / 2, pos.y + height / 2);

		if (shape == GraphConsts.DIAMOND) {
			upperLeft.x = pos.x - width / 1.5f;
			upperLeft.y = pos.y - height / 1.5f;
			lowerRight.x = pos.x + width / 1.5f;
			lowerRight.y = pos.y + height / 1.5f;
		}
		return ((upperLeft.x > a.x) && (lowerRight.x < b.x) && (upperLeft.y > a.y) && (lowerRight.y < b.y));
	}

	/**
	 * Moves itself by dx, dy.
	 */
	public void translate(float dx, float dy) {
		pos.translate(dx, dy);
		for (int i = 0; i < edgesIn.size(); i++) {
			edgesIn.get(i).updateLength();
		}
		for (int i = 0; i < edgesOut.size(); i++) {
			edgesOut.get(i).updateLength();
		}
	}

	/**
	 * Moves to p.
	 */
	public void move(Point p) {
		pos.move(p);
		for (int i = 0; i < edgesIn.size(); i++) {
			edgesIn.get(i).updateLength();
		}
		for (int i = 0; i < edgesOut.size(); i++) {
			edgesOut.get(i).updateLength();
		}
	}

	/**
	 * Finds the intersection point of this node and the "start" node
	 */
	public Point findBorder(Node start) {
		if (shape == GraphConsts.OVAL || shape == GraphConsts.ROUNDED_RECT) {
			float dx = pos.x - start.pos.x;
			float dy = pos.y - start.pos.y;

			float t = (float) (0.5 / Math.sqrt((dx * dx) / (width * width) + (dy * dy) / (height * height)));

			return new Point(pos.x - dx * t, pos.y - dy * t);
		}
		Point a = new Point(pos);
		Point b = new Point(start.pos.x, start.pos.y);
		Point c = new Point(0, 0);
		Point d = new Point(0, 0);

		Point p1 = new Point(0, 0);
		Point p2 = new Point(0, 0);
		Point p3 = new Point(0, 0);
		Point p4 = new Point(0, 0);

		if (shape == GraphConsts.RECT) {
			// The bounding points of the rectangle
			p1.move(pos.x - width / 2, pos.y - height / 2);
			p2.move(p1.x, pos.y + height / 2);
			p3.move(pos.x + width / 2, p2.y);
			p4.move(p3.x, p1.y);
		} else {
			// The bounding points of the diamond
			p1.move(pos.x, pos.y - height / (float) 1.5);
			p2.move(pos.x - width / (float) 1.5, pos.y);
			p3.move(pos.x, pos.y + height / (float) 1.5);
			p4.move(pos.x + width / (float) 1.5, pos.y);
		}

		float tmp1 = b.leftOf(p3, p1);

		if (tmp1 < 0) {
			// nw -> sw -> se
			float tmp2 = b.leftOf(p4, p2);
			if (tmp2 < 0) {
				// sw -> se
				c.move(p2);
				d.move(p3);
			} else if (tmp2 > 0) {
				// nw -> sw
				c.move(p1);
				d.move(p2);
			} else {
				// sw
				return p2;
			}
		} else if (tmp1 > 0) {
			// nw -> ne -> se
			float tmp2 = b.leftOf(p2, p4);
			if (tmp2 < 0) {
				// nw -> ne
				c.move(p4);
				d.move(p1);
			} else if (tmp2 > 0) {
				// ne -> se
				c.move(p3);
				d.move(p4);
			} else {
				// ne
				return p4;
			}
		} else {
			// nw or se
			float tmp2 = b.leftOf(p2, p4);
			if (tmp2 < 0) {
				// nw
				return p1;
			} else if (tmp2 > 0) {
				// se
				return p3;
			} else {
				// nowhere
				return a;
			}
		}

		float[][] A = new float[2][2];
		A[0][0] = b.y - a.y;
		A[0][1] = a.x - b.x;
		A[1][0] = d.y - c.y;
		A[1][1] = c.x - d.x;

		float[] B = new float[2];
		B[0] = a.x * b.y - b.x * a.y;
		B[1] = c.x * d.y - d.x * c.y;

		float denominator = A[0][0] * A[1][1] - A[0][1] * A[1][0];

		float[][] invA = new float[2][2];
		invA[0][0] = A[1][1] / denominator;
		invA[0][1] = -A[0][1] / denominator;
		invA[1][0] = -A[1][0] / denominator;
		invA[1][1] = A[0][0] / denominator;

		Point ans = new Point(invA[0][0] * B[0] + invA[0][1] * B[1],
							  invA[1][0] * B[0] + invA[1][1] * B[1]);

		return ans;
	}

	public boolean equals(Entity ent) {
		if (ent.type == GraphConsts.NODE) {
			return (((Node) ent).index == index && pos.equals(((Node) ent).pos));
		}
		return false;
	}

	public ArrayList<Node> getNeighbours() {
		ArrayList<Node> tmp = new ArrayList<Node>(edgesOut.size());
		for (int i = 0; i < edgesOut.size(); i++) {
			tmp.add((edgesOut.get(i)).otherNode(this));
		}
		return tmp;
	}

	/**
	 * Returns true if this node is connected to another node.
	 */
	public boolean isConnected() {
		return (edgesIn.size() > 0) || (edgesOut.size() > 0);
	}

	/**
	 * Returns a text representation of the node.
	 */
	public String toString() {
		return label[0];
	}

	public Color getInternalColor() {
		return internalColor;
	}

	public void setInternalColor(Color internalColor) {
		this.internalColor = internalColor;
	}
}