/*
 * File: HelpCanvas.java
 * Jun 12, 2000
 */
//Updated to SWING: Saleema Amershi, summer 2003

package AIspace.graphToolKit.help;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Stroke;

import javax.swing.JPanel;

/**
 * draws the canvas for the legend
 */
public abstract class HelpCanvas extends JPanel {
	protected final int LINE_SEPARATION = 35;
	protected final int GROUP_SEPARATION = 60;
	protected final int SUBGROUP_SEPARATION = 50;

	private int lineWidth;
	private boolean isAntiAliasingEnabled;
	private float[] dashedPattern = {10, 10, 10, 10};
	
	public HelpCanvas(boolean isAntiAliasingEnabled) {
		super();
		this.lineWidth = 2;
		this.isAntiAliasingEnabled = isAntiAliasingEnabled;
		setBackground(Color.white);
	}

	/**
	 * paint the canvas (This needs to be overridden in subclasses)
	 */
	public void paint(Graphics graphics) {
		if(isAntiAliasingEnabled) {
		    ((Graphics2D) graphics).setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		}
		super.paint(graphics);
	}

	/**
	 * draw an edge with an arrow head
	 * @param graphics - the graphics object to draw to
	 * @param x - the x center of the line
	 * @param y - the y center of the line
	 * @param length - the length of the line
	 * @param color - the color of the line
	 * @param isDashed - true if the line is a dashed one
	 * @param isLineBold - true if the line is bolded
	 * @param isArrowHeadFilled - true if the arrow head is filled in
	 */
	public void drawEdge(Graphics graphics, int x, int y, int length, Color color, boolean isDashed, boolean isLineBold, boolean isArrowHeadFilled) {
		Stroke oldStroke = null;
		graphics.setColor(color);
		
		int extraWidth = lineWidth;
		if(isLineBold)
			extraWidth += 4;		
		
		if(isDashed) {
			//if the arc is no forgetting, draw it as dashed
			oldStroke = ((Graphics2D) graphics).getStroke();
			((Graphics2D) graphics).setStroke(new BasicStroke(extraWidth, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 10, dashedPattern, 0));
		}
	    
		//draw the line
		graphics.drawLine( x-length/2, y, x+length/2-6, y );
		if(!isDashed) {
			//make the line a bit thicker to match dashed one
			extraWidth--;
			int shiftDown = extraWidth / 2;
			while(extraWidth > 0) {
				graphics.drawLine( x-length/2, y+shiftDown-extraWidth, x+length/2-6, y+shiftDown-extraWidth );
				extraWidth--;
			}
		}

		if(isDashed) {
			//reset the graphics to be normal
			((Graphics2D) graphics).setStroke(oldStroke);
		}
		
		//draw the arrow
		if(!isArrowHeadFilled) {
			graphics.drawLine(x+length/2-6, y-6, x+length/2-6, y+6);
			graphics.drawLine(x+length/2-6, y+6, x+length/2+8, y);
			graphics.drawLine(x+length/2-6, y-6, x+length/2+8, y);
		} else {
			int[] xPoints = {x+length/2-6, x+length/2+8, x+length/2-6};
			int[] yPoints = {y+6, y, y-6};
			graphics.fillPolygon(xPoints, yPoints, 3);
		}
		
		graphics.setColor(Color.BLACK);
	}
	
	/**
	 * draw an edge with no arrow head
	 * @param graphics - the graphics object to draw to
	 * @param x - the x center of the line
	 * @param y - the y center of the line
	 * @param length - the length of the line
	 * @param color - the color of the line
	 * @param isDashed - true if the line is a dashed one
	 * @param isLineBold - true if the line is bolded
	 */
	public void drawEdge(Graphics graphics, int x, int y, int length, Color color, boolean isDashed, boolean isLineBold) {
		Stroke oldStroke = null;
		graphics.setColor(color);
		
		int extraWidth = lineWidth;
		if(isLineBold)
			extraWidth += 4;		
		
		if(isDashed) {
			//if the arc is no forgetting, draw it as dashed
			oldStroke = ((Graphics2D) graphics).getStroke();
			float[] dashedPattern = { 10, 10, 10, 10 };
			((Graphics2D) graphics).setStroke(new BasicStroke(extraWidth, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 10, dashedPattern, 0));
		}
	    
		//draw the line
		graphics.drawLine( x-length/2, y, x+length/2, y );
		if(!isDashed) {
			//make the line a bit thicker to match dashed one
			extraWidth--;
			int shiftDown = extraWidth / 2;
			while(extraWidth > 0) {
				graphics.drawLine( x-length/2, y+shiftDown-extraWidth, x+length/2, y+shiftDown-extraWidth );
				extraWidth--;
			}
		}

		if(isDashed) {
			//reset the graphics to be normal
			((Graphics2D) graphics).setStroke(oldStroke);
		}
		
		graphics.setColor(Color.BLACK);
	}

	public void drawRectangleNode(Graphics graphics, int x, int y, int width, int height, Color color, boolean isBold, boolean isFilled) {
		width *= 1.2;
		height *= 1.2;
		int extraWidth = lineWidth;
		if(isBold)
			extraWidth += 2;		
		graphics.setColor(color);
		graphics.fillRect(x - width/2, y - height/2, width, height);
		if(!isFilled) {
			graphics.setColor(Color.white);
			graphics.fillRect(x - width/2 + extraWidth, y - height/2 + extraWidth, width - 2*extraWidth, height - 2*extraWidth);
		}
		graphics.setColor(Color.black);
	}
	 
	public void drawOvalNode(Graphics graphics, int x, int y, int width, int height, Color color, boolean isBold, boolean isFilled) {
		height *= 1.3;
		width *= 1.3;
		int extraWidth = lineWidth;
		if(isBold)
			extraWidth += 2;
		graphics.setColor(color);
		graphics.fillOval(x - width/2, y - height/2, width, height);
		if(!isFilled) {
			graphics.setColor(Color.white);
			graphics.fillOval(x - width/2 + extraWidth, y - height/2 + extraWidth, width - 2*extraWidth, height - 2*extraWidth);
		}
		graphics.setColor(Color.black);
	}
	 
	public void drawDiamondNode(Graphics graphics, int x, int y, int width, int height, Color color, boolean isBold, boolean isFilled) {
		int[] xVals = {x, (int)(x + width/1.5), x, (int)(x - width/1.5)};
		int[] yVals = {(int)(y - height/1.5), y, (int)(y + height/1.5), y};
		
		int extraWidth = lineWidth;
		if(isBold)
			extraWidth += 2;

		graphics.setColor(color);
		graphics.fillPolygon(xVals, yVals, 4);
		if(!isFilled) {
			graphics.setColor(Color.white);
			xVals[1] -= 2*extraWidth;
			xVals[3] += 2*extraWidth;
			yVals[0] += extraWidth;
			yVals[2] -= extraWidth;
			graphics.fillPolygon(xVals, yVals, 4);
		}
		graphics.setColor(Color.black);
	}
}