/**
 * BackPropagation.java
 * Jul 20, 2002
 */


package AIspace.neural;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;

import AIspace.graphToolKit.elements.Edge;
import AIspace.graphToolKit.elements.Node;
import AIspace.neural.elements.NeuralEdge;
import AIspace.neural.elements.NeuralNode;
import AIspace.neural.intList.IntList;


/**
 * implements the Back Propagation algorithm for Neural Network Learning
 * @author Leslie Tung, Shinjiro Sueda, Joseph Roy Santos
 */
public class BackPropagation {
   //
   // Fields
   //

	private NeuralGraph graph;

	private Double parameters[];
	private int numberOfParameters;
	private double learningRate;
	private ArrayList<Hashtable<Integer, String>> examples;
	private ArrayList<Hashtable<Integer, String>> testExamples;
	private Node outputIndex[];
	private Node inputNodes[];
	private int numberOfInputNodes;
	private int numberOfOutputNodes;
	
	private IntList xVals;		// x values of points to draw
	private IntList yVals;		// y values of points to draw
	
	// for the test examples plot
	private IntList xTestVals;
	private IntList yTestVals;

	private int iterationCount;
	private double currErr;
	private double currTestErr;
	
	// stores the previous weights (for momentum)
	private double[] momentumWeights;
	
	// ArrayList of Lists that store references to each "layer"
	// index 0 is always the output layer
	// index size()-1 is always the hidden layer below inputs
	private ArrayList<ArrayList<NeuralNode>> nodeLayers;
	
	private double[] midrange;
	private double[] range;
	private boolean standardize;
	
	private Double stoppingCondition = null;


	//
	// Methods
	//
	
	/**
	 * construct the BackPropagation object
	 * @param g			  the NeuralGraph
	 * @param outputNodes the offsetPoint of output nodes
	 * @param standardize determines if the inputs are standardized
	 * @param rate		  the learning rate
	 * @param f			  the frame to display the output
	 */
	public BackPropagation(NeuralGraph g, ArrayList<Node> outputNodes, boolean standardize) {
		graph = g;
		
		range = null;
		midrange = null;
		this.standardize = standardize;
		
		momentumWeights = null;
		// converts the ExampleList into a ArrayList
		
		examples = graph.getExampleList().getTrainingArrayList();
		testExamples = graph.getExampleList().getTestArrayList();
		
		numberOfParameters = 0;
		numberOfInputNodes = 0;

		numberOfOutputNodes = outputNodes.size();
		outputIndex = new Node[numberOfOutputNodes];
		for (int i = 0; i < numberOfOutputNodes; i++) {
			outputIndex[i] = outputNodes.get(i);
		}

		NeuralNode tmpNode;
		Iterator<Node> nodeItr = graph.getNodes();
		while (nodeItr.hasNext()) {
			tmpNode = (NeuralNode)nodeItr.next();
			if (tmpNode.getNumParentNodes() > 0)  {
				// not an input nodes
				numberOfParameters++;
			}
			if (tmpNode.getNumParentNodes() == 0 && tmpNode.getNumChildrenNodes()!=0) {
				numberOfInputNodes++;
			}
		}
										  
		numberOfParameters = numberOfParameters + graph.numEdges();
		parameters = new Double[numberOfParameters];
		inputNodes = new Node[numberOfInputNodes];

		int count1 = 0;
		int count3 = 0;
		
		NeuralNode[] inputArray = new NeuralNode[graph.getInputNodes().size()];
		NeuralNode[] outputArray = new NeuralNode[graph.getOutputNodes().size()];
		int inputCount = 0;
		int outputCount = 0;
		
		// initialize nodeLayers index 0
		nodeLayers = new ArrayList<ArrayList<NeuralNode>>();
		ArrayList<NeuralNode> outputList = new ArrayList<NeuralNode>();
		nodeLayers.add(outputList);
		
		// find the output nodes and input nodes, put them in the parameter list		
	  
		nodeItr = graph.getNodes();
		while (nodeItr.hasNext()) {
			tmpNode = (NeuralNode)nodeItr.next();
			// get output nodes
			if ((tmpNode.getNumParentNodes() > 0) && (tmpNode.getNumChildrenNodes() == 0))  {
				parameters[count1] = new Double(tmpNode.getCurrentParaValue());
				tmpNode.setParameterIndex(count1);
				outputArray[outputCount] = tmpNode;
				// add to nodeLayer index 0;
				outputList.add(tmpNode);
				outputCount++;
				count1++;
			}
			// get input nodes
			else if ((tmpNode.getNumParentNodes() == 0) && (tmpNode.getNumChildrenNodes() > 0)) {
				inputNodes[count3] = tmpNode;
				inputArray[inputCount] = tmpNode;
				inputCount++;
				count3++;
			}
		}
		
		// call a helper function to add edges then hidden nodes recursively
		// so this will iterate from output nodes up to the edges, then to the hidden nodes.
		// Recurse until the set of "hidden nodes" is actually the input nodes.
		orderBackpropParameters(count1, inputArray, outputArray, 1);

		xVals = new IntList();
		yVals = new IntList();
		xTestVals = new IntList();
		yTestVals = new IntList();
		if (standardize) {
			solveForStandardize();	
		}
		currErr = totalError( parameters );
		currTestErr = totalTestError (parameters);
		addPoint( 0, (int)(currErr*100) );
		addTestPoint(0, (int)(currTestErr*100) );
		graph.updatePlot();
		iterationCount = 0;
	}
	
	/**
	 * Find the midrange and range (for standardization). Note that this will only 
	 * rely on the training set.
	 */
	private void solveForStandardize() {
		midrange = new double[numberOfInputNodes];
		range = new double[numberOfInputNodes];
		double[] minInput = new double[numberOfInputNodes];
		double[] maxInput = new double[numberOfInputNodes];
		for (int i = 0; i < examples.size(); i++) {
			// get one example from the training set
			Hashtable<Integer, String> oneExample = examples.get(i);
			// set the parameter value of the input nodes to the example
			// from the training set. 
			for (int j = 0; j < numberOfInputNodes; j++) {
				Double inputValue = Double.valueOf(oneExample.get(inputNodes[j].getIndex()));
				if (i > 0) {
					maxInput[j] = Math.max(maxInput[j], inputValue.doubleValue());
					minInput[j] = Math.min(minInput[j], inputValue.doubleValue());
				} else {
					maxInput[j] = inputValue.doubleValue();
					minInput[j] = inputValue.doubleValue();	
				}
			}
		}
		for (int i = 0; i < numberOfInputNodes; i++) {
			midrange[i] = (maxInput[i] + minInput[i])/2;
			range[i] = maxInput[i] - minInput[i];	
		}
	}
	
 	/**
 	 * Helper function that records the parameters into the parameter array
 	 * 
 	 */  
	public void orderBackpropParameters(int count, NeuralNode[] inputArray, NeuralNode[] outputArray, int layerLevel) {
		// get the edges that
		// are connected to the 'output' parents
		ArrayList<Edge> edgesIn = null;
		
		if (count >= parameters.length) return;
		
		for (int i = 0; i < outputArray.length; i++) {
			edgesIn = outputArray[i].getAllEdges()[0];
			for (int j = 0; j < edgesIn.size(); j++) {
				NeuralEdge edge = (NeuralEdge)edgesIn.get(j);
				parameters[count] = new Double(edge.getCurrentParaValue());
				edge.setParameterIndex(count);
	
				count++;	
			}	
		}
		// do the same thing for the parents (the 'hidden' nodes for this layer)
		// this, right now, will rely on the total connecting between nodes (ie
		// we only need to check one node as it connects to all the nodes in the
		// hidden layer.
		// To make this general, just go through the outputArray and get all the parents
		// of all the nodes, and then check for duplicates. This will be slow. 
		// (Close to O(n^2))

		
		ArrayList<NeuralNode> inputs = new ArrayList<NeuralNode>();
		for (NeuralNode node: inputArray)
		{
			inputs.add(node);
		}
		
		ArrayList<NeuralNode> parents = new ArrayList<NeuralNode>();
		for (int i = 0; i < outputArray.length; i++) {

			// get the parents
			ArrayList<Node> parentNodes = outputArray[i].getParentNodes();
			// put the parents into a ArrayList
			for (int j = 0; j < parentNodes.size(); j++) {
				Iterator<Node> nodeItr = graph.getNodes();
				while (nodeItr.hasNext()) {
					NeuralNode node = (NeuralNode) nodeItr.next();
					int indexOfParent = parentNodes.get(j).getIndex();
					if ((node.index == indexOfParent) && (!parents.contains(node)) && (!inputs.contains(node))) {
						parents.add(node);
						break;
					}
				}
			}
		}
		
		// check to see if the parentArray is empty.
		// if it is, then the hidden layer == input layer
		// and the recursion can terminate
		
		if (parents.size() == 0)
			return;
		
		// initialize the next element in nodeLayers
		// this will be a layer of hidden nodes of index layerLevel		
		ArrayList<NeuralNode> hiddenLayer = new ArrayList<NeuralNode>();	
				
		NeuralNode[] hiddenArray = new NeuralNode[parents.size()];
		for (int j = 0; j < hiddenArray.length; j++) {
			NeuralNode node = parents.get(j);
			if (count >= parameters.length) return;
		  	parameters[count] = new Double(node.getCurrentParaValue());
		  	node.setParameterIndex(count);
		  	hiddenArray[j] = node;	
			hiddenLayer.add(node);			
			count++;	
		}
		nodeLayers.add(hiddenLayer);
		
		// call this function recursively on the new hidden node array
		// tail-recursive!! 
		orderBackpropParameters(count, inputArray, hiddenArray, (layerLevel+1));							
	}

	
	/**
	 * the sigmoid function
	 */
	public static double sigmoidFunction(double x) {
		double netoutput;
	
		if (x < -45)
			netoutput = 0;
		else if (x > 45)
			netoutput = 1;
		else
			netoutput = 1 / (1 + Math.exp(-x));
		return netoutput;
	}
	
	/**
	 * The hyberbolic tangent function
	 */
	public static double tanhFunction(double x) {
		double e2x = Math.exp(2*x);
		return ((e2x-1)/(e2x+1));	
	}
	
	/**
	 * The exponential function
	 */
	public static double expFunction(double x) {
		return Math.exp(x);
	}

	
	/**
	 * the symbolic derivative of the sigmoid function
	 * (not used in this back propagation algorithm)
	 */
	public static double sigmoidDerivative(double x) {
		double temp = sigmoidFunction(x);
		return (temp * (1.0 - temp));
	}
	
	public double getCurrErr() {
		return currErr;
	}
	
	public double getCurrTestErr() {
		return currTestErr;
	}
	
	public double getStoppingError() {
		return stoppingCondition.doubleValue();
	}


	public int getIterationCount() {
		return iterationCount;
	}

	
	/**
	 * set the learning rate
	 */
	public void setLearningRate(double rate) {
		learningRate = rate;
	}

	
	/**
	 * access the learning rate
	 */
	public double getLearningRate() {
		return learningRate;
	}

	
	/**
	 * access the parameter values
	 */
	public Double[] getParameters() {
		return parameters;
	}
	
	/**
	 * calculate the value of the node with the graph's parameter settings
	 * (recursively, if applicable)
	 */
	public double getValue(int nodeIndex) {
		NeuralNode currNode;
		NeuralEdge currEdge;
		int parentNodeIndex;
		ArrayList<Node> parents;
		double tmpValue;
		currNode = (NeuralNode) graph.nodeFromIndex(nodeIndex);
		if (currNode == null) {
			// error
			return -9999.99;
		}

		double value = currNode.getCurrentParaValue();
		parents = currNode.getParentNodes();
		if (parents.size() == 0) {
			// simply return its input value
			return value;
		}
		
		// calculate each parent's value
		for (int i = 0; i < parents.size(); i++) {
			parentNodeIndex = parents.get(i).getIndex();		
			tmpValue = getValue(parentNodeIndex);
			currEdge = (NeuralEdge) graph.getEdge(parentNodeIndex, nodeIndex);
			if (currEdge != null) {
				value = value + (currEdge.getCurrentParaValue() * tmpValue);
			} else {
				// error
				return -9999.99;
			}						  
		}
		switch(currNode.functionType) {
			case NeuralNode.SIGMOID:
				return BackPropagation.sigmoidFunction(value);
			case NeuralNode.LINEAR:
				return value;	
			case NeuralNode.TANH:
				return BackPropagation.tanhFunction(value);	
			case NeuralNode.EXP:
				return BackPropagation.expFunction(value);	
			default:
				return 0;
		}
	}
	
	/**
	 * calculate the value of the node with the parameter values given
	 * (recursively, if applicable)
	 */
	public double getValue(int nodeIndex, Double[] para) {
		NeuralNode currNode;
		NeuralEdge currEdge;
		int parentNodeIndex;
		ArrayList<Node> parents;
		double tmpValue;
		int paraIndex;

		currNode = (NeuralNode) graph.nodeFromIndex(nodeIndex);
		if (currNode == null) {
			// error
			return -9999.99;
		}

		paraIndex = currNode.getParameterIndex();
		if (paraIndex == -1) {
			return currNode.getCurrentParaValue();
		}
		double value = para[paraIndex].doubleValue();
		  
		parents = currNode.getParentNodes();
		if (parents.size() == 0) {
			// simply return its input value
		} else {
			// calculate each parent's value
			for (int i = 0; i < parents.size(); i++) {
				parentNodeIndex = parents.get(i).getIndex();		
				tmpValue = getValue(parentNodeIndex, para);
				currEdge = (NeuralEdge) graph.getEdge(parentNodeIndex, nodeIndex);
				if (currEdge != null) {
					paraIndex = currEdge.getParameterIndex();
					value = value + para[paraIndex].doubleValue() * tmpValue;
				} else {
					// error
					return -9999.99;
				}
			}
		}

		switch(currNode.functionType) {
			case NeuralNode.SIGMOID:
				return BackPropagation.sigmoidFunction(value);
			case NeuralNode.LINEAR:
				return value;	
			case NeuralNode.TANH:
				return BackPropagation.tanhFunction(value);	
			case NeuralNode.EXP:
				return BackPropagation.expFunction(value);	
			default:
				return 0;
		}
	}


	/**
	 * iterate numIter iterations of backpropagation
	 */
	public Double[] nnlearn(int numIter, double rate, double momentum) {
		Double newPara[] = new Double[numberOfParameters];
		for (int i= 0; i < numberOfParameters; i++) {
			newPara[i] = new Double(parameters[i].doubleValue());
		}

		int i = 0;
		while( i < numIter ) {
			iterationCount++;
			newPara =  updateEach(newPara, rate, momentum);
		
			for (int j= 0; j < numberOfParameters; j++) {
				parameters[j] = new Double(newPara[j].doubleValue());
			}
			NeuralNode tmpNode;
			int paraIndex;
			Iterator<Node> nodeItr = graph.getNodes();
			while (nodeItr.hasNext()) {
				tmpNode = (NeuralNode)nodeItr.next();
				paraIndex = tmpNode.getParameterIndex();
				if (paraIndex != -1) {
					tmpNode.setCurrentParaValue(parameters[paraIndex].doubleValue());
				}
			}

			Iterator<Edge> itr = graph.getEdges();
			NeuralEdge edge;
			while(itr.hasNext()) {
				edge = (NeuralEdge) itr.next();
				paraIndex = edge.getParameterIndex();
				edge.setCurrentParaValue(parameters[paraIndex].doubleValue());
			}
			
			i++;
		}

		currErr = totalError( parameters );
		currTestErr = totalTestError (parameters);
		addPoint( iterationCount, (int)(currErr*100) );
		addTestPoint( iterationCount, (int)(currTestErr*100) );
		graph.updatePlot();

		// set the parameters back to the nodes and edges
		return newPara;
	}				

	
	/**
	 * update each of the node and edge to the most current parameter values
	 */
	public Double[] updateEach(Double[] para, double rate, double momentum) {
		Double newPara[] = new Double[para.length];
		Double tmpPara[] = new Double[para.length];
		// check to see if this is a new run, which would be if
		// momentumWeights is null. If it is, initialize momentumWeights
		if (momentumWeights ==  null) {
			momentumWeights = new double[para.length];
			for (int i = 0; i < momentumWeights.length; i++) {
				momentumWeights[i] = 0;	
			}
		}	
		// copy the contents of the old parameters into tmpPara
		for (int i = 0; i < para.length; i++) {
		  tmpPara[i] = new Double(para[i].doubleValue()); 
		}
		learningRate = rate;
		double[] derrors = solvePartials(tmpPara);
		// update the output nodes and edges going into the output nodes  
		// get the output layer and update the weights
		ArrayList<NeuralNode> outputLayer = nodeLayers.get(0);
		for (int i = 0; i < outputLayer.size(); i++) {
			NeuralNode output = outputLayer.get(i);
			int paraIndex = output.getParameterIndex();
	
			double derror = derrors[paraIndex];
			double delta = (rate*derror) + (momentum*momentumWeights[paraIndex]);
			newPara[paraIndex] = new Double(para[paraIndex].doubleValue() + delta );
			momentumWeights[paraIndex] = delta;
			// get the output edges and update the weights
			
			ArrayList<Edge> outputEdges = output.getAllEdges()[0];
			for (int j = 0; j < outputEdges.size(); j++) {
				NeuralEdge outputEdge = (NeuralEdge) outputEdges.get(j);
				int paraIndexEdge = outputEdge.getParameterIndex();
				double derrorEdge =  derrors[paraIndexEdge];
				delta = (rate*derrorEdge)+(momentum*momentumWeights[paraIndexEdge]);
				newPara[paraIndexEdge] = new Double(para[paraIndexEdge].doubleValue() + delta );
				momentumWeights[paraIndexEdge] = delta;
			}
		}
	  

		// update the hidden nodes and edges going into the hidden nodes
		for (int i = 1; i < nodeLayers.size(); i++) {
			// get the hidden nodes
	
			ArrayList<NeuralNode> hiddenLayer = nodeLayers.get(i);
			for (int j = 0; j < hiddenLayer.size(); j++) {
				NeuralNode output = hiddenLayer.get(j);
				int paraIndex = output.getParameterIndex();
				double derror =  derrors[paraIndex];
				double delta = (rate*derror) + (momentum*momentumWeights[paraIndex]);
				newPara[paraIndex] = new Double(para[paraIndex].doubleValue() + delta);
				momentumWeights[paraIndex] = delta;
				// get the output edges and update the weights
	
				ArrayList<Edge> hiddenEdges = output.getAllEdges()[0];
				for (int k = 0; k < hiddenEdges.size(); k++) {
					NeuralEdge outputEdge = (NeuralEdge) hiddenEdges.get(k);
					int paraIndexEdge = outputEdge.getParameterIndex();
					double derrorEdge =  derrors[paraIndexEdge];
					delta = (rate*derrorEdge)+(momentum*momentumWeights[paraIndexEdge]);
					newPara[paraIndexEdge] = new Double(para[paraIndexEdge].doubleValue() +delta );
					momentumWeights[paraIndexEdge] = delta;
				}
			}
		}

		return newPara;
	}
	
	
	/**
	 * Function to solve for a node's output. 
	 * This is given an example, and the node to be solved for.
	 * 
	 */
	public double solveNodeOutput(int index, Hashtable<Integer, String> example, Double[] para) {
		// set the parameter value of the input nodes to the example
 		double value;
 		int nodeIndex;
 		//get the list of parameters
 		String[] params = graph.getExampleList().parameters;
		for (int j = 0; j < numberOfInputNodes; j++) {
			nodeIndex = -1;
			NeuralNode node = (NeuralNode) inputNodes[j];
			//get the index of the node in the parameter list
			for(int i = 0; i < params.length; i++) {
				if(node.getLabel().equals(params[i])) {
					nodeIndex = i;
					break;
				}
			}
			Double inputValue = Double.valueOf(example.get(new Integer(nodeIndex)));
			
				
			value = inputValue.doubleValue();
			if (standardize) {
				value = (value - midrange[j])/(range[j]/2);	
			}
			node.setCurrentParaValue(value);
		}	

		// get the value
		return getValue(index, para);	
	}

	/**
	 * Function to solve for a node's output. 
	 * This is given an example, and the node to be solved for.
	 * 
	 */
	public double solveNodeOutput(int index, Double[] values) {
		// set the parameter value of the input nodes to the example
 
		NeuralNode node;
		double value;
		for (int i = 0; i < numberOfInputNodes; i++) {
			node = (NeuralNode)inputNodes[i];
			value = values[i].doubleValue();
			if (standardize) {
				value = (value - midrange[i])/(range[i]/2);	
			}
			node.setCurrentParaValue(value);
		}	

		// get the value	
		return getValue(index);
	}
	
	/**
	 * Helper function that solves all the nodes' outputs and saves them
	 * to the node
	 */
	private void storeAllNodeOutputs(ArrayList<NeuralNode> outputNodes, Hashtable<Integer, String> oneExample, Double[] para) {
 		double value;
 		int nodeIndex;
 		//get the list of parameters
 		String[] params = graph.getExampleList().parameters;
		for (int j = 0; j < numberOfInputNodes; j++) {
			nodeIndex = -1;
			NeuralNode node = (NeuralNode) inputNodes[j];
			//get the index of the node in the parameter list
			for(int i = 0; i < params.length; i++) {
				if(node.getLabel().equals(params[i])) {
					nodeIndex = i;
					break;
				}
			}
			Double inputValue = Double.valueOf(oneExample.get(new Integer(nodeIndex)));
				
			value = inputValue.doubleValue();
			if (standardize) {
				value = (value - midrange[j])/(range[j]/2);	
			}
			node.setCurrentParaValue(value);
		}
		for (int j = 0; j < outputNodes.size(); j++) {
			// get an output node
			NeuralNode output = outputNodes.get(j);
			// get the predicted output from the neural net
			storeNodeOutput(output.index,  para);			
		}		
	}
	
	/**
	 * Helper function that solves a specific node and stores its output.
	 */
	public double storeNodeOutput(int nodeIndex, Double[] para) {
		NeuralNode curNode;
		NeuralEdge currEdge;
		int parentNodeIndex;
		ArrayList<Node> parents;
		double tmpValue;
		int paraIndex;

		curNode = (NeuralNode) graph.nodeFromIndex(nodeIndex);
		if (curNode == null) {
			// error
			return -9999.99;
		}

		paraIndex = curNode.getParameterIndex();
		if (paraIndex == -1) {
			return curNode.getCurrentParaValue();
		}
		double value = para[paraIndex].doubleValue();
		  
		parents = curNode.getParentNodes();
		if (parents.size() == 0) {
			// simply return its input value
		} else {
			// calculate each parent's value
			for (int i = 0; i < parents.size(); i++) {
				parentNodeIndex = parents.get(i).getIndex();		
				tmpValue = storeNodeOutput(parentNodeIndex, para);
				currEdge = (NeuralEdge) graph.getEdge(parentNodeIndex, nodeIndex);
				if (currEdge != null) {
					paraIndex = currEdge.getParameterIndex();
					value = value + para[paraIndex].doubleValue() * tmpValue;
				} else {
					// error
					return -9999.99;
				}
			}
		}
		
		switch(curNode.functionType) {
			case NeuralNode.SIGMOID:
				value = BackPropagation.sigmoidFunction(value);
				break;
			case NeuralNode.LINEAR:
				break;
			case NeuralNode.TANH:
				value = BackPropagation.tanhFunction(value);
				break;
			case NeuralNode.EXP:
				value = BackPropagation.expFunction(value);
				break;
		}
		curNode.setOutputValue(value);
		return value;
	}
	/**
	 * Function to compute the partial derivatives of the error wrt to the outputs.
	 * This should backpropagate the error and accumulate the derivatives.
	 * Returns an array of the same size as para that stores the accumulated
	 * derivatives. This is to avoid iterating over the examples again.
	 */
	
	private double[] solvePartials(Double[] para) {
		// initialize the accumulator array
		double[] derrAccumulator = new double[para.length];
		// initialize the list of error pairs
		// if it hasn't been initalized yet, set a ArrayList
		// and note that it has been initialized
		Iterator<Node> itr = graph.getNodes();
		while(itr.hasNext()) {
			NeuralNode node = (NeuralNode) itr.next();
			if (node.currentError == null) {
				node.currentError = new double[examples.size()];
			}
		}
		
		// go through all the examples
		for (int i = 0; i < examples.size(); i++) {
			Double actual;
			// get one example from the training set
			Hashtable<Integer, String> oneExample = examples.get(i);
			ArrayList<NeuralNode> outputNodes = nodeLayers.get(0);
			// solve all the nodes' outputs and save the outputs to the nodes
			// This is for efficiency purposes (ie avoid calling solveNodeOutput too many times
			storeAllNodeOutputs(outputNodes, oneExample, para);
			// solve the output nodes
			for (int j = 0; j < outputNodes.size(); j++) {
				// get an output node
				NeuralNode output = outputNodes.get(j);
				// find the index of the output node relative to the example list indexing
				int realIndex = graph.getInputNodes().size() + j;
				// get the actual value of the output node from the example
				actual = Double.valueOf(oneExample.get(new Integer(realIndex)));
				double actualvalue = actual.doubleValue();
				// get the predicted output from the neural net
				double predicted = output.getOutputValue();
				// find the delta value of the error (partial derivative)
				// depending on the activation function selected for the node
				double error = 0;
				switch(output.functionType) {
					case NeuralNode.SIGMOID:
						error = (actualvalue - predicted) * predicted * (1 - predicted);
						break;
					case NeuralNode.LINEAR:
						error = (actualvalue - predicted);
						break;
					case NeuralNode.TANH:
						error = (actualvalue - predicted) * (1 - (predicted * predicted));
						break;
					case NeuralNode.EXP:
						error = (actualvalue - predicted) * expFunction(predicted);
						break;
				}
				// record the derivative in the accumulator
				// this is to make the algorithm faster
				
				// accumulate the bias
				derrAccumulator[output.getParameterIndex()] += error;
				// accumulate the edges that feed into this node
				ArrayList<Edge> edgesIn = output.getAllEdges()[0];
				for (int k = 0; k < edgesIn.size(); k++) {
					NeuralNode sourceNode = (NeuralNode)((NeuralEdge)edgesIn.get(k)).otherNode(output);
					derrAccumulator[((NeuralEdge)edgesIn.get(k)).getParameterIndex()]
							+= error*getValue(sourceNode.index, para);	
				}
				
				// store the result in the result object
				// note that this doesn't remove any elements, just replaces them
				// this is to reduce initialization penalties, which can be significant
				// in large example sets
				output.currentError[i] = error;			
			}
	
			// solve the hidden nodes
			// this does the same thing as the previous code, except that
			// the error factor (actualvalue-predicted) is now a summation
			// of all the errors of the nodes the current node feeds into,
			// weighted by the edge weights.
			
			// get all the hidden layers
			for (int j = 1; j < nodeLayers.size(); j++) {
				// get one hidden layer
				ArrayList<NeuralNode> hiddenLayer = nodeLayers.get(j);
				for (int k = 0; k < hiddenLayer.size(); k++) {
					// get a hidden node
					NeuralNode hidden = hiddenLayer.get(k);
					// get the predicted value of the neural net
					double predicted = hidden.getOutputValue();
					// get all edges coming out of the node
					ArrayList<Edge> edgesOut = hidden.getAllEdges()[1];
					double weightCorrection = 0;
					
					// go through all the output edges and sum all their errors,
					// multiplied by their weights
					for (int l = 0; l < edgesOut.size(); l++) {
						int edgeIndex = ((NeuralEdge) edgesOut.get(l)).getParameterIndex();
						double weight = para[edgeIndex].doubleValue();
						NeuralNode otherNode = (NeuralNode) ((NeuralEdge) edgesOut.get(l)).otherNode(hidden);
						double otherError = otherNode.currentError[i];
						weightCorrection += (weight * otherError);
					}
					double error = 0;
					
					// find the delta of the current node, depending on the 
					// activation function of the node
					switch(hidden.functionType) {
						case NeuralNode.SIGMOID:
							error = weightCorrection * predicted * (1 - predicted);
							break;
						case NeuralNode.LINEAR:
							error = weightCorrection;
							break;
						case NeuralNode.TANH:
							error = weightCorrection * (1 - (predicted * predicted));
							break;
						case NeuralNode.EXP:
							error = weightCorrection * expFunction(predicted);
							break;
					}
					// record the derivative in the accumulator
					// this is to make the algorithm faster
					derrAccumulator[hidden.getParameterIndex()] += error;
					// accumulate the edges that feed into this node
					ArrayList<Edge> edgesIn = hidden.getAllEdges()[0];
					for (int l = 0; l < edgesIn.size(); l++) {
						NeuralNode sourceNode = (NeuralNode)((NeuralEdge) edgesIn.get(l)).otherNode(hidden);
						derrAccumulator[((NeuralEdge) edgesIn.get(l)).getParameterIndex()] += error*getValue(sourceNode.index, para);	
					}				
					// store the result
					hidden.currentError[i] = error;
				}
			}
		}
		// return the accumulator array
		return derrAccumulator; 
	}
 
	
	/**
	 * calculate the total sum of squares error for all the examples
	 * using the parameter values given.
	 * This returns the stopping criteria value (sum of squares value)
	 * by reference.
	 */
	public double totalError(Double[] para) {
		  // accumulator for the error
		double acc = 0.0;
		double stopAcc = 0.0;
		for (int i = 0; i < examples.size(); i++) {
			Double actual;
			// get one example from the training set
			Hashtable<Integer, String> oneExample = examples.get(i);
			// go through the output nodes
			for (int k = 0; k < numberOfOutputNodes; k++) {
				// get the index of the output node
				int realIndex = graph.getInputNodes().size() + k;

				int unadjustedIndex = outputIndex[k].getIndex();
				// get the actual value from the example list
				actual = Double.valueOf(oneExample.get(new Integer(realIndex)));
				// calculate the sum of squares error of this
				// and add it to the accumulator
				// this looks like summation over (estimatedvalue-actualvalue)^2
				// for every output node.
				if (graph.getErrorType() == NeuralGraph.AVERAGE_SUMSQUARES) {
					double nodeOutputErrSquared = Math.pow(solveNodeOutput(unadjustedIndex, oneExample, para)
											- actual.doubleValue(), 2);
					acc += (.5*nodeOutputErrSquared);
					stopAcc += nodeOutputErrSquared;
				} else if (graph.getErrorType() == NeuralGraph.SUMSQUARES) {
					double nodeOutputErrSquared = Math.pow(solveNodeOutput(unadjustedIndex, oneExample, para)
											- actual.doubleValue(), 2);
					acc += nodeOutputErrSquared;  
					stopAcc += nodeOutputErrSquared;			 	
				}
			}
		}
		stoppingCondition = new Double (stopAcc);

		if (graph.getErrorType() == NeuralGraph.AVERAGE_SUMSQUARES) {
			return acc/examples.size();
		}
		return acc;	
	}
	
	/**
	 * calculate the total sum of squares error for all the test examples
	 * using the parameter values given
	 */
	public double totalTestError(Double[] para) {
		double acc = 0.0;
		for (int i = 0; i < testExamples.size(); i++) {
			Double actual;
			Hashtable<Integer, String> oneExample = testExamples.get(i);
			for (int k = 0; k < numberOfOutputNodes; k++) {
				int realIndex = graph.getInputNodes().size()+k;
				int unadjustedIndex = outputIndex[k].getIndex();

				actual = Double.valueOf(oneExample.get(new Integer(realIndex)));
				if (graph.getErrorType() == NeuralGraph.AVERAGE_SUMSQUARES) {
					double nodeOutputErrSquared = Math.pow(solveNodeOutput(unadjustedIndex, oneExample, para)
											- actual.doubleValue(), 2);
					acc += (.5*nodeOutputErrSquared);
				} else if (graph.getErrorType() == NeuralGraph.SUMSQUARES) {
					double nodeOutputErrSquared = Math.pow(solveNodeOutput(unadjustedIndex, oneExample, para)
											- actual.doubleValue(), 2);
					acc += nodeOutputErrSquared;			  	
				}
			}
		}
		if (graph.getErrorType() == NeuralGraph.AVERAGE_SUMSQUARES) {
			return acc/examples.size();
		}
		return acc;	
	}


	/**
	 * Adds a point to the list of drawing points.
	 * This is for the training set
	 */
	public void addPoint( int x, int y ) {
		xVals.add( x );
		yVals.add( y );
	}
	
	/**
	 * Adds a point to the list of drawing points.
	 * This is for the test set
	 */	
	public void addTestPoint( int x, int y ) {
		xTestVals.add( x );
		yTestVals.add( y );
	}


	/**
	 * Returns values required for plotting.
	 * This is for the training set.
	 */
	public IntList[] getPlotVals() {
		IntList[] tmp = {xVals, yVals};
		return tmp;
	}

	/**
	 * Returns values required for plotting.
	 * This is for the test set.
	 */
	public IntList[] getTestPlotVals() {
		IntList[] tmp = {xTestVals, yTestVals};
		return tmp;
	}


	/**
	 * Clears the plotting points.
	 */
	public void clearPlotPts() {
		xVals = new IntList();
		yVals = new IntList();
		xTestVals = new IntList();
		yTestVals = new IntList();
	}	
}