/**
 * This class manages training and test set data for the Decision Tree Learning and Neural Networks applets.
 * It is also responsible for parsing and generating data files.
 * 
 * Note: Examples are currently stored in hash tables that map an index number to a data value.  The indices
 * seem to just number the values from zero to the number of parameters - 1.  I don't yet know why this class
 * doesn't just use a offsetPoint.
 * 
 * 
 */

package AIspace.neural;

import java.awt.Font;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.StringTokenizer;

import javax.swing.DefaultListModel;
import javax.swing.JList;
import javax.swing.ListSelectionModel;

import AIspace.XMLReader.Pair;
import AIspace.XMLReader.XMLBlock;
import AIspace.XMLReader.XMLParseException;
import AIspace.XMLReader.XMLTree;

public class ExampleList implements Cloneable {
	private DefaultListModel testModel;
	private DefaultListModel trainingModel;
	/**ArrayList of hash tables. Each hash table represents a training example.*/
	protected ArrayList<Hashtable<Integer, String>> trainingData;
	/**ArrayList of hash tables. Each hash table represents a test example. */
	protected ArrayList<Hashtable<Integer, String>> testData;
	/**Appears to be an array of integers from 0 to the number of parameters for hash table indexing*/
	protected Integer[] indices;
	/**This contains the names of the input and output values*/
	protected String[] parameters;
	/**This will identify whether each parameter is an input or output value*/
	protected boolean[] isInput;
	/**This is the probability that an unclassified raw data example will be added to the test set*/
	protected float probTest;
	/**This is a graphical list that probably shouldn't be here */
	private JList trainingList;
	/**This is a graphical  list that probably shouldn't be here */
	private JList testList;

	/**
		* Returns a new, empty ExampleList object.
		* 
		*/
	public ExampleList() {
		trainingData = new ArrayList<Hashtable<Integer, String>>(5);
		testData = new ArrayList<Hashtable<Integer, String>>(5);
		indices = null;
		parameters = null;
		isInput = null;
		probTest = 0.0F;

		trainingList = new JList();
		trainingList.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
		trainingModel = new DefaultListModel();
		trainingList.setModel(trainingModel);

		testList = new JList();
		testList.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
		testModel = new DefaultListModel();
		testList.setModel(testModel);

		//currently the font has to be these dimensions to line up correclty when displayed
		Font font = new Font("courier", Font.PLAIN, 12);
		trainingList.setFont(font);
		testList.setFont(font);
	}
	
	public ExampleList(ExampleList list) {
		trainingData = new ArrayList<Hashtable<Integer, String>>(list.getTrainingArrayList());
		testData = new ArrayList<Hashtable<Integer, String>>(list.getTestArrayList());
		indices = list.indices.clone();
		parameters = list.parameters.clone();
		isInput = list.isInput;
		probTest = list.probTest;
		
		trainingModel = list.trainingModel;
		testModel = list.testModel;
		trainingList = new JList();
		trainingList.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
		trainingModel = new DefaultListModel();
		trainingList.setModel(trainingModel);

		testList = new JList();
		testList.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
		testModel = new DefaultListModel();
		testList.setModel(testModel);

		//currently the font has to be these dimensions to line up correclty when displayed
		Font font = new Font("courier", Font.PLAIN, 12);
		trainingList.setFont(font);
		testList.setFont(font);
	}

	/**
	* Removes all elements from the example list and also removes all parameter names and input/output assignments
	* 
	*/
	public void reset() {
		indices = null;
		parameters = null;
		isInput = null;
		probTest = 0.0F;

		trainingModel.clear();
		testModel.clear();
	}
	
	public void resetAll() {
		reset();
		testData.clear();
		trainingData.clear();
		trainingList = getTrainingList();
		testList = getTestList();
	}

	/**
	* Parses a text file for data to add to the Example List.  This method is capable of handling raw
	* comma-delimited data and uses the following format:
	* 	Comments begin with the % character 
	*   Lines are terminated with a ; character for compatibility with other applets
	* 
	* 	Line indicator tags are letters followed by a colon.  The following line indicator tags are valid:
	* 
	* 	N:		Node	-	Used in other applets, ignored here
	* 	E:		Edge	-	Used in other applets, ignored here
	* 	T:		Comma delimited attribute title string
	* 	A:		Training Set Example (Comma Delimited) (or X: for backward compatibility)
	* 	B:		Test Set Example (Comma Delimited)
	* 
	* 	Lines that do not start with any of the following tags and are not comments are assumed
	* 	to be raw data examples.  These examples are randomly assigned to either the test set or the training set
	* 	based on the assignment probability setting of this class. (probTest)
	* 
	*/
	public String parse(String file) {
		String errorMessage = new String("");
		String inString;

		boolean rawDataFound = false;
		//True if raw data was found in the file and must be re-distributed to the test set randomly.
		int p1 = -1;
		int p2 = file.indexOf("\n");
		int lineCount = 0;
		try {
			while (p2 != -1) {
				inString = file.substring(p1 + 1, p2).trim();
				if (inString.length() > 2) {

					errorMessage = parseLine(inString.trim());

					if (errorMessage.equals("RawDataFound")) {
						rawDataFound = true;
					} else if(errorMessage.equals("You must identify parameters before loading examples")) {
						return errorMessage;
					} else if (errorMessage.length() > 0) {
						throw (new Exception());
					}
				}
				p1 = p2;
				p2 = file.indexOf("\n", p1 + 1);

				lineCount++;
			}

			//Re-distribute examples if raw unidentified data was found.
			if (rawDataFound) {

				int numToMove = (int) (trainingData.size() * probTest);
				int[] inds = new int[numToMove];

				if (numToMove > 0) {
					int n = (int) (trainingData.size() * Math.random());
					inds[0] = n;
					boolean different;

					for (int i = 1; i < numToMove; i++) {
						different = false;
						while (!different) {
							n = (int) (trainingData.size() * Math.random());
							different = true;
							for (int j = 0; j < i; j++) {
								if (inds[j] == n) {
									different = false;
									break;
								}
							}
						}
						inds[i] = n;
					}
				}

				moveToTestForLoad();
			}
			return "OK";

		} catch (Exception e) {
			String message = "";
			if (errorMessage.length() > 0) {
				message = "Error at line " + (lineCount + 1) + " -- " + errorMessage;
			} else {
				message = "Error at line " + (lineCount + 1) + " -- " + e.toString();
			}
			return message;
		}
	}

	/**
	* Parses an individual line of input from a file (or other data source).
	* Note that this method uses both the leading character and a colon to identify
	* line identification tags so that these tags don't interfere with raw data.
	* 
	* @see parse
	*/
	public String parseLine(String inString) {
		String errorMessage = new String("");
		int numParameters = getNumParameters();
		StringTokenizer tokenizer;
		int currTokenNum;

		if (inString.charAt(0) == 'T' && inString.charAt(1) == ':') {
			//Title String

			ArrayList<String> titles = new ArrayList<String>();
			String dataString = inString.substring(2);
			//Remove the line identifier tag
			tokenizer = new StringTokenizer(dataString);

			while (tokenizer.hasMoreTokens()) {
				titles.add((tokenizer.nextToken(",;")).trim());
			}

			numParameters = titles.size();

			setParameters(titles);
		} else if ((inString.charAt(0) == 'A' || inString.charAt(0) == 'B' || inString.charAt(0) == 'X') && inString.charAt(1) == ':') {
			//Return an error if there are no parameters specified.  This condition
			//should be checked before this method is called.  If there are no parameters,
			//then the ParameterInputDialog should be used to input them.

			if (numParameters == 0) {
				return ("You must identify parameters before loading examples");
			}

			String dataString = inString.substring(2);
			//Remove the line identifier tag

			tokenizer = new StringTokenizer(dataString);
			String[] values = new String[numParameters];

			currTokenNum = 0;
			while (tokenizer.hasMoreTokens() && currTokenNum < numParameters) {
				values[currTokenNum] = (tokenizer.nextToken(",;")).trim();
				currTokenNum++;
			}

			//Make sure enough values have been found
			if (currTokenNum < (numParameters - 1)) {
				return ("Not enough values found");
			}

			if (inString.charAt(0) == 'A' || inString.charAt(0) == 'X') {
				addTrainingExample(indices, values);
			} else {
				addTestExample(indices, values);
			}
		} else if ((inString.charAt(0) == '%') || (inString.charAt(0) == ' ') || (inString.charAt(0) == 'N')
				|| (inString.charAt(0) == 'E') || (inString.charAt(0) == '\n')) {
			// ignore comments and blank lines
		} else {
			//Assume that this is a line of raw data and try to parse it.
			//This is slightly different from the identified case above. No errors are reported
			//if this line turns out not to be a useful example.  If a proper example is found,
			//then it is added to the training set and "RawDataFound" is returned so that the 
			//calling parse method can re-distribute a random percentage to the test set if requred.

			//Return an error if there are no parameters specified.  This condition
			//should be checked before this method is called.  If there are no parameters,
			//then the ParameterInputDialog should be used to input them.
			if (numParameters == 0) {
				return ("You must identify parameters before loading examples");
			}

			tokenizer = new StringTokenizer(inString);
			String[] values = new String[numParameters];

			currTokenNum = 0;
			while (tokenizer.hasMoreTokens() && currTokenNum < numParameters) {
				values[currTokenNum++] = tokenizer.nextToken(", ;");
			}

			//Make sure enough values have been found
			if (currTokenNum < (numParameters - 1)) {
				return ("");
			}

			addTrainingExample(indices, values);

			return "RawDataFound";
		}
		return errorMessage;
	}

	/**
	 * Parses an XML file with <EXAMPLES> tags. It takes that subtree and
	 * reads training and test examples. The <EXAMPLES> subtree may or may not
	 * include parameter information. If it doesn't, this function returns 
	 * "You must identify parameters before loading examples" as an error string.
	 * The calling function should handle this by querying the user for parameter
	 * names.
	 */
	public String parseXML(XMLTree tree) {
		String errString = "";
		try {
			

			// get the <EXAMPLES> subtree
			XMLBlock exTree = tree.findNetworkTree("examples");
			// get the parameters
			ArrayList<XMLBlock> para = exTree.searchChildTag("parameter");
			// if there are no parameters, check to see if parameters have been
			// declared. If not, return a special signal string. If parameters
			// have been declared, continue.
			if (para == null) {
				if (getNumParameters() == 0) {
					return "You must identify parameters before loading examples";
				}
			} else {
				// process all the parameter declarations
				// note that since these have to be ordered as input, then output,
				// this sorts all the parameters into input and output vectors.
				ArrayList<String> paras = null;
				ArrayList<String> in = new ArrayList<String>();
				ArrayList<String> out = new ArrayList<String>();
				for (int i = 0; i < para.size(); i++) {
					XMLBlock onePara = para.get(i);
					String name = onePara.getText();
					// if there is no name, return an error
					if (name == "") {
						errString = "A parameter name has an invalid name.";
						throw (new Exception());
					}
					ArrayList<Pair> typeVect = onePara.getProperties();
					// if there is no type declaration, assume it is
					// an input
					if (typeVect == null) {
						in.add(name);
					} else {
						// get the type declaration pair
						Pair type = typeVect.get(0);
						if (type.name.equalsIgnoreCase("type")) {
							if (type.value.equalsIgnoreCase("output")) {
								out.add(name);
							} else {
								in.add(name);
							}
						} else {
							in.add(name);
						}
					}
				}
				// put the input and output vectors together

				// record up to what index the inputs are
				int inputIndex = in.size() - 1;
				// if the inputIndex is < 0, then there are
				// no inputs. Return an error
				if (inputIndex < 0) {
					errString = "No input parameters";
					throw (new Exception());
				}
				paras = in;
				// append the two vectors together
				for (int i = 0; i < out.size(); i++) {
					paras.add(out.get(i));
				}
				setParameters(paras);
				// set parameters as input or output
				for (int i = 0; i < paras.size(); i++) {
					setInputParam((i <= inputIndex), i);
				}
			}
			// It doesn't matter if there aren't any examples - DON'T throw
			// exceptions if this is the case. However, if there is a parameter
			// missing, do throw an exception.

			// get the training and test sets 
			ArrayList<XMLBlock> exampleSet = exTree.searchChildTag("example");
			ArrayList<XMLBlock> trainSet = new ArrayList<XMLBlock>();
			ArrayList<XMLBlock> testSet = new ArrayList<XMLBlock>();

			// iterate over all the examples and separate the test and training sets
			if (exampleSet != null) {
				for (int i = 0; i < exampleSet.size(); i++) {
					XMLBlock oneExample = exampleSet.get(i);
					ArrayList<Pair> propVect = oneExample.getProperties();
					if ((propVect == null) || propVect.get(0).value.equalsIgnoreCase("training")) {
						trainSet.add(oneExample);
					} else if (propVect.get(0).value.equalsIgnoreCase("test")) {
						testSet.add(oneExample);
					} else {
						trainSet.add(oneExample);
					}
				}
			}

			// input the training examples 
			for (int i = 0; i < trainSet.size(); i++) {
				Integer[] trainIndices = new Integer[parameters.length];
				String[] trainItems = new String[parameters.length];
				XMLBlock oneExample = trainSet.get(i);
				// get the values for each of the parameters
				ArrayList<XMLBlock> values = oneExample.searchChildTag("value");
				// check to see if the number of parameters are equal to the number
				// of values. If not, return an error
				if (values.size() != parameters.length) {
					errString = "Training example " + i + "does not have the correct number of parameters";
					throw (new Exception());
				}
				for (int j = 0; j < values.size(); j++) {
					// get the parameter name for this example
					ArrayList<Pair> paraList = values.get(j).getProperties();
					// there is no associated name for this example, return an error
					if (paraList == null) {
						errString = "Training example " + i + " has no associated parameter";
						throw (new Exception());
					}
					// assume only one property
					Pair paraPair = paraList.get(0);
					trainItems[j] = new String(values.get(j).getText());
					trainIndices[j] = new Integer(indexOf(paraPair.value));
				}
				addTrainingExample(trainIndices, trainItems);
			}
			// input the test examples 
			for (int i = 0; i < testSet.size(); i++) {
				Integer[] testIndices = new Integer[parameters.length];
				String[] testItems = new String[parameters.length];
				XMLBlock oneExample = testSet.get(i);
				// get the values for each of the parameters
				ArrayList<XMLBlock> values = oneExample.searchChildTag("value");
				// check to see if the number of parameters are equal to the number
				// of values. If not, return an error
				if (values.size() != parameters.length) {
					errString = "Test example " + i + "does not have the correct number of parameters";
					throw (new Exception());
				}
				for (int j = 0; j < values.size(); j++) {
					// get the parameter name for this example
					ArrayList<Pair> paraList = values.get(j).getProperties();
					// there is no associated name for this example, return an error
					if (paraList == null) {
						errString = "Test example " + i + " has no associated parameter";
						throw (new Exception());
					}
					// assume only one property
					Pair paraPair = paraList.get(0);
					testItems[j] = new String(values.get(j).getText());
					testIndices[j] = new Integer(indexOf(paraPair.value));
				}
				addTestExample(testIndices, testItems);
			}
		} catch (Exception e) {
			if (errString.length() > 0) {
				return "Error: " + errString;
			}
			e.printStackTrace();
			return "Error: " + e.toString();
		}
		return "OK";
	}

	/**
	 * Function that returns the index of a parameter. Returns -1 if the parameter is not found
	 */
	public int indexOf(String name) {
		for (int i = 0; i < parameters.length; i++) {
			if (parameters[i].equalsIgnoreCase(name)) {
				return i;
			}
		}
		return -1;
	}

	/**
	 * Takes a string and checks to see if it is an XML file by checking if
	 * the first non-whitespace character is a "<" symbol.
	 * If it is, then this will build an XML tree and pass it to parseXML().
	 * If it is a normal text file, this will pass the string to parse().
	 * Any errors will just be passed up to the calling function
	 */
	public String parseString(String all) {
		String s = all.trim();
		String errString = "";
		//reset();	We don't want to do a full reset here because there may be parameter names we need
		
		trainingList.removeAll();
		testList.removeAll();
		trainingModel.clear();
		testModel.clear();

		// an XML file
		if (s.startsWith("<")) {
			XMLTree tree = new XMLTree();
			try {
				tree.readString(all);
			} catch (XMLParseException e) {
				return e.getLocalizedMessage();
			}
			errString = parseXML(tree);
		}
		// non-XML
		else {
			errString = parse(all);
		}
		return errString;
	}

	/**
	 * Generates the text representation of the examples in this ExampleList.
	 * The representation is as described in the parse method.
	 * 
	 * @see parse
	 * 
	 */

	public String generateTextRep() {
		// identifier lines are not needed because
		// it is already taken care of in NeuralGraph
		/*Calendar now = Calendar.getInstance();
		outString += "% Auto-generated on " + now.getTime().toString() + "\n\n";
		
		outString += "%  'T:' Identifies the Attribute Title Line  " + "\n";
		outString += "%  'A:' Identifies Training Set Examples  " + "\n";
		outString += "%  'B:' Identifies Test Set Examples  " + "\n\n";*/

		if (indices == null) {
			return "\nNO DATASET LOADED, NO GRAPH GENERATED";
		}
		StringBuffer outBuffer = new StringBuffer();
		ArrayList<Hashtable<Integer, String>> examples;

		boolean isTraining = true;
		// Used to indicate whether we're on the first or second iteration.
		//This for just loops twice, once for training examples, and again for test examples
		for (int k = 0; k < 2; k++) {
			if (isTraining) {
				//Output the title string
				examples = trainingData;
				outBuffer.append("T: ");

				for (int i = 0; i < indices.length; i++) {
					if (i > 0) {
						outBuffer.append(", ");
					}
					outBuffer.append(parameters[i]);
				}

				outBuffer.append(";\n");
			} else {
				examples = testData;
			}

			if (examples.size() > 0) {
				Integer indices[] = this.indices;

				for (int i = 0; i < examples.size(); i++) {
					if (isTraining) {
						outBuffer.append("A: ");
					} else {
						outBuffer.append("B: ");
					}

					Hashtable<Integer, String> oneExample = examples.get(i);
					for (int j = 0; j < indices.length; j++) {
						String v = oneExample.get(indices[j]);
						if (v == null)
							v = "Error";
						if (j > 0)
							outBuffer.append(", ");
						outBuffer.append(v);
					}
					outBuffer.append(";\n");
				}
			} else {
				if (isTraining) {
					outBuffer.append("\n% NO TRAINING EXAMPLES\n");
				} else {
					outBuffer.append("\n% NO TEST EXAMPLES\n");
				}
			}

			isTraining = false;
			//Set to false for the next/second/last iteration
		}

		return outBuffer.toString();
	}

	public String generateXMLTextRep() {
		if (indices == null) {
			return "<!-- No Examples Loaded -->\n\n";
		}
		StringBuffer outBuffer = new StringBuffer("<!-- Example Database -->\n");
		outBuffer.append("<EXAMPLES>\n\n");
		outBuffer.append("\t<!-- Parameter Definition -->\n");
		for (int i = 0; i < parameters.length; i++) {
			outBuffer.append("\t<PARAMETER type=");
			if (isInput[i]) {
				outBuffer.append("input>");
			} else {
				outBuffer.append("output>");
			}
			outBuffer.append(parameters[i]).append("</PARAMETER>\n");
		}
		outBuffer.append("\n\t<!-- Examples -->\n");

		boolean isTraining = true;
		ArrayList<Hashtable<Integer, String>> examples;
		// Used to indicate whether we're on the first or second iteration.
		//This for just loops twice, once for training examples, and again for test examples
		for (int k = 0; k < 2; k++) {
			if (isTraining) {
				//Output the title string
				examples = trainingData;
			} else {
				examples = testData;
			}
			if (examples.size() > 0) {
				Integer indices[] = this.indices;
				for (int i = 0; i < examples.size(); i++) {
					if (isTraining) {
						outBuffer.append("\t<EXAMPLE type=training>\n");
					} else {
						outBuffer.append("\t<EXAMPLE type=test>\n");
					}

					Hashtable<Integer, String> oneExample = examples.get(i);
					for (int j = 0; j < indices.length; j++) {
						String v = oneExample.get(indices[j]);
						if (v == null)
							v = "Error";

						outBuffer.append("\t\t<VALUE parameter=\"").append(parameters[j]).append("\">");
						outBuffer.append(v).append("</VALUE>\n");
					}
					outBuffer.append("\t</EXAMPLE>\n");
				}
			}
			isTraining = false;
			//Set to false for the next/second/last iteration
		}
		return outBuffer.append("</EXAMPLES>\n\n").toString();
	}

	public String inValue(int i, int j) {
		//assume correct use!
		return trainingData.get(i).get(indices[j]);
	}

	public String outValue(int i) {
		//assume correct use!
		return trainingData.get(i).get(indices[indices.length - 1]);
	}

	/**
	* function to add an element from the test set
	* 
	*/
	public void addTestExample(Integer ind[], String values[]) {
		indices = ind.clone();
		Hashtable<Integer, String> oneExample = new Hashtable<Integer, String>();
		for (int i = 0; i < indices.length; i++) {
			oneExample.put(indices[i], values[i]);
		}
		testData.add(oneExample);
	}

	/**
	* function to remove an element from the test set
	* 
	*/
	public void removeTestExamples(int[] inds) {
		for (int i = inds.length - 1; i >= 0; i--)
			testData.remove(inds[i]);
		// disabled for now
		testList = getTestList();
	}

	/** 
	 * Sets the parameters for the Example Set.  This method converts 
	 * the para parameter array to a offsetPoint and calls the other
	 * <code>setParameters</code> method.
	 */
	public void setParameters(String[] para) {
		ArrayList<String> parameters = new ArrayList<String>(para.length);
		for (int i = 0; i < para.length; i++) {
			parameters.add(para[i]);
		}

		setParameters(parameters);
	}

	public int getTestSize() {
		return testModel.getSize();
	}

	public int getTrainingSize() {
		return trainingModel.getSize();
	}

	/**
		* Sets the parameter array for the Example set.  The parameter array
		* is the list of attribute names for data items.  This method also
		* clears all data in the ExampleList.
		* 
		* @param newParams		A ArrayList of new parameter names (Strings)
		*/
	public void setParameters(ArrayList<String> newParams) {
		reset();
		parameters = new String[newParams.size()];
		indices = new Integer[newParams.size()];
		isInput = new boolean[newParams.size()];

		for (int i = 0; i < newParams.size(); i++) {
			parameters[i] = newParams.get(i);
			indices[i] = new Integer(i);
			isInput[i] = true;
		}

		isInput[newParams.size() - 1] = false;
		//Set last param as output by default	
	}
	
	public void clearParams() {
		parameters = null;
	}

	/**
	 * Sets a parameter as either an input or an output.
	 * (This is provided for the Neural Networks applet)
	 * PRE: The parameters of this example list have been initialized.
	 * 
	 * @param input		true if you wish to set the parameter as in input, false if the parameter is an output
	 * @param index		The index of the parameter to be set as an input or output.
	 * 
	 */
	public void setInputParam(boolean boolInput, int index) {
		//Check if this operation is possible
		if (isInput == null || index >= isInput.length || index < 0) {
			return;
		}

		isInput[index] = boolInput;
	}

	/**
	 * Returns a boolean that indicates whether a parameter is an input or output.
	 * (This is provided for the Neural Networks applet)
	 * PRE: The parameters of this example list have been initialized.
	 * 
	 * @param index		The index of the parameter whose setting is to be returned
	 */
	public boolean isInputParam(int index) {
		//Check if this operation is possible
		if (isInput == null || index >= isInput.length || index < 0) {
			return false; //This should not occur, check pre-condition
		}

		return isInput[index];
	}

	/**
	 *  function that returns the parameters
	 */
	public String[] getParameters() {
		return parameters;
	}

	/**
	* Adds a single training example to the example list.
	* Builds a hash table mapping indices to values and
	* adds the hash table to the trainingData offsetPoint.
	* 
	*/
	public void addTrainingExample(Integer ind[], String values[]) {
		indices = ind.clone();
		Hashtable<Integer, String> oneExample = new Hashtable<Integer, String>();
		for (int i = 0; i < indices.length; i++) {
			oneExample.put(indices[i], values[i]);
		}
		trainingData.add(oneExample);
	}
	
	/**
	 * updates the test example at index to have indicies ind and values
	 * @param index - the index of the example to update
	 * @param ind - the indices to put in the example
	 * @param values - the values to put in the example
	 */
	public void updateTestExample(int index, Integer ind[], String values[]) {
		testData.remove(index);
		indices = ind.clone();
		Hashtable<Integer, String> oneExample = new Hashtable<Integer, String>();
		for (int i = 0; i < indices.length; i++) {
			oneExample.put(indices[i], values[i]);
		}
		
		testData.add(index, oneExample);
		testList = getTestList();
	}
	
	/**
	 * updates the training example at index to have indicies ind and values
	 * @param index - the index of the example to update
	 * @param ind - the indices to put in the example
	 * @param values - the values to put in the example
	 */
	public void updateTrainingExample(int index, Integer ind[], String values[]) {
		trainingData.remove(index);
		indices = ind.clone();
		Hashtable<Integer, String> oneExample = new Hashtable<Integer, String>();
		for (int i = 0; i < indices.length; i++) {
			oneExample.put(indices[i], values[i]);
		}
		
		trainingData.add(index, oneExample);
		trainingList = getTrainingList();
	}
	
	public void updateTrainingExample(int index, Hashtable<Integer, String> oneExample) {
		trainingData.remove(index);
		trainingData.add(index, oneExample);
		trainingList = getTrainingList();
	}
	
	public void updateTestExample(int index, Hashtable<Integer, String> oneExample) {
		testData.remove(index);
		testData.add(index, oneExample);
		testList = getTestList();
	}

	public void removeTrainingExamples(int[] inds) {
		for (int i = inds.length - 1; i >= 0; i--)
			trainingData.remove(inds[i]);
		// disabled for now.
		trainingList = getTrainingList();
	}

	public void moveToTestForLoad() {
		trainingList = getTrainingList();
		testList = getTestList();
	}

	/**
	* Moves training examples with indices in the array inds
	* to the test set.  
	* 
	*/
	public void moveToTest(int[] inds) {
		sortIntArray(inds);

		//Note, two for loops are used intentionally, the first iterates forward
		//so that data examples appear in the same order when moved.  The second
		//iterates backwards on a sorted array of indexes so that the indexes don't
		//change while deleting	

		for (int i = 0; i < inds.length; i++) {
			testData.add(trainingData.get(inds[i]));
		}

		for (int i = inds.length - 1; i >= 0; i--) {
			trainingData.remove(inds[i]);
		}

		trainingList = getTrainingList();
		testList = getTestList();
	}

	/**
	* Moves test examples with indices in the array inds
	* to the training set.  
	* 
	*/
	public void moveToTraining(int[] inds) {
		sortIntArray(inds);

		//Note, two for loops are used intentionally, the first iterates forward
		//so that data examples appear in the same order when moved.  The second
		//iterates backwards on a sorted array of indexes so that the indexes don't
		//change while deleting	

		for (int i = 0; i < inds.length; i++) {
			trainingData.add(testData.get(inds[i]));
		}

		for (int i = inds.length - 1; i >= 0; i--) {
			testData.remove(inds[i]);
		}

		trainingList = getTrainingList();
		testList = getTestList();
	}


	public JList getTestList() {
		testModel.clear();

		if (indices == null) {
			testModel.addElement("NO DATA");
			return testList;
		}
		StringBuffer heading = new StringBuffer("");
		if (parameters != null) {
			for (int i = 0; i < indices.length; i++) {
				heading.append(parameters[i]);
				for (int j = parameters[i].length(); j < 8; j++)
					heading.append(" ");
			}
		}
		// testList.add(heading);     
		testModel.addElement(heading.toString());

		for (int i = 0; i < testData.size(); i++) {
			StringBuffer display = new StringBuffer("");
			Hashtable<Integer, String> oneExample = testData.get(i);
			for (int j = 0; j < indices.length; j++) {
				String value = oneExample.get(indices[j]);
				if (value != null) {
					display.append(value);
					for (int k = value.length(); k < 8; k++)
						display.append(" ");
				}
			}
	
			testModel.addElement(display.toString());
		}
		return testList;
	}

	/** returns the testData offsetPoint */
	public ArrayList<Hashtable<Integer, String>> getTestArrayList() {
		return testData;
	}

	/** returns the training offsetPoint
	 * 
	 */
	public ArrayList<Hashtable<Integer, String>> getTrainingArrayList() {
		return trainingData;
	}

	/** returns the indices
	 */
	public Integer[] getIndices() {
		return indices;
	}
	
	/**
	 * returns the hashtable of values for the test example at index
	 * @param index - the index of the example to get
	 * @return - the hashtable of values
	 */
	public Hashtable getTestExampleValues(int index) {
		return testData.get(index);
	}
	
	/**
	 * returns the hashtable of values for the training example at index
	 * @param index - the index of the example to get
	 * @return - the hashtable of values
	 */
	public Hashtable getTrainingExampleValues(int index) {
		return trainingData.get(index);
	}

	public JList getTrainingList() {
		//  trainingList.removeAll();
		trainingModel.clear();

		if (indices == null) {
			trainingModel.addElement("NO DATA");
			return trainingList;
		}
		StringBuffer heading = new StringBuffer("");
		if (parameters != null) {
			for (int i = 0; i < indices.length; i++) {
				heading.append(parameters[i]);
				for (int j = parameters[i].length(); j < 8; j++)
					heading.append(" ");
			}
		}
		trainingModel.addElement(heading.toString());
		//trainingList.add(heading);     

		for (int i = 0; i < trainingData.size(); i++) {
			StringBuffer display = new StringBuffer("");
			Hashtable<Integer, String> oneExample = trainingData.get(i);
			for (int j = 0; j < indices.length; j++) {
				String value = oneExample.get(indices[j]);
				if (value != null) {
					display.append(value);
					for (int k = value.length(); k < 8; k++)
						display.append(" ");
				}
			}
			trainingModel.addElement(display.toString());
		}
		return trainingList;
	}

	public int getNumParameters() {
		if (parameters != null) {
			return parameters.length;
		}
		return 0;
	}

	/**
	 * Returns the set of values that appear in the training set for a particular parameter.
	 * 
	 * Note: This method's performance is not an issue, so there is no reason to build a more
	 * elaborate version to improve performance.
	 * 
	 * @param		paramIndex			The index of the parameter whose values are to be returned
	 */
	public ArrayList getValueSet(int paramIndex) {
		ArrayList<String> outputList = new ArrayList<String>(5); //ArrayList of values to return
		String value; //Holds values to be inserted in the output offsetPoint
		boolean valueExists = false;
		//Used as a flag when searching the output offsetPoint

		//Iterate over all training examples
		for (int i = 0; i < trainingData.size(); i++) {
			Hashtable<Integer, String> oneExample = trainingData.get(i);
			value = oneExample.get(new Integer(paramIndex));

			//Add the value to the output offsetPoint if it is not already there
			valueExists = false;
			//Iterate over the output offsetPoint to check for duplicates
			for (int j = 0; j < outputList.size(); j++) {
				if (outputList.get(j).equals(value)) {
					valueExists = true;
					break;
				}
			}

			if (!valueExists) {
				outputList.add(value);
			}
		}

		sortStringList(outputList);
		return outputList;
	}

	/**
	* insertion sorts an integer Array
	*/
	public void sortIntArray(int[] inputArray) {
		for (int i = 0; i < inputArray.length; i++) {
			for (int j = i; j < inputArray.length; j++) {
				if (inputArray[j] < inputArray[i]) {
					//Swap
					int temp = inputArray[i];
					inputArray[i] = inputArray[j];
					inputArray[j] = temp;
				}
			}
		}
	}

	/**
		* insertion sorts a string offsetPoint
		* PRE: inputList is an ArrayList of strings
		*/
	public void sortStringList(ArrayList<String> inputList) {
		for (int i = 0; i < inputList.size(); i++) {
			for (int j = i; j < inputList.size(); j++) {
				if (inputList.get(j).compareTo(inputList.get(i)) < 0) {
					//Swap
					String temp = inputList.get(i);
					inputList.set(i, inputList.get(j));
					inputList.set(j, temp);
				}
			}
		}
	}

	/**
		* makes a copy of this example list.
		*/
	public Object clone() {
		ExampleList copy;
		copy = new ExampleList();
		copy.trainingData = new ArrayList<Hashtable<Integer, String>>(trainingData);
		copy.indices = new Integer[indices.length];
		copy.parameters = new String[indices.length];
		for (int i = 0; i < indices.length; i++) {
			copy.indices[i] = new Integer(indices[i].intValue());
			copy.parameters[i] = new String(parameters[i]);
		}
		return copy;
	}
}