/**
 * NeuralGraph.java
 * July 25, 2002
 */

package AIspace.neural;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Hashtable;
import java.util.Iterator;

import AIspace.XMLReader.XMLBlock;
import AIspace.XMLReader.XMLTree;
import AIspace.graphToolKit.Graph;
import AIspace.graphToolKit.GraphConsts;
import AIspace.graphToolKit.elements.Edge;
import AIspace.graphToolKit.elements.Node;
import AIspace.graphToolKit.elements.Point;
import AIspace.neural.elements.NeuralEdge;
import AIspace.neural.elements.NeuralNode;
import AIspace.neural.intList.IntList;

/**
 * NeuralGraph subclasses CIGraph, implementing functionality
 * specific to the neural network applet.
 *
 * @author  Leslie Tung, Shinjiro Sueda, & Joseph Roy Santos
 * @version 1.1
 * @since JDK1.1
 */
public class NeuralGraph extends Graph {
	private ArrayList<Node> inputNodes;
	private ArrayList<Node> outputNodes;
	private double learningRate;
	/**
	 * double which sets the bounds for random parameter initialization.
	 */
	private double paramBound;
	/**
	 * double which sets the value for parameter initialization.
	 */
	private double paramValue;
	/** 
	 * boolean which is set to true if parameters should be initialized to random values 
	 * (paramBound), or false if parameters should be initialized to a set value (paramValue).
	 */
	private boolean paramRandom;
	private int numberOfIterations;
	private ExampleList exampleList;
	public BackPropagation backPropagation;
	private int startingCount;
	public PlotFrame pFrame;
	private AutoStep auto;
	private double stoppingErrVal;
	public  boolean showParam;
	private double momentum;
	public boolean standardized;
	public  int dt = 100;
	private int errorType;
	private String shortDescription;
	private String detailedDescription;
	
	public static int CORRECT = 110;
	public static int INCORRECT = 111;
	public static int NO_PREDICTION = 112;

	public static int AVERAGE_SUMSQUARES = -1;
	public static int SUMSQUARES = -2;	

	//
	// Methods
	//
	
	/**
	 * Initializes a new NeuralGraph
	 */
	public NeuralGraph( NeuralCanvas canvas ) {
		super( canvas );
		setLineWidth(2);
		learningRate = 0.2;
		numberOfIterations = 50;
		stoppingErrVal = 0.1;
		startingCount = 0;
		inputNodes = new ArrayList<Node>();
		outputNodes = new ArrayList<Node>();
		exampleList = new ExampleList();
		showParam = true;
		paramBound = 1.0;
		paramValue = 0.0;
		paramRandom = true;
		momentum = 0;
		standardized = false;
		errorType = SUMSQUARES;
		shortDescription = new String();
		detailedDescription = new String();
	}
	
	public void setErrorType( int errConst) {
		errorType = errConst;
	}
	
	public int getErrorType() {
		return errorType;
	}

	public void addNode( Node node ) {
		super.addNode( node );
		resetBackPropagation();
	}

	public void addEdge( Edge edge ) {
		super.addEdge( edge );
		
		if(paramsChanged())
			rebuildData();
		resetBackPropagation();
	}
	
	public void addEdge( Edge edge, boolean rebuildData ) {
		super.addEdge( edge );
		if(rebuildData && paramsChanged())
			rebuildData();
		resetBackPropagation();
	}
	
	/**
	 * Checks if there is a cycle in the network.
	 * @return true if there is a cycle in the network, false otherwise.
	 * 
	 * @author Byron Knoll
	 */
	public boolean cycleCheck ()
	{
		boolean returnValue = false;
		for (Node inputNode: inputNodes)
		{
			// Perform DFS from the input nodes to detect cycles in the network.
			returnValue = returnValue || cycleDFS((NeuralNode)inputNode, null);
		}
		return returnValue;
	}
	
	/**
	 * Override the "moveNodeToTop" method in the superclass. This is to prevent
	 * the order of nodes from being changed. In other AIspace tools it is OK for
	 * the node order to change, but the training/test examples here are dependent
	 * on node order.
	 * @author Byron Knoll
	 */
	@Override
	public void moveNodeToTop (Node node)
	{
		
	}
	
	/**
	 * Performs depth-first search on a neural node to check for cycles
	 * 
	 * @param node - the node to check
	 * @param path - a collection of nodes on the current path
	 * @return true if a cycle is detected, false otherwise
	 * 
	 * @author Byron Knoll
	 */
	private boolean cycleDFS (NeuralNode node, ArrayList <NeuralNode> path)
	{
		if (path == null) path = new ArrayList <NeuralNode> ();
		if (path.contains(node)) return true;
		if (node.getNumChildrenNodes()==0) return false;
		path.add(node);
		boolean returnValue = false;
		for (Node child : node.getChildrenNodes())
		{
			returnValue = returnValue || cycleDFS ((NeuralNode)child, path);
		}
		path.remove(node);
		return returnValue;
	}
	
	/**
	 * returns true if the parameters have been changed
	 * parameters being input/output nodes
	 */
	public boolean paramsChanged() {
		ArrayList<String> inputParameters = new ArrayList<String>();
		Iterator<Node> nodeItr = getNodes();
		while (nodeItr.hasNext()) {
			NeuralNode tmpNeuralNode = (NeuralNode)nodeItr.next();
			
			//if node has no children or no parents, it is a parameter
			if((tmpNeuralNode.getParentNodes().size() + tmpNeuralNode.getChildrenNodes().size()) == 0) {
				//if it has no parents and children, don't include it
			} else if((tmpNeuralNode.getParentNodes().size() == 0) || (tmpNeuralNode.getChildrenNodes().size() == 0)) {
				inputParameters.add(tmpNeuralNode.getLabel());
			}
		}
		
			//test if any parameters have been added/removed
		boolean paramsChanged = false;
		String[] params = exampleList.getParameters();
		if((params != null) && (params.length == inputParameters.size())) {
			for(int i = 0; i < params.length; i++) {
				if(!params[i].equals(inputParameters.get(i))) {
					paramsChanged = true;
					break;
				}
			}
		} else
			paramsChanged = true;
		
		return paramsChanged;
	}
	
	/**
	 * rebuilds the data in the trainingData and testData vectors
	 * newly added parameters get 0 default values, missing parameters
	 * get their values deleted
	 * @param newEdge - the edge that connects the new nodes - used to determine
	 * if the node being added is an input or output node. pass in null if no edge
	 * available.
	 */
	public void rebuildData() {
		//algorithm:
		// 1) find the list of new parameters in the graph (string array)
		// 2) get the list of old parameters in the graph (from exampleList)
		// 3) update the exampleList parameters with the new list
		// 4) if less new parameters than old parameters, one or more parameters deleted
		//		a) call remove example update
		// 5) if more new parameters than old parameters, one parameter added
		//		a) call add example update

		int numInputs = 0;
		//get new parameter list
		ArrayList<String> nodeParams = new ArrayList<String>();
		ArrayList<String> outParams = new ArrayList<String>();
		Iterator<Node> nodeItr = getNodes();
		while (nodeItr.hasNext()) {
			NeuralNode tmpNeuralNode = (NeuralNode)nodeItr.next();
			
			//if node has no children or no parents, it is a parameter
			if((tmpNeuralNode.getParentNodes().size() + tmpNeuralNode.getChildrenNodes().size()) == 0) {
				//if it has no parents and no children, ignore it
			} else if(tmpNeuralNode.getParentNodes().size() == 0) {
				nodeParams.add(tmpNeuralNode.getLabel());
				numInputs++;
			} else if(tmpNeuralNode.getChildrenNodes().size() == 0) {
				outParams.add(tmpNeuralNode.getLabel());
			}
		}
		
		//done this way so output parameters always appear at the end
		nodeParams.addAll(outParams);
		
		//determine which ones have been added or removed
		String[] params = exampleList.getParameters();
		if(params != null) {
			if(params.length > nodeParams.size()) {
				removeExamplesUpdate(params, nodeParams);
			} else if (params.length < nodeParams.size()) {
				addExamplesUpdate(params, nodeParams, numInputs);
			} else {
				hiddenExamplesUpdate(params, nodeParams);
			}
		}
		
		if(nodeParams.size() == 0) {
			exampleList.resetAll();
		}
		
		//update the parameters if needed
		if(nodeParams.size() > 0) {
			exampleList.setParameters(nodeParams.toArray(new String[0]));
		} else {
			exampleList.clearParams();
		}
	}
	
	/** 
	 * transfer examples down the line
	 * 
	 * @param params - the list of old parameters
	 * @param nodeParams - the list of new parameters
	 */
	private void hiddenExamplesUpdate(String[] params, ArrayList<String> nodeParams) {
		//two types of hidden example updates that don't change parameter size
		//either node is at beginning of edge to a hidden node, or node is
		//at end of edge from a hidden node
		//in either case, transfer the parameters from the old node to the new one
		
		//find index of removed item in old parameters list
		int oldIndex = -1;
		int newIndex = -1;
		for(int j = 0; j < params.length; j++) {
			//if index in old params list not in new one, it was removed,
			//so store the index
			if(!nodeParams.contains(params[j])) {
				oldIndex = j;
				break;
			}
		}
		
		//find index of added item in new parameters list
		boolean found;
		for(int i = 0; i < nodeParams.size(); i++) {
			//parameter hasn't been found in list of old parameters yet
			found = false;
			//iterate through old parameters looking for current parameter
			for(int j = 0; j < params.length; j++) {
				//if found, then stop iterating
				if(params[j].equals(nodeParams.get(i))) {
					found = true;
					break;
				}
			}
			//node wasn't found, so it was newly added
			if(found == false) {
				newIndex = i;
				break;
			}
		}
		
		if(newIndex == -1 || oldIndex == -1)
			return;
		
		params[oldIndex] = nodeParams.get(newIndex);
		for(int i = 0; i < nodeParams.size(); i++) {
			nodeParams.set(i, params[i]);
		}
	}
	
	/**
	 * handles updating the training and test examples when an item is added.
	 * preserves order of items when deleting, input variables kept at the
	 * start of the list, output variables kept at the end. a default value of
	 * 0.0 is used for new values
	 * 
	 * @param params - the list of old parameters
	 * @param nodeParams - the list of new parameters
	 * @param numInputs - the number of input parameters
	 */
	private void addExamplesUpdate(String[] params, ArrayList<String> nodeParams, int numInputs) {
		//determine if one added is an input or output node
		int[][] isOutNode = isOutputNode(params, nodeParams);
		//hashtable to hold the current example
		Hashtable<Integer, String> oneExample;
		
		//ensure that the output values remain at the end
		int boolIndex = 0;
		boolean isNodeOutput;
		int nodeIndex;
		int nextNodeIndex;
		//add each of the new parameters (can have max of one new input, one new output)
		while(boolIndex < isOutNode.length) {
			//see if node is an output node
			isNodeOutput = (isOutNode[boolIndex][1] != 0);
		
			//get index of node in nodes offsetPoint
			nodeIndex = isOutNode[boolIndex][0];
			nextNodeIndex = -1;
			
			//find parameter index of node with the next index greater than nodeIndex
			for(int i = nodeIndex+1; i < nodes.size(); i++) {
				if(isNodeOutput && (((NeuralNode)this.nodeAt(i)).getNumChildrenNodes() == 0) && (((NeuralNode)this.nodeAt(i)).getNumParentNodes() > 0)) {
					//if output node, only consider output nodes
					nextNodeIndex = i;
					break;
				} else if(!isNodeOutput && (((NeuralNode)this.nodeAt(i)).getNumParentNodes() == 0) && (((NeuralNode)this.nodeAt(i)).getNumChildrenNodes() > 0)) {
					//if input node, only consider input nodes
					nextNodeIndex = i;
					break;
				}
			}
			
			//no node with lower index, so insert examples at end of inputs or end of outputs
			if(nextNodeIndex == -1) {
				if(isNodeOutput)
					nextNodeIndex = nodeParams.size()-1;
				else
					nextNodeIndex = numInputs-1;
			} else {
				//currently nextNodeIndex holds index of node, need old parameter index of node
				String nodeLabel = ((NeuralNode)this.nodeAt(nextNodeIndex)).getLabel();
				for(int i = 0; i < params.length; i++) {
					if(nodeLabel.equals(params[i])) {
						nextNodeIndex = i;
						break;
					}				
				}
			}
			
			//go through all the training data and update each example to include new parameter
			for(int i = 0; i < exampleList.trainingData.size(); i++) {
				oneExample = exampleList.trainingData.get(i);
				exampleList.updateTrainingExample(i, updateInsertExample(oneExample, nextNodeIndex));
			}
			
			//go through all the test data and update each example to include new parameter
			for(int i = 0; i < exampleList.testData.size(); i++) {
				oneExample = exampleList.testData.get(i);
				exampleList.updateTestExample(i, updateInsertExample(oneExample, nextNodeIndex));
			}
			boolIndex++;
		}
	}
	
	/**
	 * handles updating the training and test examples when an item is removed.
	 * preserves order of items when deleting
	 * 
	 * @param params - the list of old parameters
	 * @param nodeParams - the list of new parameters
	 */
	private void removeExamplesUpdate(String[] params, ArrayList nodeParams) {
		/*
		 * algorithm:
		 * 1) get the indexes of the parameters to remove
		 * 2) starting with the last index
		 * 	a) remove the parameter info, update all indexes in _example_ beyond last index
		 *  b) repeat for next highest index to remove
		 */
		
		//index of the parameters to remove
		int[] indices = new int[params.length - nodeParams.size()];
		//current index into the indices
		int indicesIndex = -1;
		//one has been removed, find the index in the params list 
		//of the removed one
		for(int j = 0; j < params.length; j++) {
			//if index in old params list not in new one, it was removed,
			//so store the index
			if(!nodeParams.contains(params[j])) {
				indices[++indicesIndex] = j;
			}
		}
		
		//remove the data from the training and test vectors
		Hashtable<Integer, String> oneExample;
		//work backwards so data order is preserved
		//(remove last one, update past that, then 2nd last one, update past that, etc)
		while(indicesIndex >= 0) {
			//go through all the training data and remove the index, while updating the others
			for(int i = 0; i < exampleList.trainingData.size(); i++) {
				oneExample = exampleList.trainingData.get(i);
				exampleList.updateTrainingExample(i, updateRemoveExample(oneExample, indices[indicesIndex]));
			}
			
			//go through all the test data and remove the index, while updating the others
			for(int i = 0; i < exampleList.testData.size(); i++) {
				oneExample = exampleList.testData.get(i);
				exampleList.updateTestExample(i, updateRemoveExample(oneExample, indices[indicesIndex]));
			}
			indicesIndex--;
		}
	}
	
	/** return true if the last parameter added was an output node 
	 * 
	 * @param params - the list of old parameters
	 * @param nodeParams - the list of new parameters
	 * @return 2d index, first is index into nodes offsetPoint. second is 0 if false, 1 if true
	 */
	private int[][] isOutputNode(String[] params, ArrayList<String> nodeParams) {
		boolean found;
		int[][] isOutputNode = new int[nodeParams.size() - params.length][2];
		
		int boolIndex = 0;
		//iterate through the new parameters
		for(int i = 0; i < nodeParams.size(); i++) {
			//parameter hasn't been found in list of old parameters yet
			found = false;
			//iterate through old parameters looking for current parameter
			for(int j = 0; j < params.length; j++) {
				//if found, then stop iterating
				if(params[j].equals(nodeParams.get(i))) {
					found = true;
					break;
				}
			}
			//node wasn't found, so it was newly added
			if(found == false) {
				//look through list of nodes for newly added one
				//needed as parameters only store node name
				for(int k = 0; k < nodes.size(); k++) {
					NeuralNode node = (NeuralNode) nodes.get(k);
					//return true if node found and has no children (output node)
					if(node.getLabel().equals(nodeParams.get(i))) {
						isOutputNode[boolIndex][0] = k;
						if(node.getChildrenNodes().size() == 0)
							isOutputNode[boolIndex][1] = 1;
					}
				}
				boolIndex++;
			}
		}
		
		//default
		return isOutputNode;
	}
	
	/** update the hashtable correctly to insert the new node example information.
	 * default value to insert is 0.0
	 * 
	 * @param oneExample - the hashtable of the example
	 * @param isOutputNode - boolean true if new node is an output node
	 * @param nextNodeIndex - the index of the parameter to place the info before (ie the prev column)
	 * @return the updated hashtable of the example
	 */
	private Hashtable<Integer, String> updateInsertExample(Hashtable<Integer, String> oneExample, int nextNodeIndex) {
		//update the indexes of the rest of the items
		for(int j = oneExample.size()-1; j >= nextNodeIndex; j--) {
			oneExample.put(new Integer(j+1), oneExample.get(new Integer(j)));
			oneExample.remove(new Integer(j));
		}
		oneExample.put(new Integer(nextNodeIndex), "0.0");
		
		return oneExample;
	}
	
	/** update the hashtable correctly to remove the node example information
	 * 
	 * @param oneExample - the hashtable of the example
	 * @param index - the index in the example of the node information
	 * @return the updated hashtable of the example
	 */
	private Hashtable<Integer, String> updateRemoveExample(Hashtable<Integer, String> oneExample, int index) {
		oneExample.remove(new Integer(index));
		
		//update the indexes of the rest of the items
		for(int j = index; j < oneExample.size(); j++) {
			oneExample.put(new Integer(j), oneExample.get(new Integer(j+1)));
			oneExample.remove(new Integer(j+1));
		}
		
		return oneExample;
	}
	
	public void setLearningRate(double rate) {
		learningRate = rate;
		if (backPropagation != null) {
			backPropagation.setLearningRate(rate);
		}
	}
	
	public double getLearningRate() {
		return learningRate;
	}

	public void setNumberOfIterations(int iter) {
		numberOfIterations = iter;
	}
	
	public int getNumberOfIterations() {
		return numberOfIterations;
	}

	public void setStoppingErrVal( double val ) {
		stoppingErrVal = val;
	}

	public double getStoppingErrVal() {
		return stoppingErrVal;
	}

	public void setInitParamBound( double val ) {
		paramBound = val;
	}

	public double getInitParamBound() {
		return paramBound;
	}
	
	public void setInitParamValue ( double val ) {
		paramValue = val;
	}

	public double getInitParamValue() {
		return paramValue;
	}
	
	/**
	 * Sets the parameter initialization option.
	 * @param val a boolean which is set to true if parameters should be initialized to random 
	 * values, or false if parameters should be initialized to a set value.
	 */
	public void setInitParamType(boolean val) {
		paramRandom = val;
	}
	
	/**
	 * Gets the parameter initialization option.
	 * @returna boolean which is true if parameters should be initialized to random values, 
	 * or false if parameters should be initialized to a set value.
	 */
	public boolean getInitParamType () {
		return paramRandom;
	}

	public ExampleList getExampleList() {
		return exampleList;
	}

	/** add an example to the training set
	 */
	public void addTrainingExample(Integer[] indices, Double[] values) {
		// convert the double array into a string array
		// to make it compatible with the Decision Tree Learning exampleList
		String [] valueStringArray = new String[values.length];
		for (int i = 0; i < values.length; i++) {
			valueStringArray[i] = values[i].toString();	
		}
		exampleList.addTrainingExample(indices, valueStringArray);
	}
	
	/** add an example to the test set
	 */
	public void addTestExample(Integer[] indices, Double[] values) {
		// convert the double array into a string array
		// to make it compatible with the Decision Tree Learning exampleList
		String [] valueStringArray = new String[values.length];
		for (int i = 0; i < values.length; i++) {
			valueStringArray[i] = values[i].toString();	
		}
		exampleList.addTestExample(indices, valueStringArray);
	}
	
	/**
	 * access the input nodes (with no parents) in the graph
	 */
	public ArrayList<Node> getInputNodes() {
		return inputNodes;
	}
	
	/**
	 * access the output nodes (with no children) in the graph
	 */
	public ArrayList<Node> getOutputNodes() {
		return outputNodes;
	}
	
	/**
	 * set the input nodes and output nodes in the graph
	 */
	public void setInputOutputNodes(boolean paint) {
		inputNodes = new ArrayList<Node>(5);
		outputNodes = new ArrayList<Node>(5);
		NeuralNode tmpNode;
		Iterator<Node> nodeItr = getNodes();
		while (nodeItr.hasNext())
		{
			tmpNode = (NeuralNode)nodeItr.next();
			tmpNode.shape = GraphConsts.OVAL;
			tmpNode.setLabel( showParam );
			if (paint) {
				tmpNode.setColoring(true);
			}
			if ((tmpNode.getNumParentNodes() == 0) && (tmpNode.getNumChildrenNodes() > 0)) {
				// input nodes
				inputNodes.add(tmpNode);
				if (paint) {
					tmpNode.shape = GraphConsts.DIAMOND;
					tmpNode.setColoring(false); // no colors for input nodes
				}
			} else if ((tmpNode.getNumChildrenNodes() == 0) && (tmpNode.getNumParentNodes() > 0)) {
				// output nodes
				tmpNode.shape = GraphConsts.RECT;
				outputNodes.add(tmpNode);
			} else if (tmpNode.getNumChildrenNodes() + tmpNode.getNumParentNodes() == 0) {
				//normal nodes not deleted but no edges out
				tmpNode.shape = GraphConsts.RECT;
				tmpNode.setLabel(false);
			}
		}
		for( int i = 0; i < numEdges(); i++ ) {
			edgeAt(i).updateLength();
		}
		setIOParameters();
	}
	
	/**
	 * Set the example list of the graph
	 */
	public void setExampleList(ExampleList e) {
		exampleList = e;
	}
	
	/**
	 * set the input nodes and output nodes in the graph (color and shape)
	 */
	public void resetInputOutputNodes() {
		NeuralNode tmpNode;
		Iterator<Node> nodeItr = getNodes();
		while (nodeItr.hasNext()) {
			tmpNode = (NeuralNode)nodeItr.next();
			tmpNode.updateSize();
			tmpNode.color = Color.black;
			tmpNode.setColoring(false);
		}
		for( int i = 0; i < numEdges(); i++ ) {
			edgeAt(i).updateLength();
		}
	}
	
	/**
	 * Initialize the parameter values based on the initialization options.
	 */
	public void initializeParaValues() {
		if (paramRandom)
		{
			randomParaValues();
		}
		else
		{
			clearParaValues();
		}
	}

	/**
	 * Resets the parameter values to a random value in [-b, b].
	 */
	public void randomParaValues() {
		stop();
		NeuralNode tmpNode;		
		Iterator<Node> nodeItr = getNodes();
		while (nodeItr.hasNext()) {
			tmpNode = (NeuralNode)nodeItr.next();
			if( tmpNode.getNumParentNodes() > 0 ) {
				tmpNode.setCurrentParaValue( 2*paramBound*Math.random() - paramBound );
			}
		}

		for(Edge e : edges) {
			((NeuralEdge) e).setCurrentParaValue( 2*paramBound*Math.random() - paramBound );
		}
		
		resetBackPropagation();
		repaint();
	}
	
	/**
	 * Resets the parameter values to a constant initial value.
	 */
	public void clearParaValues() {
		stop();
		NeuralNode tmpNode;
		
		// Iterate through all the nodes.
		Iterator<Node> nodeItr = getNodes();
		while (nodeItr.hasNext()) {
			tmpNode = (NeuralNode)nodeItr.next();
			// If the node has parents, set its parameter value to the constant value.
			if( tmpNode.getNumParentNodes() > 0 ) 
			{
				tmpNode.setCurrentParaValue(paramValue);
			}
		}

		// Iterate through all the edges and set the parameter values to the constant value.
		for(Edge e : edges) 
		{
			((NeuralEdge) e).setCurrentParaValue(paramValue);
		}
		
		resetBackPropagation();
		repaint();
	}

	/**
	 * Reset the parameter values to their initial state.
	 */
	public void resetParaValues() {
		stop();
		NeuralNode tmpNode;
		Iterator<Node> nodeItr = getNodes();
		while (nodeItr.hasNext()) {
			tmpNode = (NeuralNode)nodeItr.next();
			if (tmpNode.getNumParentNodes() > 0) {
				// only change parameter if node is not input node
				tmpNode.setCurrentParaValue(tmpNode.getInitialParaValue());
			}
		}

		for(Edge e : edges) {
			((NeuralEdge) e).setCurrentParaValue(((NeuralEdge) e).getInitialParaValue());
		}

		resetBackPropagation();
		repaint();
	}

	/**
	 * Toggles the parameter display.
	 */
	public void setShowParam( boolean show ) {
		showParam = show;
		Iterator<Node> nodeItr = nodes.iterator();
		while(nodeItr.hasNext()) {
			((NeuralNode) nodeItr.next()).setLabel(showParam);
		}
		NeuralEdge tmpEdge;
		for( int i = 0; i < numEdges(); i++ ) {
			tmpEdge = (NeuralEdge)edgeAt(i);
			tmpEdge.setLabel( showParam );
			tmpEdge.updateLength();
		}
		repaint();
	}

	/**
	 * Sets backpropagation to null.
	 */
	public void resetBackPropagation() {
		backPropagation = null;
		// reset all nodes' errors
		for (int i = 0; i < nodes.size(); i++) {
			((NeuralNode) nodes.get(i)).currentError = null;	
		}
	}

	/**
	 * Enables or disables the "Auto Step" and "Stop" buttons in the
	 * NeuralWindow and the PlotFrame classes.
	 */
	public void enableButtons( boolean bool ) {
		if( !canvas.inline ) {
			((NeuralWindow)canvas.parent).enableButtons( bool );
		}
		else
		{
			((InlineNeuralApplet)canvas.parent).enableButtons( bool );
		}
		if( pFrame != null ) {
			pFrame.auto.setEnabled( bool );
			pFrame.run.setEnabled( bool);
			pFrame.stop.setEnabled( !bool );
		}
	}

	/**
	 * Starts/continues back propagation.
	 */
	public void step() {
		if( backPropagation == null ) {
			initBackProp();
		}
		backPropagation.nnlearn(1, learningRate, momentum);
		setPromptLabel("Step " + backPropagation.getIterationCount() 
				+ ", " + "Training Error: " 
				+Math.round(backPropagation.getCurrErr()*10000)/(double)10000 
				+ "\nTest Error: "
				+ Math.round(backPropagation.getCurrTestErr()*10000)/(double)10000);
			
	}
	
	public void setTotalErrorAfterInitialization(){
		backPropagation = new BackPropagation(this, outputNodes, standardized );
		setPromptLabel("Initialized parameters.\n" 
						+ "Step " + backPropagation.getIterationCount()
						+ ": " + "Training Error: "
						+ Math.round(backPropagation.getCurrErr()*10000)/(double)10000 
						+ ", Test Error: "
						+ Math.round(backPropagation.getCurrTestErr()*10000)/(double)10000);  	
	}

	/**
	 * Starts/continues back propagation x amount of times.
	 */
	public void stepTo(int x) {
		if( backPropagation == null ) {
			initBackProp();
		}
		backPropagation.nnlearn(x, learningRate, momentum);
		setPromptLabel("Step " + backPropagation.getIterationCount()
						+ ", " + "Training Error: "
						+ Math.round(backPropagation.getCurrErr()*10000)/(double)10000 
						+ "\nTest Error: "
						+ Math.round(backPropagation.getCurrTestErr()*10000)/(double)10000);
	}

	public void autoStep() {
		if( autoStepAlive() ) 
			return;
		enableButtons( false );
		setPromptLabel( "Starting Auto Step..." ); 
		if( backPropagation == null ) {
			initBackProp();
		}
		startingCount = backPropagation.getIterationCount();
		auto = new AutoStep( this , false);
		auto.start();
	}
	
	public void initBackProp() {
		backPropagation = new BackPropagation(this, outputNodes, standardized );	
	}

	public boolean autoStepContinue() {
		if( backPropagation == null ) {
			// There is something wrong if it gets here!
			return false;
		}
		if( backPropagation.getIterationCount() - startingCount >= numberOfIterations ) {
			setPromptLabel( "Finished " + backPropagation.getIterationCount() 
					+ " steps.  Training Error: " 
					+ Math.round(backPropagation.getCurrErr()*10000)/(double)10000 
					+ "\nTest Error: "
					+ Math.round(backPropagation.getCurrTestErr()*10000)/(double)10000);
			return false;
		}

		return true;
	}

	public boolean runContinue() {
		if( backPropagation == null ) {
			// There is something wrong if it gets here!
			return false;
		}
		if( backPropagation.getStoppingError() <= stoppingErrVal ) {
			setPromptLabel("Reached target training error of " + Math.round(backPropagation.getCurrErr()*10000)/(double)10000 +
							" after " + backPropagation.getIterationCount() + " steps." 
							+ "\nTest Error: "
							+ Math.round(backPropagation.getCurrTestErr()*10000)/(double)10000);
			return false;
		}

		return true;
	}

	public void runUntilFinished() {
		if( autoStepAlive() ) 
			return;
		enableButtons( false );
		setPromptLabel( "Running until finished..." );
		if( backPropagation == null ) {
			initBackProp();
		}
		startingCount = backPropagation.getIterationCount();
		auto = new AutoStep( this, true ); 
		auto.start();
	}
	
	public void stop() {
		if( auto != null ) 
			auto.quit();
		enableButtons( true );
		repaint();
	}

	public void autoStepEnd() {
		enableButtons( true );
		if( pFrame != null && pFrame.isVisible() ) 
			pFrame.repaint();
		repaint();
	}

	public boolean autoStepAlive() {
		if( auto == null ) 
			return false;
		return auto.isAlive();
	}

	/**
	 * Loads a graph from a text description stored as a URL
	 *
	 * @param filename			  The name of the file to load
	 */
	public String parseText( String file ) {
		nodes = new ArrayList<Node>();
		edges = new ArrayList<Edge>();
		selectedNodes = new ArrayList<Node>();
		selectedEdges = new ArrayList<Edge>();
		
		String errorMessage = new String("");
		String inString = new String("");
		int p1 = -1;
		int p2 = file.indexOf("\n");
		int lineCount = 0;
		try {
			while (p2 != -1) {
				inString = file.substring(p1 + 1, p2).trim();
				if ((inString).length() > 0) {
					if ((inString.charAt(0) != '%') && (inString.indexOf(";") == -1)) {
						errorMessage = new String("Missing semicolon");
						throw (new Exception());
					}

					errorMessage = parseLine( inString.trim() );

					if (errorMessage.length() > 0) {
						throw (new Exception());
					}
				}
				p1 = p2;
				p2 = file.indexOf("\n", p1 + 1);
				lineCount++;
			}
			
			exampleList = new ExampleList();
			errorMessage = exampleList.parseString(file);
			while (errorMessage.equals("You must identify parameters before loading examples")) {
				errorMessage = exampleList.parseString(file);
			}	
			if (!errorMessage.equals("OK") ) {
				throw (new Exception());	 	
			}

			resetBackPropagation();
			learningRate = 0.2;
			return "OK";
		} catch (Exception e) {
			if (errorMessage.length() > 0) {
				return "Error at line " + (lineCount + 1) + " -- " + errorMessage;
			}
			return "Error at line " + (lineCount + 1) + " -- " + e.toString();
		}
	}

	/**
	 * Loads a graph from an XML  description stored as a URL
	 *
	 * @param filename			 The name of the file to load
	 */
	public String parseXMLText( String file ) {
		nodes = new ArrayList<Node>();
		edges = new ArrayList<Edge>();
		selectedNodes = new ArrayList<Node>();
		selectedEdges = new ArrayList<Edge>();
		
		String errorMessage = new String("");
		
		try {
		
			XMLTree tree = new XMLTree();
			tree.readString(file);
			errorMessage = parseXML(tree);
		 
			if (!errorMessage.equals("OK")) {
				throw (new Exception());	 	
			}

			exampleList = new ExampleList();
			errorMessage = exampleList.parseString(file);
			while (errorMessage.equals("You must identify parameters before loading examples")) {
				errorMessage = exampleList.parseString(file);
			}	
			if (!errorMessage.equals("OK")) {
				throw (new Exception());	 	
			}

			resetBackPropagation();
			learningRate = 0.2;
			return "OK";
		} catch (Exception e) {
			String message = "";
			if (errorMessage.length() > 0) {
				message = errorMessage;
			} else {
				message = e.toString();
			}
			return message;
		}
	}	
	
	/**
	 * Parses a single line of text from the text representation
	 *
	 * @param g				The NeuralGraph being built
	 * @param inString		The line of text to parse
	 */
	public String parseLine(String inString) {
		String errorMessage = new String("");
		if (inString.charAt(0) == 'N') {
			// a node
			int p1, p2;

			// get the name
			p1 = inString.indexOf(",");
			p2 = inString.indexOf(",", p1 + 1);
			String name = inString.substring(p1 + 1, p2).trim();	

			// get the position
			p1 = p2;
			p2 = inString.indexOf(",", p1 + 1);
			float x = Float.parseFloat(inString.substring(p1 + 1, p2).trim());

			p1 = p2;
			p2 = inString.indexOf(",", p1 + 1);
			float y = Float.parseFloat(inString.substring(p1 + 1, p2).trim());
			Point pos = new Point(x, y);

			// get the initial parameter value
			p1 = p2;
			p2 = inString.indexOf(";", p1 + 1);
			double paraValue = Double.parseDouble(inString.substring(p1 + 1, p2).trim());
								
			// get the index
			p1 = inString.indexOf(":");
			p2 = inString.indexOf(",");
			Integer ind = new Integer(inString.substring(p1 + 1, p2).trim());
			int index = ind.intValue();

			NeuralNode node = new NeuralNode( this );
			node.setLabel(name);
			node.pos = pos;
			node.setInitialParaValue(paraValue);
			node.setCurrentParaValue(paraValue);
			addNode( node );
			node.index = index;
		} else if (inString.charAt(0) == 'E') {
			// an edge
			int p1, p2;
			// get from node index
			p1 = inString.indexOf(":");	 
			p2 = inString.indexOf(",");
			int from = Integer.parseInt(inString.substring(p1 + 1, p2).trim());
								
			p1 = p2;
			p2 = inString.indexOf(",", p1 + 1);
			int to = Integer.parseInt(inString.substring(p1 + 1, p2).trim());

			p1 = p2;
			p2 = inString.indexOf(";", p1 + 1);												
			double paraValue = Double.parseDouble(inString.substring(p1 + 1, p2).trim());

			NeuralNode fromNode = null;
			NeuralNode toNode = null;
			fromNode = (NeuralNode) nodeFromIndex(from);
			toNode = (NeuralNode) nodeFromIndex(to);
			if (fromNode == null) {
				errorMessage = new String("Error: Node " + from + " does not exist!");
			} else if (toNode == null) {
				errorMessage = new String("Error: Node " + to + " does not exist!");
			} else {
				fromNode.addChild(nodeAt(to));
				toNode.addParent(nodeAt(from));
				NeuralEdge edge = new NeuralEdge( this, fromNode, toNode );
				edge.setInitialParaValue(paraValue);
				edge.setCurrentParaValue(paraValue);
				edge.setLabel( showParam );
				addEdge( edge );
			}
		} else if (inString.charAt(0) == 'T') {			
		// done by ExampleList class
		} else if (inString.charAt(0) == 'X' || inString.charAt(0) == 'A' || inString.charAt(0) == 'B') {
		 // done by ExampleList class						 
		} else if ((inString.charAt(0) == '%') || (inString.charAt(0) == ' ') || (inString.charAt(0) == '\n')) {
			// ignore comments and blank lines
		} else {
			// invalid file
			errorMessage = new String("Invalid format");
		}
		return errorMessage;
	}
	

	/**
	 * Function to parse an XML tree for Neural Network information.
	 * Note that this assumes that example information has already been parsed
	 * by exampleList.
	 */
	public String parseXML(XMLTree tree) {
		String errString = "";
		try {
			// find the <DESCRIPTION> tag
			
			XMLBlock desc = tree.findNetworkTree("description");
			ArrayList<XMLBlock> sdescs = new ArrayList<XMLBlock>();
			ArrayList<XMLBlock> ldescs = new ArrayList<XMLBlock>();
			
			if (desc == null) {
				setShortDesc("");
				setDetailedDesc("");
			}
			
			else {
				sdescs = desc.searchChildTag("short");
				ldescs = desc.searchChildTag("detailed");
				
				if (sdescs == null ||sdescs.size() == 0)
					setShortDesc("");
				
				else 
					setShortDesc(sdescs.get(0).getText());

				if (ldescs == null || ldescs.size()== 0)
					setDetailedDesc("");	

				else 
					setDetailedDesc(ldescs.get(0).getText());
			}
			
			// find the network information
			XMLBlock net = tree.findNetworkTree("network");
			// if there is no network defined, return an error
			if (net == null) {
				errString = "No Neural Network Defined";
				throw (new Exception());	
			}
			// get the nodes
			ArrayList<XMLBlock> nodes = net.searchChildTag("node");
			// if there are no nodes defined, return an error
			if (nodes == null) {
				errString = "No nodes defined";
				throw (new Exception());	
			}
			for (int i = 0; i < nodes.size(); i++) {
				// instantiate new nodes
			 	XMLBlock thisNode = nodes.get(i);
			 	// get the name
			 	ArrayList<XMLBlock> n = thisNode.searchChildTag("name");
				// if there is no name defined, return an error
				if (n == null) {
					errString = "No name defined for a node";
					throw (new Exception());	
				}
				String name = n.get(0).getText().trim();

				// get the position
				ArrayList<XMLBlock> xpos = thisNode.searchChildTag("xpos");
				ArrayList<XMLBlock> ypos = thisNode.searchChildTag("ypos");
				// if there is no xpos defined, return an error
				if (xpos == null) {
					errString = "Node " + name + " has no x-position defined";
					throw (new Exception());	
				}
				// if there is no ypos defined, return an error
				if (ypos == null) {
					errString = "Node "	+ name + " has no y-position defined";
					throw (new Exception());
				}				
				float x = Float.parseFloat( xpos.get(0).getText().trim() );
				float y = Float.parseFloat( ypos.get(0).getText().trim() );
				Point pos = new Point(x, y);
				
				// get the initial parameter value
				ArrayList<XMLBlock> paraVal = thisNode.searchChildTag("weight");
				double paraValue = 0;
				// if there is no weight defined, set weight to 0
				if (paraVal != null) {
					paraValue = Double.parseDouble(paraVal.get(0).getText().trim());		
				}
								
					// get the index
				ArrayList<XMLBlock> ind = thisNode.searchChildTag("index");
				// if there is no index defined, return an error
				if (ind == null) {
					errString = "Node " + name + " does not have an index defined";
					throw (new Exception()); 	
				}
				int index = Integer.parseInt(ind.get(0).getText().trim());
				
				//get the function type
				//if the activation function is not defined, then
				// set it to sigmoid
				ArrayList<XMLBlock> func = thisNode.searchChildTag("function");
				int funcType = NeuralNode.SIGMOID;
				if (func != null) {
					String type = func.get(0).getText().trim();
					if (type.equalsIgnoreCase("exp")) {
						funcType = NeuralNode.EXP;
					} else if (type.equalsIgnoreCase("tanh")) {
						funcType = NeuralNode.TANH;	
					} else if (type.equalsIgnoreCase("linear")) {
						funcType = NeuralNode.LINEAR;
					} else {
						funcType = NeuralNode.SIGMOID;	
					}
				}
				
				NeuralNode node = new NeuralNode( this );
				node.setLabel(name);
				node.pos = pos;
		  		node.setInitialParaValue(paraValue);
		  		node.setCurrentParaValue(paraValue);
		  		node.functionType = funcType;
				addNode( node );
				node.index = index;			
			}
			// get the edges
			ArrayList<XMLBlock> edges = net.searchChildTag("edge");
			// it doesn't matter if there are no edges defined
			if (edges != null) {
				// go through all the edges
				for (int i = 0; i < edges.size(); i++) {
					XMLBlock thisEdge = edges.get(i);
					ArrayList<XMLBlock> start = thisEdge.searchChildTag("startindex");
					// if the start node is not defined, return an error
					if (start == null) {
						errString = "Start node for an edge is not defined";
						throw (new Exception());	
					}
					ArrayList<XMLBlock> end = thisEdge.searchChildTag("endindex");
					// if the end node is not defined, return an error
					if (end == null) {
						errString = "End node for an edge is not defined";
						throw (new Exception());	
					}
					int startIndex = Integer.parseInt(start.get(0).getText().trim());
					int endIndex = Integer.parseInt(end.get(0).getText().trim());
					// get the weight
					ArrayList<XMLBlock> paraVal = thisEdge.searchChildTag("weight");
					double paraValue = 0;
					// if there is no weight defined, set weight to 0
					if (paraVal != null) {
						paraValue = Double.parseDouble(paraVal.get(0).getText().trim());
					}
					// find the nodes associated with the edge
					NeuralNode fromNode = (NeuralNode) nodeFromIndex(startIndex);
					NeuralNode toNode = (NeuralNode) nodeFromIndex(endIndex);
					if (fromNode == null) {
						errString = "Node " + startIndex + " does not exist!";
						throw (new Exception());
					} else if (toNode == null) {
						errString = "Node " + endIndex + " does not exist!";
						throw (new Exception());
					}
					// add the node
					fromNode.addChild(nodeAt(endIndex));
					toNode.addParent(nodeAt(startIndex));
					NeuralEdge edge = new NeuralEdge( this, fromNode, toNode );
					edge.setInitialParaValue(paraValue);
					edge.setCurrentParaValue(paraValue);
					edge.setLabel( showParam );
					addEdge( edge );
				}
			}	
		} catch (Exception e) {
			if (errString.length() > 0) {
				return "Error: " + errString;	
			}
			return "Error: " + e.toString();	
		}
		return "OK";			
	}
	
	/** generate the nnlearn prolog code */
	private String getNNLearnCode() {
		StringBuffer prologText = new StringBuffer();
		
		prologText.append("% Neural-net style learning for parameterized logic\n");
		prologText.append("% programs.  Given a set of examples, a parameterized\n");
		prologText.append("% logic program and a measure of error for each\n");
		prologText.append("% example, this program used back-propagation to tune\n");
		prologText.append("% the parameters. Derivatives are calculated analytically.\n");
		prologText.append("% The applet does NOT use this algorithm, although this will work.\n\n");

		prologText.append("% nnlearn(N,DX,LC,Exs,P0,P1)\n% N is the number of iterations to do\n");
		prologText.append("% DX is the increment to evaluate derivatives\n");
		prologText.append("% LC is the learning constant for gradient descent\n");
		prologText.append("% Exs is a list of all of the examples\n");
		prologText.append("% P0 is the list of parameter settings before the learning\n");
		prologText.append("% P1 is the list of parameter settings after the learning\n");
		prologText.append("nnlearn(0,_,_,Exs,P0,P0) :-\n	total_error(Exs,P0,0,Err0),\n	writeln(['Error = ', Err0]).\n");
		prologText.append("nnlearn(N,DX,LC,Exs,P0,P2) :-\n	update_parms(DX,LC,Exs,P0,P1),\n");
		prologText.append("	N1 is N-1,\n	nnlearn(N1,DX,LC,Exs,P1,P2).\n\n");

		prologText.append("% update_parms(DX,LC,Exs,P0,P1).\n");
		prologText.append("% updates all of parameter in P0 to P1\n");
		prologText.append("update_parms(DX,LC,Exs,P0,P1) :-\n	total_error(Exs,P0,0,Err0),\n");
		prologText.append("	writeln(['Error = ', Err0]),\n	update_each(P0,P0,Err0,Exs,DX,LC,P1).\n\n");

		prologText.append("% update_each(PR,P0,Err0,Exs,DX,LC,P1)\n");
		prologText.append("% updates each parameter in PR.\n");
		prologText.append("% P0 is the initial parameter setting with error Err0\n% Exs is a list of all of the examples\n");
		prologText.append("% DX is the increment to evaluate derivatives\n");
		prologText.append("% LC is the learning constant for gradient descent\n");
		prologText.append("% P1 is the updated parameter settings\nupdate_each([],_,_,_,_,_,[]).\n");
		prologText.append("update_each([val(P,V)|RPs],P0,Err0,Exs,DX,LC,\n										[val(P,NV)|NPs]) :-\n");
		prologText.append("	V1 is V+DX,\n	total_error(Exs,[val(P,V1)|P0],0,Nerr),\n");
		prologText.append("	NV is V+LC*(Err0-Nerr)/DX,\n	update_each(RPs,P0,Err0,Exs,DX,LC,NPs).\n\n");

		prologText.append("% total_error(Exs,P,Err0,Err1)\n% P is a list of parameter settings\n");
		prologText.append("% Exs is a list of examples\n% Err1 is Err0 plus the sum of the errors on examples\n");
		prologText.append("total_error([],_,Err,Err).\ntotal_error([Ex|Exs],P,PErr,TErr) :-\n");
		prologText.append("	pprove(error(Ex,Err),P),!,\n	NErr is PErr+Err,\n	total_error(Exs,P,NErr,TErr).\n\n");

		prologText.append("% `<-' is the object-level `if'\n:- op(1150, xfx, <- ).\n");
		prologText.append("% `&' is the object level conjunction.\n:- op(950,xfy, &).\n\n");

		prologText.append("% pprove(G,P) means prove G with parameter settings P.\n");
		prologText.append("% P is a list of the value assignments of the\n% form val(P,V)\n\n");

		prologText.append("pprove(true,_) :- !.\npprove((A & B),P) :-\n	!,\n	pprove(A,P),\n");
		prologText.append("	pprove(B,P).\npprove((A is E),P) :-\n	!,\n");
		prologText.append("	eval(E,A,P).\npprove(H,P) :-\n	(H <- B),\n	pprove(B,P).\n\n");

		prologText.append("% eval(E,V,P) true if expression E has value V\n");
		prologText.append("% given parameter settings P\neval((A+B),S,P) :-\n");
		prologText.append("	!,\n	eval(A,VA,P),\n	eval(B,VB,P),\n	S is VA+VB.\n");
		prologText.append("eval((A*B),S,P) :-\n	!,\n	eval(A,VA,P),\n	eval(B,VB,P),\n");
		prologText.append("	S is VA*VB.\neval((A-B),S,P) :-\n");
		prologText.append("	!,\n	eval(A,VA,P),\n	eval(B,VB,P),\n	S is VA-VB.\n");
		prologText.append("eval((A/B),S,P) :-\n	!,\n	eval(A,VA,P),\n	eval(B,VB,P),\n");
		prologText.append("	S is VA/VB.\neval(sigmoid(E),V,P) :-\n	!,\n	eval(E,EV,P),\n");
		prologText.append("	V is 1/(1+exp(-EV)).\neval(N,N,_) :-\n	number(N),!.\n");
		prologText.append("eval(N,V,P) :-\n	member(val(N,V1),P),!,V=V1.\n		% it only sees the first value for the parameter\n");
		prologText.append("eval(N,V,_) :-\n	writeln(['Arithmetic Error: ',V,' is ',N]),fail.\n\n");

		prologText.append("member(A, [A|_]).\nmember(A, [H|L]) :- member(A, L).\n\n");

		prologText.append("writeln([]) :- nl.\nwriteln([X|Xs]) :- write(X), writeln(Xs).\n");
		
		return prologText.toString();
	}

	public String generatePrologTextRep() {
		StringBuffer prologText = new StringBuffer("% Auto-generated on ");
		prologText.append(Calendar.getInstance().getTime().toString()).append("\n\n");
		//put in the nnlearn prolog code
		prologText.append(getNNLearnCode()).append("\n\n");
		
		NeuralNode curNode, parentNode;
		NeuralEdge curEdge;
		
		/* *****************************************
		 * EXAMPLE DATA
		 * *****************************************/
		prologText.append("% Example Data:\n% prop(Object,Attribute,Value)\n");
		
		Iterator<Hashtable<Integer, String>> exampleItr = getExampleList().getTrainingArrayList().iterator();
		//the input/output nodes iterator
		Iterator<Node> ioNodeItr;
		Hashtable<Integer, String> oneExample;
		//the index of the current input or output node
		Integer curIONodeIndex;
		int index = 1;
		while(exampleItr.hasNext()) {
			oneExample = exampleItr.next();
			
			// INPUT NODE EXAMPLES
			ioNodeItr = inputNodes.iterator();
			while(ioNodeItr.hasNext()) {
				curIONodeIndex = ioNodeItr.next().getIndex();
				curNode = (NeuralNode) nodeFromIndex(curIONodeIndex.intValue());
				
				prologText.append("prop(e").append(index).append(", \'");
				if(curNode.getLabel().equals(""))
					prologText.append("node").append(curIONodeIndex);
				else
					prologText.append(curNode.getLabel());
				prologText.append("\', ").append(oneExample.get(curIONodeIndex)).append(") <- true.\n");
			}
			
			// OUTPUT NODE EXAMPLES
			ioNodeItr = outputNodes.iterator();
			while(ioNodeItr.hasNext()) {
				curIONodeIndex = ioNodeItr.next().getIndex();
				curNode = (NeuralNode) nodeFromIndex(curIONodeIndex.intValue());
				
				if(oneExample.get(curIONodeIndex) == null)
					continue;
				
				prologText.append("prop(e").append(index).append(", \'");
				if(curNode.getLabel().equals(""))
					prologText.append("node").append(curIONodeIndex);
				else
					prologText.append(curNode.getLabel());
				
				prologText.append("\', ").append(oneExample.get(curIONodeIndex)).append(") <- true.\n");
			}
			prologText.append("\n");
			index++;
		}
		
		prologText.append("\n");
		
		/* *****************************************
		 * OUTPUT NODES
		 * *****************************************/
		Iterator<Node> indexItr;
		StringBuffer textBuffer;
		ioNodeItr = outputNodes.iterator();
		while(ioNodeItr.hasNext()) {
			curIONodeIndex = ioNodeItr.next().getIndex();
			curNode = (NeuralNode) nodeFromIndex(curIONodeIndex.intValue());
			
			prologText.append("predicted_prop(Obj, \'");
			if(curNode.getLabel().equals(""))
				prologText.append("node").append(curIONodeIndex);
			else
				prologText.append(curNode.getLabel());
			prologText.append("\', V) <-\n");
			textBuffer = new StringBuffer("	V is sigmoid(w_" + curNode.getParameterIndex());
			
			indexItr = curNode.getParentNodes().iterator();
			index = 0;
			while(indexItr.hasNext()) {
				curIONodeIndex = indexItr.next().getIndex();
				parentNode = (NeuralNode) nodeFromIndex(curIONodeIndex.intValue());
				
				prologText.append("	prop(Obj, \'");
				if(parentNode.getLabel().equals(""))
					prologText.append("node").append(curIONodeIndex);
				else
					prologText.append(parentNode.getLabel());
				prologText.append("\', I_").append(index).append(") &\n");

				curEdge = (NeuralEdge) getEdge(parentNode.getIndex(), curNode.getIndex());
				if (curEdge != null) {
					textBuffer.append(" + w_").append(curEdge.getParameterIndex());
				}
				textBuffer.append(" * I_").append(index++);
			}
			prologText.append(textBuffer).append(").\n\n");
		}
		
		/* *****************************************
		 * HIDDEN NODES
		 * *****************************************/
		Iterator<Node> nodeItr = nodes.iterator();
		while(nodeItr.hasNext()) {
			curNode = (NeuralNode) nodeItr.next();
			if((curNode.getNumParentNodes() > 0) && (curNode.getNumChildrenNodes() > 0)) {
				//node is a hidden node
				prologText.append("prop(Obj, \'");
				if(curNode.getLabel().equals(""))
					prologText.append("node").append(curNode.getIndex());
				else
					prologText.append(curNode.getLabel());
				prologText.append("\', V) <-\n");
				textBuffer = new StringBuffer("	V is sigmoid(w_" + curNode.getParameterIndex());
				
				indexItr = curNode.getParentNodes().iterator();
				index = 0;
				while(indexItr.hasNext()) {
					curIONodeIndex = indexItr.next().getIndex();
					parentNode = (NeuralNode) nodeFromIndex(curIONodeIndex.intValue());
					
					prologText.append("	prop(Obj, \'");
					if(parentNode.getLabel().equals(""))
						prologText.append("node").append(curIONodeIndex);
					else
						prologText.append(parentNode.getLabel());
					prologText.append("\', I_").append(index).append(") &\n");

					curEdge = (NeuralEdge) getEdge(parentNode.getIndex(), curNode.getIndex());
					if (curEdge != null) {
						textBuffer.append(" + w_").append(curEdge.getParameterIndex());
					}
					textBuffer.append(" * I_").append(index++);
				}
				prologText.append(textBuffer).append(").\n\n");
			}
		}
		
		/* *****************************************
		 * ERROR FUNCTION
		 * *****************************************/
		prologText.append("error(Obj, E) <-\n");
		textBuffer = new StringBuffer("	E is ");
		ioNodeItr = outputNodes.iterator();
		index = 1;
		String name;
		while(ioNodeItr.hasNext()) {
			curIONodeIndex = ioNodeItr.next().getIndex();
			curNode = (NeuralNode) nodeFromIndex(curIONodeIndex.intValue());

			if(curNode.getLabel().equals(""))
				name = "node" +curIONodeIndex;
			else
				name = curNode.getLabel();
			prologText.append("	predicted_prop(Obj, \'").append(name).append("\', VC_").append(index).append(") &\n");
			prologText.append("	prop(Obj, ").append(name).append("\', V_").append(index).append(") &\n");
			
			textBuffer.append("(VC_").append(index).append(" - V_").append(index).append(") * (VC_").append(index).append(" - V_").append(index).append(")");
			if(ioNodeItr.hasNext())
				textBuffer.append(" + ");
		}
		prologText.append(textBuffer).append(".\n\n\n");
		
		/* *****************************************
		 * EXAMPLE QUERY
		 * *****************************************/
		prologText.append("% Example Query:\n% ?nnlearn(50, 0.01, 1.0, [");
		exampleItr = getExampleList().getTrainingArrayList().iterator();
		index = 1;
		while(exampleItr.hasNext()) {
			prologText.append("e").append(index++);
			exampleItr.next();
			if(exampleItr.hasNext())
				prologText.append(", ");
		}
		prologText.append("], [");
		
		nodeItr = nodes.iterator();
		while(nodeItr.hasNext()) {
			curNode = (NeuralNode) nodeItr.next();
			if(curNode.getNumParentNodes() > 0) {
				prologText.append("val(w_").append(curNode.getParameterIndex()).append(", ");
				prologText.append(curNode.getInitialParaValue()).append(")");
				if(nodeItr.hasNext())
					prologText.append(", ");
			}
		}
		
		Iterator<Edge> edgeItr = edges.iterator();
		while(edgeItr.hasNext()) {
			curEdge = (NeuralEdge) edgeItr.next();
			prologText.append(", val(w_").append(curEdge.getParameterIndex()).append(", ");
			prologText.append(curEdge.getInitialParaValue()).append(")");
		}
		
		prologText.append("], P1).\n");
		return prologText.toString();
	}

	public String generateTextRep() {
		Calendar now = Calendar.getInstance();
		StringBuffer textBuffer = new StringBuffer("% Auto-generated on ");
		textBuffer.append(now.getTime().toString()).append("\n\n");

		textBuffer.append("% Nodes\n% N: index, node_name, x_position, y_position, parameter_value;\n");
		NeuralNode curNode;
		Iterator<Node> itr = nodes.iterator();
		while(itr.hasNext()) {
			curNode = (NeuralNode) itr.next();
			textBuffer.append("N: ").append(curNode.getIndex()).append(", ");
			if (curNode.getLabel().equals("")) {
				textBuffer.append("Node ").append(curNode.getIndex());
			} else {
				textBuffer.append(curNode.getLabel());
			}
			textBuffer.append(", ").append(curNode.pos.x).append(", ").append(curNode.pos.y);
			textBuffer.append(", ").append(curNode.getCurrentParaValue()).append(";\n");
		}

		textBuffer.append("\n% Edges\n");
		textBuffer.append("% E: from_node_index, to_node_index, parameter_value;\n");
		NeuralEdge curEdge;
		Iterator<Edge> edgeItr = edges.iterator();
		while(edgeItr.hasNext()) {
			curEdge = (NeuralEdge) edgeItr.next();
			textBuffer.append("E: ").append(((NeuralNode) curEdge.start).getIndex()).append(", ");
			textBuffer.append(((NeuralNode) curEdge.end).getIndex()).append(", ").append(curEdge.getCurrentParaValue()).append(";\n");
		}

		textBuffer.append("\n% Examples\n");
		textBuffer.append("% T: input1's title, input2's title, ..., output1's title, output2's title, ...;\n");
		textBuffer.append("% \t note that the order must be input nodes in ascending index order, \n% \t followed by output nodes in ascending index order.\n");
		textBuffer.append("% A: or X: training - input1's value, input2's value, ..., output1's value, output2's value, ...;\n");
		textBuffer.append("% B: test - input1's value, input2's value, ..., output1's value, output2's value, ...;\n");			
		if ((exampleList != null) && (exampleList.getTrainingArrayList().size() > 0 || exampleList.getTestArrayList().size() > 0)) {
			textBuffer.append(exampleList.generateTextRep());
		} else {
			textBuffer.append("% NO TEST AND TRAINING EXAMPLES\n");
		}
		
		return textBuffer.toString();
	}
	
	/**
	 * Generates an XML representation of the neural network
	 */
	public String generateXMLTextRep() {
		StringBuffer xmlText = new StringBuffer("<?xml version=\"1.0\" ?>\n<MLDBIF>\n");
		xmlText.append("<DB>\n\n");
		
		//short description
		xmlText.append("<DESCRIPTION>\n\t<SHORT>").append(getShortDesc()).append("</SHORT>");
		//detailed description
		xmlText.append("\n\t<DETAILED>").append(getDetailedDesc()).append("</DETAILED>\n</DESCRIPTION>\n\n");
		
		xmlText.append("<!-- Neural Network Definition -->\n<NETWORK>\n");
		// iterate through all the nodes
		if (numNodes() > 0) {
			xmlText.append("\n\t<!-- Node Definitions -->\n\n");	
		}
		Iterator<Node> itr = nodes.iterator();
		NeuralNode curNode;
		while(itr.hasNext()) {
			curNode = (NeuralNode) itr.next();
			
			xmlText.append("\t<NODE>\n\t\t<NAME>");
			if (curNode.getLabel().equals("")) {
				xmlText.append("Node ").append(curNode.getIndex());
			} else {
				xmlText.append(curNode.getLabel());
			}
			xmlText.append("</NAME>\n\t\t<WEIGHT>").append(curNode.getCurrentParaValue()).append("</WEIGHT>\n");
			xmlText.append("\t\t<XPOS>").append(curNode.pos.x).append("</XPOS>\n");
			xmlText.append("\t\t<YPOS>").append(curNode.pos.y).append("</YPOS>\n");
			xmlText.append("\t\t<INDEX>").append(curNode.getIndex()).append("</INDEX>\n");
			switch(curNode.functionType) {
				case NeuralNode.LINEAR:
					xmlText.append("\t\t<FUNCTION>linear</FUNCTION>\n");
					break;
				case NeuralNode.TANH:
					xmlText.append("\t\t<FUNCTION>tanh</FUNCTION>\n");	
					break;
				case NeuralNode.EXP:
					xmlText.append("\t\t<FUNCTION>exp</FUNCTION>\n");	
					break;
				case NeuralNode.SIGMOID:
					xmlText.append("\t\t<FUNCTION>sigmoid</FUNCTION>\n");	
					break;
			}
			xmlText.append("\t</NODE>\n");
		}
		// iterate over all edges
		if (numEdges() > 0) {
			xmlText.append("\n	<!-- Edge Definitions -->\n\n");	
		}
		NeuralEdge edge;
		for( int i = 0; i < numEdges(); i++ ) {
		  	edge = (NeuralEdge)edgeAt(i);
			xmlText.append("\t<EDGE>\n");
			xmlText.append("\t\t<STARTINDEX>").append(((NeuralNode)edge.start).getIndex()).append("</STARTINDEX>\n");
			xmlText.append("\t\t<ENDINDEX>").append(((NeuralNode)edge.end).getIndex()).append("</ENDINDEX>\n");
			xmlText.append("\t\t<WEIGHT>").append(edge.getCurrentParaValue()).append("</WEIGHT>\n");
			xmlText.append("\t</EDGE>\n");
		}
		
		xmlText.append("\n</NETWORK>\n\n");
		xmlText.append(exampleList.generateXMLTextRep());			
		xmlText.append("</DB>\n");
		xmlText.append("</MLDBIF>\n\n");
		
		return xmlText.toString();
	}

	/**
	 * Delete the selected entities.
	 * @author Byron Knoll
	 */
	public void deleteSelected() {
		
		// Remove the selected edges.
		Iterator<Edge> edgeItr = selectedEdges.iterator();
		Edge curEdge;
		while(edgeItr.hasNext()) {
			curEdge = edgeItr.next();
			
			// Remove the parent-child relationships at the nodes at each end of the edge.
			NeuralNode tmpNode = (NeuralNode) curEdge.end;
			NeuralNode tmpNode2 = (NeuralNode) curEdge.start;
			tmpNode.removeParent(tmpNode2);
			tmpNode2.removeChild(tmpNode);
			
			// Remove itself from all connected nodes
			curEdge.removeFromNodes();
			// Remove from edges in graph
			edges.remove(curEdge);
			if(paramsChanged())
				rebuildData();
		}
		selectedEdges.clear();

		// Remove the selected nodes.
		Iterator<Node> nodeItr = selectedNodes.iterator();
		Node curNode;
		ArrayList[] allNodeEdges;
		while(nodeItr.hasNext()) {
			curNode = nodeItr.next();
			
			// Remove this node from all the parent-child relationships of other nodes.
			Iterator<Node> nodeItr2 = getNodes();
			while (nodeItr2.hasNext()) {
				NeuralNode tmpNode = (NeuralNode)nodeItr2.next();
				tmpNode.removeParent(curNode);
				tmpNode.removeChild(curNode);
			}
			
			// Get all connected edges.
			allNodeEdges = curNode.getAllEdges();
			// Remove edges in.
			while (allNodeEdges[0].size() > 0) {
				curEdge = (Edge) allNodeEdges[0].get(0);
				// Remove itself from all connected nodes.
				curEdge.removeFromNodes();
				// Remove from edges in graph.
				edges.remove(curEdge);
			}
			// Remove edges out.
			while (allNodeEdges[1].size() > 0) {
				curEdge = (Edge) allNodeEdges[1].get(0);
				// Remove itself from all connected nodes.
				curEdge.removeFromNodes();
				// Remove from edges in graph.
				edges.remove(curEdge);
			}
			nodes.remove(curNode);
			
			if(paramsChanged())
				rebuildData();
		}
		selectedNodes.clear();
		
		setInputOutputNodes( true );
		backPropagation = null;
	}

	public void showPlot() {
		if( pFrame == null ) {
			pFrame = new PlotFrame( this );
		}
		pFrame.open();
	}

	/**
	 * Updates the y-axis of the plot frame.
	 */
	public void clearPlotFrame() {
		if( pFrame != null ) {
			pFrame.clear();
		}
	}
	
	/**
	  * Updates the y-axis of the plot frame.
	  */
	 public void clear() {
		if( pFrame != null ) {
			pFrame.clear();
			pFrame.setVisible(false);
		}
	 }

	/**
	 * Returns values required for plotting.
	 */
	public IntList[] getPlotVals() {
		if( backPropagation == null ) 
			return null;
		return backPropagation.getPlotVals();
	}

	/**
	 * Returns values required for plotting.
	 */
	public IntList[] getTestPlotVals() {
		if( backPropagation == null ) 
			return null;
		return backPropagation.getTestPlotVals();
	}

	/**
	 * Returns the current total error.
	 */
	public double totalError() {
		if( backPropagation == null ) 
			return 0.1;
		return backPropagation.getCurrErr();
	}


	public void updatePlot() {
		if( pFrame != null && pFrame.isVisible() ) 
			pFrame.repaint();
	}
	
	/**
	 *	Function to determine the answer of the neuralgraph given
	 *	a set of input values and the index of the output node
	 *	that the user is interested in (this index should correspond to
	 *	the outputnodes ArrayList indices). 
	 * 
	 * @param valueArray	an array of strings representing double values to be solved.
	 * @param outputIndex	the index of the output node to be solved (0 is the first, 1 
	 * 						is the second, etc...)
	 */
	public String outputValue (String[] valueArray, int outputIndex) {
		return nodeValue(valueArray, outputNodes.get(outputIndex));
	}
	
	/**
	 * Function to determine the answer of the neuralgraph given a set of input values 
	 * and the node that the user is interested in.
	 * 
	 * @param valueArray	an array of strings representing double values to be solved.
	 * @param node			the node to get the value for.
	 */
	public String nodeValue (String[] valueArray, Node node) {
		double value = 0;
		try {
			int numInputs = inputNodes.size();
			
			Double values[] = new Double[numInputs];
			// record and cast the values passed in valueArray to doubles
			for (int i = 0; i < numInputs; i++) {
				values[i] = Double.valueOf(valueArray[i]);
			}
	
			if (backPropagation == null) {
				initBackProp();	
			}
			
			value = backPropagation.solveNodeOutput(node.getIndex(), values);
			value = Math.round(value * 10000.0) / 10000.0;
		} catch (NumberFormatException e) {
			System.err.println("Number format error");
		}
		return Double.toString(value);
	}

	/**
	 * Function to find the corresponding indices of correct and incorrect
	 * values in the test set. There is an efficiency issue when there is a lot
	 * of examples - this will get called three times. To do: add a function that
	 * does the computation, and store the indices, so when this function is
	 * called, it will just return the already-created indices.
	 * 
	 * @param indexMode	Determines if the function will return correct
	 * incorrect, or undefined indices (constants from NeuralGraph)
	 * @param range	Determines the range within which an answer is considered correct.
	 * @param outputIndex	The index of the output node to be calculated.
	 * 
	 * @see NeuralGraph
	 * 	
	 */
	public ArrayList<Integer> getTestExampleIndices(int indexMode, double range, int outputIndex) {
		ArrayList<Integer> exampleIndices = new ArrayList<Integer>();

		// get the test set offsetPoint
		ArrayList<Hashtable<Integer, String>> testSet = exampleList.getTestArrayList();

		// for each member of the test set, get the 'real' value
		// and the predicted value, and compare the two. If the 
		// predicted neuralvalue is within value-range <= value < value+range,
		// then it is 'correct'. Else, it is incorrect. 
		for (int i = 0; i < testSet.size(); i++) {
			int currIndex = inputNodes.size() + outputIndex;
			double value = Double.parseDouble( testSet.get(i).get(new Integer(currIndex)));

			String[] stringInputs = new String[inputNodes.size()];
			for (int j = 0; j < stringInputs.length; j++) {
				stringInputs[j] = testSet.get(i).get(new Integer(j));
			}
			double neuralValue = Double.parseDouble(outputValue(stringInputs, outputIndex));

			if (indexMode == NeuralGraph.CORRECT) {
				if ((Math.abs(value - neuralValue)) < range) {
					exampleIndices.add(new Integer(i));
				}
			} else if (indexMode == NeuralGraph.INCORRECT) {
				if ((Math.abs(value - neuralValue)) >= range) {
					exampleIndices.add(new Integer(i));
				}
				// I'm not sure this actually happens, but this is here
				// just in case (this is for unclassified answers)
			} else if (!((Math.abs(value - neuralValue)) < range) && !((Math.abs(value - neuralValue)) >= range)) {
				exampleIndices.add(new Integer(i));
			}
		}

		return exampleIndices;
	}
	
	/**
	 * Function to set the input and output parameters in examplelist
	 */
	public void setIOParameters() {
		for (int i = 0; i < (inputNodes.size()+outputNodes.size()); i++)
		{
			if (i < inputNodes.size()) exampleList.setInputParam(true,i);
			else exampleList.setInputParam(false,i);
		}
	}

	/**
	 * Clears the plotting points.
	 */
	public void clearPlotPts() {
		if( backPropagation == null ) 
			return;
		backPropagation.clearPlotPts();
	}

	public void disposeWindows() {
		if( pFrame != null )  
			pFrame.dispose();
	}
	
	public void setMomentum(double m) {
		momentum = m;	
	}
	
	public double getMomentum() {
		return momentum;
	}
	
	public String getShortDesc() {
		return shortDescription;
	}
	
	public void setShortDesc(String description) {
		shortDescription = description;
		if (description.trim().equals("") || description == null) {
			shortDescription = "";
		}
	}
	
	public String getDetailedDesc() {
		return detailedDescription;
	}
	
	public void setDetailedDesc(String description) {
		detailedDescription = description;
		if (description.trim().equals("") || description == null) {
			detailedDescription = "";
		}
	}
}