/**
 * File: NeuralWindow.java
 * Sep 9, 2000
 *
 ***
 * @ arksey
 * This class is a subclass of GraphWindow class and is the is the main interface class.
 * The user directly interacts with this class and the functionality is implemented in other classes. 
 */

package AIspace.neural;

import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.awt.event.WindowEvent;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.URL;

import javax.swing.ButtonGroup;
import javax.swing.ImageIcon;
import javax.swing.JApplet;
import javax.swing.JButton;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JToggleButton;

import AIspace.graphToolKit.GraphConsts;
import AIspace.graphToolKit.GraphWindow;
import AIspace.graphToolKit.dialogs.MessageDialog;
import AIspace.graphToolKit.dialogs.OpenLocationDialog;
import AIspace.graphToolKit.help.HelpFrame;
import AIspace.graphToolKit.help.HelpMenu;
import AIspace.neural.dialogs.CalculateOutputDialog;
import AIspace.neural.dialogs.ExampleDialog;
import AIspace.neural.dialogs.InitializationOptionsDialog;
import AIspace.neural.dialogs.LearningOptionsDialog;
import AIspace.neural.dialogs.NeuralDescriptionDialog;
import AIspace.neural.dialogs.NeuralProblemDialog;
import AIspace.neural.dialogs.NeuralWizard;
import AIspace.neural.dialogs.ParameterInputDialog;
import AIspace.neural.dialogs.StoppingOptionsDialog;
import AIspace.neural.dialogs.TestStatisticsDialog;
import AIspace.neural.help.NeuralHelpCanvas;

/** The main window class.
* @author Shinjiro Sueda
*/
public class NeuralWindow extends GraphWindow {

	private static final long serialVersionUID = 1L;
	
	//Dialog boxes
	private StoppingOptionsDialog sod;
	private LearningOptionsDialog lod;
	private InitializationOptionsDialog iod;

	private JMenu speed;
	private JMenuItem initialization;
	private JMenuItem stopping;
	private JMenuItem learning;
	private JMenuItem textRepresentationMenuItem;
	private JMenuItem xmlRepresentationMenuItem;
	private JMenuItem descriptionMenuItem;
	private JCheckBoxMenuItem showParam;

	private JButton fakeButton; //in solve,used to toggle buttons off

	private boolean showParameters;

	protected NeuralWizard wizard;

	//	solve tool bars icons for buttons
	protected ImageIcon stepIcon;
	protected ImageIcon fineStepIcon;
	protected ImageIcon autoSearchIcon;
	protected ImageIcon stepTargetIcon;
	protected ImageIcon stopIcon;
	protected ImageIcon resetIcon;
	protected ImageIcon quizIcon;
	protected ImageIcon invertGraphIcon;
	protected ImageIcon sumStatIcon;
	protected ImageIcon showPlotIcon;
	protected ImageIcon editIcon;
	protected ImageIcon randomIcon;
	

	//
	// Methods
	//

	/**
	 * Constructor.
	 */
	public NeuralWindow(JApplet parent) {
		super(parent);
		
		// Set the medium line width as selected (default for neural applet).
		narrowLine.setSelected(false);
		mediumLine.setSelected(true);
		wideLine.setSelected(false);
		
		initializeAppletInfo();

		setTitle(appletTitle + " --- untitled.txt");
		setSize(955, 700);
		switchTab("Create");

		initialize();
		construct();

		centerWindow();
		setVisible(true);
	}
	
	/** sets the applet information */
	private void initializeAppletInfo() {
		appletTitle = "Neural Applet Version 4.3.7";
		appletName = "neural";
		aboutText = appletTitle + "\n\n" +
			"Inspired by neurons and their connections in the brain, neural network\n" +
			"is a representation used in machine learning. After running the\n" +
			"back-propagation learning algorithm on a given set of examples,\n" +
			"the neural network can be used to predict outcomes for any set of\n" +
			"input values.\n\n"+
			"This applet was written by Kevin O'Neill, Shinjiro Sueda, Leslie Tung,\n" +
			"Joseph Roy Santos, Nicole Arksey, Kyle Porter, and Byron Knoll, with\n" +
			"help from Mike Cline, Holger Hoos, Peter Gorniak, Cristina Conati,\n" +
			"Alan Mackworth, and David Poole.";
	}
	
	/**
	 * Constructs the rest of the window not constructed in graph tool kit
	 *
	 */
	private void construct() {
		JPanel bottom = new JPanel();

		bottom.add(new JLabel("Color Key:", JLabel.RIGHT));
		bottom.add(new Legend());
		bottomPanel.add(bottom);
	}
	/**
	 * Initializes variables and sets mode to C_CREATE_NODE 
	 *
	 */
	private void initialize() {
		stepIcon = createImageIcon("images/neural/step.png");
		fineStepIcon = createImageIcon("images/common/finestep.png");
		randomIcon = createImageIcon("images/neural/randomPara.png");
		stopIcon = createImageIcon("images/neural/stop.gif");
		editIcon = createImageIcon("images/neural/Edit24.gif");
		showPlotIcon = createImageIcon("images/neural/showPlot.png");
		sumStatIcon = createImageIcon("images/neural/History24.gif");
		quizIcon = createImageIcon("images/neural/quiz.gif");
		stepTargetIcon = createImageIcon("images/neural/stepTarget.png");

		solveToolBarWithText();
		toolBar.setVisible(false);
		createToolBarWithText();
		toolBar.setVisible(true);

		stopping.setEnabled(false);
		learning.setEnabled(false);
		speed.setEnabled(false);

		returnCanvas().setSubmode(GraphConsts.C_CREATE_NODE);
		setPromptLabel("Click the canvas to create a node.");
		createButModes[1].setSelected(true);
	}

	/**
	 *initializes the canvas with the scroll pane, overrides getCanvas() in graphWindow
	 */
	protected void getCanvas() {
		canvas = new NeuralCanvas(this, false);
		canvas.setPreferredSize(new Dimension(15000, 10000));

		scrollPanel = new JScrollPane(canvas, JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		scrollPanel.setPreferredSize(new Dimension(500, 400));

		scrollPanel.getHorizontalScrollBar().setMaximum(15000);
		scrollPanel.getHorizontalScrollBar().setValue(7500);

		scrollPanel.getVerticalScrollBar().setMaximum(10000);
		scrollPanel.getVerticalScrollBar().setValue(5000);
	}
	
	/** clear any existing graph from the canvas */
	protected void resetCanvas() {
		int speed = ((NeuralGraph) canvas.graph).dt;
		canvas.reset();
		// Reset the parameter initialization options dialog.
		iod = new InitializationOptionsDialog (this, (NeuralGraph) canvas.graph, false);
		((NeuralGraph) canvas.graph).dt = speed;
	}

	/**
	 *  @see AIspace.graphToolKit.GraphWindow#createToolBar()
	 */
	protected void createToolBar() {
		//create tool bar
		createButModes = new JToggleButton[7];

		setCreateButtons(0, editIcon, "View/Edit Examples");
		setCreateButtons(1, createNodeIcon, "Create Node");
		setCreateButtons(2, createEdgeIcon,"Create Edge" );
		setCreateButtons(3, selectIcon,"Select" );
		setCreateButtons(4, deleteIcon, "Delete");
		setCreateButtons(5, setPropsIcon, "Set Properties");
		
		createToolBarLayout();
	}

	/**
	 * @see AIspace.graphToolKit.GraphWindow#createToolBarWithText()
	 */
	protected void createToolBarWithText() {
		//create tool bar
		createButModes = new JToggleButton[7];

	    setCreateButtonsWithText(0,  "View/Edit Examples",   editIcon,  "View/Edit Examples");
		setCreateButtonsWithText(1, "Create Node", createNodeIcon, "Create Node");
		setCreateButtonsWithText(2, "Create Edge", createEdgeIcon,"Create Edge" );
		setCreateButtonsWithText(3, "   Select   ", selectIcon, "Select" );
		setCreateButtonsWithText(4,"   Delete   ", deleteIcon, "Delete");
		setCreateButtonsWithText(5,"Set Properties", setPropsIcon, "Set Properties");

		createToolBarLayout();
	}
	
	private void createToolBarLayout() {
		toolBar.removeAll();

		createModeButGroup = new ButtonGroup();
		createModeButGroup.add(fakeButton);
	
		for (int i = 0; i <= 5; i++) {
			if(i == 1)
				toolBar.addSeparator();
			createModeButGroup.add(createButModes[i]);
			toolBar.add(createButModes[i]);
			createButModes[i].setFont(toolBarFont);
		}

		algoText.setText(" ");
	}

	/**
	 * @see AIspace.graphToolKit.GraphWindow#solveToolBarWithText()
	 */
	public void solveToolBarWithText() {
		super.solveToolBar();

		solveButModes = new JToggleButton[10];

		setSolveButtonsWithText(0, "View/Edit Examples", editIcon, "View/Edit Examples");
		setSolveButtonsWithText(1, " Initialize Parameters", randomIcon, "Initialize Parameters");
		setSolveButtonsWithText(2, "      Step      ", stepIcon, "Step");
		setSolveButtonsWithText(3, "Step " + ((NeuralGraph) canvas.graph).getNumberOfIterations() + "X", fineStepIcon, 
									"Step " + ((NeuralGraph) canvas.graph).getNumberOfIterations() + "X");
		setSolveButtonsWithText(4, "Step to Target Error", stepTargetIcon, "Step to Target Error");
		setSolveButtonsWithText(5, "     Stop      ", stopIcon, "Stop ");
		setSolveButtonsWithText(6, " Calculate Output  ", quizIcon, "Calculate Output");
		setSolveButtonsWithText(7, "     Show Plot   ", showPlotIcon, "Show Plot");
		setSolveButtonsWithText(8, "Summary Statistics", sumStatIcon, "Summary Statistics");

		solveToolBarLayout();
	}

	/**
	 * @see AIspace.graphToolKit.GraphWindow#solveToolBar()
	 */
	public void solveToolBar() {
		super.solveToolBar();

		solveButModes = new JToggleButton[9];

		setSolveButtons(0, editIcon, "View/Edit Examples");
		setSolveButtons(1, randomIcon, "Initialize Parameters");
		setSolveButtons(2, stepIcon, "Step");
		setSolveButtons(3, fineStepIcon, "Step " + ((NeuralGraph) canvas.graph).getNumberOfIterations() + "X");
		setSolveButtons(4, stepTargetIcon, "Step to Target Error");
		setSolveButtons(5, stopIcon, "Stop Search");
		setSolveButtons(6, quizIcon, "Calculate Output");
		setSolveButtons(7, showPlotIcon, "Show Plot");
		setSolveButtons(8, sumStatIcon, "Summary Statistics");

		solveToolBarLayout();
	}
	
	private void solveToolBarLayout() {
		solveModeButGroup = new ButtonGroup();

		fakeButton = new JButton();
		solveModeButGroup.add(fakeButton);	
		
		for (int i = 0; i < 9; i++) {
			if(i == 1)
				toolBar.addSeparator();
			solveModeButGroup.add(solveButModes[i]);
			toolBar.add(solveButModes[i]);
			solveButModes[i].setFont(toolBarFont);
		}
		solveButModes[5].setEnabled(false);
	}

	/**
	 * @see AIspace.graphToolKit.GraphWindow#createNewGraph()
	 */
	protected void createNewGraph() {
		setTitle(appletTitle + " --- untitled.txt");
		returnCanvas().reset();
	}

	/**
	 * called when solve or create tabs are called.
	 *  @see AIspace.graphToolKit.GraphWindow#switchTab(java.lang.String)
	 */
	public void switchTab(String str) {
		super.switchTab(str);

		if (str == "Create") {
			textRepresentationMenuItem.setText("View/Edit Text Representation");
			textRepresentationMenuItem.setActionCommand(textRepresentationMenuItem.getText());
			xmlRepresentationMenuItem.setText("View/Edit XML Representation");
			xmlRepresentationMenuItem.setActionCommand(xmlRepresentationMenuItem.getText());
			descriptionMenuItem.setText("View/Edit Graph and Data Description");
			descriptionMenuItem.setActionCommand(descriptionMenuItem.getText());
			speed.setEnabled(false);
			stopping.setEnabled(false);
			learning.setEnabled(false);
		} else if (str == "Solve") {
			textRepresentationMenuItem.setText("View Text Representation");
			textRepresentationMenuItem.setActionCommand(textRepresentationMenuItem.getText());
			xmlRepresentationMenuItem.setText("View XML Representation");
			xmlRepresentationMenuItem.setActionCommand(xmlRepresentationMenuItem.getText());
			descriptionMenuItem.setText("View Graph and Data Description");
			descriptionMenuItem.setActionCommand(descriptionMenuItem.getText());
			setPromptLabel("Click on an entity to view its properties.");
			solveButModes[5].setEnabled(false);
			speed.setEnabled(true);
			stopping.setEnabled(true);
			learning.setEnabled(true);
			returnCanvas().repaint();
		}
	}

	//Creating Menu Bar
	protected JMenu createFileMenu() {
		JMenuItem mItem;
		JMenu m = new JMenu("File");;
		m.setMnemonic(KeyEvent.VK_F);
				
		mItem = new JMenuItem("Create New Graph");
		mItem.setActionCommand(mItem.getText());
		mItem.addActionListener(this);
		mItem.setMnemonic(KeyEvent.VK_N);
		m.add(mItem);
		
		mItem = new JMenuItem("Load Sample Graph and Data");
		mItem.setActionCommand(mItem.getText());
		mItem.setMnemonic(KeyEvent.VK_G);
		mItem.addActionListener(this);
		m.add(mItem);
		JMenuItem loadsamplegraphwizard = new JMenuItem("Load Sample Data");
		loadsamplegraphwizard.setActionCommand(loadsamplegraphwizard.getText());
		loadsamplegraphwizard.setMnemonic(KeyEvent.VK_D);
		loadsamplegraphwizard.setDisplayedMnemonicIndex(12);
		loadsamplegraphwizard.addActionListener(this);
		m.add(loadsamplegraphwizard);
		
		boolean hasLocalAccess = hasLocalAccess();
		if(hasLocalAccess) {
			mItem = new JMenuItem("Load Graph and Data From File");
			mItem.setActionCommand(mItem.getText());
			mItem.setMnemonic(KeyEvent.VK_L);
			mItem.addActionListener(this);
			m.add(mItem);
			JMenuItem opengraphwizard = new JMenuItem("Load Data From File");
			opengraphwizard.setActionCommand(opengraphwizard.getText());
			opengraphwizard.addActionListener(this);
			opengraphwizard.setMnemonic(KeyEvent.VK_O);
			m.add(opengraphwizard);
		}
		// applets can't connect to hosts on the internet other than the originating host
		// so we should only allow this option when running as an application - ry Nov2005
		mItem = new JMenuItem("Load Graph and Data From URL");
		mItem.setActionCommand(mItem.getText());
		mItem.setMnemonic(KeyEvent.VK_U);
		mItem.addActionListener(this);
		m.add(mItem);
		JMenuItem openlocwizard = new JMenuItem("Load Data From URL");
		openlocwizard.setActionCommand(openlocwizard.getText());
		openlocwizard.setMnemonic(KeyEvent.VK_R);
		openlocwizard.setDisplayedMnemonicIndex(16);
		openlocwizard.addActionListener(this);
		m.add(openlocwizard);
		m.addSeparator();
		
		if(hasLocalAccess) {
			mItem = new JMenuItem("Save Graph and Data");
			mItem.setActionCommand(mItem.getText());
			mItem.addActionListener(this);
			mItem.setMnemonic(KeyEvent.VK_S);
			m.add(mItem);
		}
		
		mItem = new JMenuItem("Print");
		mItem.setActionCommand(mItem.getText());
		mItem.addActionListener(this);
		mItem.setMnemonic(KeyEvent.VK_P);
		m.add(mItem);

		mItem = new JMenuItem("Quit");
		mItem.setActionCommand(mItem.getText());
		mItem.addActionListener(this);
		mItem.setMnemonic(KeyEvent.VK_Q);
		m.add(mItem);

		return m;
	}

	protected JMenu createEditMenu() {
		JMenu m = new JMenu("Edit");

		JMenuItem mItem = new JMenuItem("View Prolog Code");
		mItem.addActionListener(this);
		m.add(mItem);
		textRepresentationMenuItem = new JMenuItem("View/Edit Text Representation");
		textRepresentationMenuItem.setMnemonic(KeyEvent.VK_T);
		textRepresentationMenuItem.setDisplayedMnemonicIndex(10);
		textRepresentationMenuItem.addActionListener(this);
		m.add(textRepresentationMenuItem);
		xmlRepresentationMenuItem = new JMenuItem("View/Edit XML Representation");
		xmlRepresentationMenuItem.setMnemonic(KeyEvent.VK_X);
		xmlRepresentationMenuItem.addActionListener(this);
		m.add(xmlRepresentationMenuItem);
		
		descriptionMenuItem = new JMenuItem( "View/Edit Graph and Data Description" );
		descriptionMenuItem.setMnemonic(KeyEvent.VK_D);
		descriptionMenuItem.setDisplayedMnemonicIndex(10);
		descriptionMenuItem.setActionCommand( descriptionMenuItem.getText() );
		descriptionMenuItem.addActionListener( this );
		m.add( descriptionMenuItem );
		
		return m;
	}

	protected JMenu createViewMenu() {
		JMenu m = super.createViewMenu();

		m.addSeparator();

		showParam = new JCheckBoxMenuItem("Show Parameters", true);
		showParam.setMnemonic(KeyEvent.VK_P);
		showParam.addActionListener(this);
		m.add(showParam);

		return m;
	}

	protected JMenu createGraphOptionsMenu() {
		JMenu m = new JMenu("Neural Options");
		ButtonGroup group = new ButtonGroup();
		speed = new JMenu("Auto Step Speed");
		JRadioButtonMenuItem xFast = new JRadioButtonMenuItem("Very Fast (0 s)", false);
		xFast.addActionListener(this);
		speed.add(xFast);
		group.add(xFast);
		JRadioButtonMenuItem fast = new JRadioButtonMenuItem("Fast (0.1 s)", true);
		fast.addActionListener(this);
		speed.add(fast);
		group.add(fast);
		JRadioButtonMenuItem medium = new JRadioButtonMenuItem("Medium (0.5 s)", false);
		medium.addActionListener(this);
		speed.add(medium);
		group.add(medium);
		JRadioButtonMenuItem slow = new JRadioButtonMenuItem("Slow (1 s)", false);
		slow.addActionListener(this);
		speed.add(slow);
		group.add(slow);

		m.add(speed);
		m.addSeparator();

		initialization = new JMenuItem("Parameter Initialization Options...");
		initialization.addActionListener(this);
		m.add(initialization);
		learning = new JMenuItem("Learning Options...");
		learning.addActionListener(this);
		m.add(learning);
		stopping = new JMenuItem("Stopping Conditions...");
		stopping.addActionListener(this);
		m.add(stopping);
		m.addSeparator();
		JCheckBoxMenuItem standardize = new JCheckBoxMenuItem("Normalize Inputs", false);
		standardize.addActionListener(this);
		m.add(standardize);

		return m;
	}

	protected JMenu createHelpMenu() {
		return new HelpMenu("neural", this);
	}

	/**
	 * Makes the menu bar.
	 */
	protected JMenuBar createMenuBar() {
		JMenuBar menuBar = super.createMenuBar();

		//create and add the new menus in the right place
		menuBar.add(createEditMenu(), 1);
		menuBar.add(createGraphOptionsMenu(), 3);

		return menuBar;
	}

	public void loadWizard(File file) {
		try {
			loadWizard(new BufferedReader(new FileReader(file.getPath())));
		} catch (Exception e) {
			fileName = null;
			showMessage("Error", e.toString());
		}
	}

	/**
	 * Loads a preset graph.
	 */
	public void load(BufferedReader in) {
		try {
			((NeuralGraph) returnCanvas().graph).disposeWindows();
			String line;
			StringBuffer allText = new StringBuffer();

			while ((line = in.readLine()) != null) {
				allText.append(line).append("\n");
			}
			in.close();

			ExampleList example = new ExampleList();
			String errorMessage = example.parseString(allText.toString());
			if ((errorMessage.length() > 0) && (!errorMessage.equals("OK"))) {
				new MessageDialog(this).open("Error Loading File", errorMessage);
				return;
			}
			resetCanvas();
			((NeuralCanvas) returnCanvas()).parseXML(allText.toString());
			((NeuralGraph) ((NeuralCanvas) returnCanvas()).graph).setInputOutputNodes(true);
			if (canvas.getMode()==GraphConsts.SOLVE)
			{
				((NeuralGraph) canvas.graph).setTotalErrorAfterInitialization();
			}
			canvas.autoscale();
			setTitle(appletTitle + " --- " + fileName);
			((NeuralGraph)canvas.graph).repaint();
		} catch (Exception e) {
			e.printStackTrace(System.out);
			super.setPromptLabel("Error" + e.toString());
		}
	}

	public void save(File file) {
		if (!file.getName().endsWith(".xml"))
			file = new File(file.getAbsolutePath() + ".xml");
		try {
			PrintWriter out = new PrintWriter(new FileOutputStream(file));
			out.println(((NeuralGraph) canvas.graph).generateXMLTextRep());
			out.close();
			fileName = file.getName();
			if (!file.getName().equals("undo.xml"));
				setTitle(appletTitle + " --- " + fileName);
		} catch (java.io.FileNotFoundException e) {
			System.out.println(e);
		}
	}
	
	/**
	 * * @returns true if stop is enabled
	 */
	public boolean isStopEnabled() {
		return solveButModes[5].isEnabled();
	}
	/**
	 * * @param bool
	 * sets stepX, step to target to bool and stop to ! bool
	 */
	public void enableButtons(boolean bool) {
		solveButModes[3].setEnabled(bool);
		solveButModes[4].setEnabled(bool);
		solveButModes[5].setEnabled(!bool);
	}
	//
	// The following implements ActionListener
	//

	/**
	 * Invoked when an action occurs.
	 */
	public void actionPerformed(ActionEvent e) {
		String arg = e.getActionCommand().trim();
		
		///////////////////create mode/////////////////
		if (e.getSource() == createButModes[3]) {
			returnCanvas().setSubmode(GraphConsts.C_SELECT);
			setPromptLabel("Click on an entity to select or drag the mouse to select multiple entities.");
			return;
		} else if (e.getSource() == createButModes[4]) {
			returnCanvas().setSubmode(GraphConsts.C_DELETE);
			setPromptLabel("Click on an entity to delete.");
			return;
		///////////////////solve mode///////////////////
		} else if (arg.equals("Initialize Parameters")) {
			if (!validGraph()) {
				return;
			}
			((NeuralGraph) canvas.graph).initializeParaValues();
			((NeuralGraph) canvas.graph).setTotalErrorAfterInitialization();

			solveModeButGroup.setSelected(fakeButton.getModel(), true);
			return;
		} else if (arg.equals("Step")) {
			if (!validGraph()) {
				System.out.println("Not valid step");
				return;
			}

			((NeuralGraph) canvas.graph).step();
			canvas.repaint();
			solveModeButGroup.setSelected(fakeButton.getModel(), true);
			return;
		} else if (arg.equals("Step " + ((NeuralGraph) canvas.graph).getNumberOfIterations() + "X")) {
			if (!validGraph()) {
				return;
			}
			((NeuralGraph) canvas.graph).autoStep();
			solveButModes[5].setEnabled(true);
			solveModeButGroup.setSelected(fakeButton.getModel(), true);
			return;
		} else if (arg.equals("Step to Target Error")) {
			if (!validGraph()) {
				return;
			}
			((NeuralGraph) canvas.graph).runUntilFinished();
			solveButModes[5].setEnabled(true);
			solveModeButGroup.setSelected(fakeButton.getModel(), true);
			return;
		} else if (arg.equals("Stop")) {
			((NeuralGraph) canvas.graph).stop();
			solveButModes[5].setEnabled(false);
			solveModeButGroup.setSelected(fakeButton.getModel(), true);
			return;
		} else if (arg.equals("Calculate Output")) {
			if (!validGraph()) {
				return;
			}

			//new InputExampleDialog(this, null, (NeuralGraph) canvas.graph, "Calculate Output Values");
			new CalculateOutputDialog(this, (NeuralGraph) canvas.graph);
			solveModeButGroup.setSelected(fakeButton.getModel(), true);
			return;
		} else if (arg.equals("Show Plot")) {
			if (!validGraph()) {
				return;
			}

			((NeuralGraph) canvas.graph).showPlot();
			solveModeButGroup.setSelected(fakeButton.getModel(), true);
			return;
		} else if (arg.equals("Summary Statistics")) {
			new TestStatisticsDialog(this, (NeuralGraph) canvas.graph);
			solveModeButGroup.setSelected(fakeButton.getModel(), true);

			return;
			////////////////////used in solve and create modes////////////////////
		} else if (
			arg.equals("View/Edit Examples")) {

			if (validGraph()) {
				new ExampleDialog(this, (NeuralGraph) canvas.graph, ((NeuralGraph) canvas.graph).getExampleList());
			} else {
				showMessage("View/Edit Examples", "At least one input and output node are needed to manipulate examples.");
			}
			solveModeButGroup.setSelected(fakeButton.getModel(), true);
			if (returnCanvas().getMode() == GraphConsts.CREATE) {
				int subMode = returnCanvas().getSubmode();
				switch (subMode) {

					case GraphConsts.C_CREATE_NODE :
						createButModes[1].setSelected(true);
						break;
					case GraphConsts.C_CREATE_EDGE :
						createButModes[2].setSelected(true);
						break;
					case GraphConsts.C_SELECT :
						createButModes[3].setSelected(true);
						break;
					case GraphConsts.C_DELETE :
						createButModes[4].setSelected(true);
						break;
					case GraphConsts.C_SET_PROP :
						createButModes[5].setSelected(true);
						break;
				}
			}
			return;
			//////////////////menu options////////////////////
		} else if (arg.equals("Parameter Initialization Options...")) {
			if (iod == null) {
				iod = new InitializationOptionsDialog(this, (NeuralGraph) canvas.graph, true);
			} else
				iod.open();
			return;
		} 
		else if (arg.equals("Learning Options...")) {
			if (lod == null) {
				lod = new LearningOptionsDialog(this, (NeuralGraph) canvas.graph);
			} else
				lod.open();
			return;
		}
		else if (arg.equals("Stopping Conditions...")) {
			if (sod == null) {
				sod = new StoppingOptionsDialog(this, (NeuralGraph) canvas.graph, "Stopping Conditions");
			} else
				sod.open();

			solveButModes[3].setText("Step " + ((NeuralGraph) returnCanvas().graph).getNumberOfIterations() + "X");
			solveButModes[3].setToolTipText("Step " + ((NeuralGraph) returnCanvas().graph).getNumberOfIterations() + "X");
			solveButModes[3].setActionCommand(solveButModes[3].getToolTipText());
			if (super.isButtonTextShowing()) {
				solveToolBarWithText();
			} else
				solveToolBar();

			return;
		} else if (arg.equals("Load Data From File")) {
			JFileChooser load = new JFileChooser();
			load.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
			load.setLocation(0, 0);
			int returnval = load.showOpenDialog(this);
			if (returnval == JFileChooser.APPROVE_OPTION) {
				File file = load.getSelectedFile();
				fileName = file.getName();
				loadWizard(file);
			}
			return;
		} else if (arg.equals("Load Sample Data")) {
			new NeuralProblemDialog(this, true).open();
			return;
		} else if (arg.equals("Load Graph and Data From File")) {
			openGraph();
			return;
		} else if (arg.equals("Load Graph and Data From URL")) {
			openLocation();
			return;
		} else if (arg.equals("Save Graph and Data")) {
			saveGraph();
			return;
		} else if (arg.equals("Load Sample Graph and Data")) {
			new NeuralProblemDialog(this, false).open();
			return;
		} else if (arg.equals("Load Data From URL")) {
			OpenLocationDialog old = new OpenLocationDialog(this);
			String urlString = old.url;
			if (urlString != null) {
				InputStream iStream = null;
				fileName = urlString.substring(urlString.lastIndexOf("/") + 1);
				try {
					iStream = new URL(urlString).openStream();
				} catch (Exception exc) {
					super.setPromptLabel("Error" + exc.toString());
				}
				if (iStream != null) {
					BufferedReader file = new BufferedReader(new InputStreamReader(iStream));
					loadWizard(file);
				}
			}
			return;
		} else if (arg.equals("View Prolog Code")) {
			((NeuralCanvas) canvas).openPrologTextRep();
		} else if (arg.equals("View/Edit Text Representation")) {
			((NeuralCanvas) canvas).openTextRep(true);
		} else if (arg.equals("View/Edit XML Representation")) {
			((NeuralCanvas) canvas).openXMLTextRep(true);
		} else if (arg.equals("View Text Representation")) {
			((NeuralCanvas) canvas).openTextRep(false);
		} else if (arg.equals("View XML Representation")) {
			((NeuralCanvas) canvas).openXMLTextRep(false);
		} else if (arg.equals("View/Edit Graph and Data Description")) {
			new NeuralDescriptionDialog(this,(NeuralGraph) canvas.graph, true).open();
		} else if (arg.equals("View Graph and Data Description")) {
			new NeuralDescriptionDialog(this,(NeuralGraph) canvas.graph, false).open();
		} else if (arg.equals("Legend for Nodes/Edges")) {
			new HelpFrame("Legend", new NeuralHelpCanvas(isAntiAliasingEnabled()), 450, 675);
			return;
		} else if (arg.equals("Very Fast (0 s)")) {
			((NeuralGraph) canvas.graph).dt = 0;
			return;
		} else if (arg.equals("Fast (0.1 s)")) {
			((NeuralGraph) canvas.graph).dt = 100;
			return;
		} else if (arg.equals("Medium (0.5 s)")) {
			((NeuralGraph) canvas.graph).dt = 500;
			return;
		} else if (arg.equals("Slow (1 s)")) {
			((NeuralGraph) canvas.graph).dt = 1000;
			return;
		} else if (arg.equals("Average Sum of Squares Error")) {
			((NeuralGraph) canvas.graph).setErrorType(NeuralGraph.AVERAGE_SUMSQUARES);
			return;
		} else if (arg.equals("Total Sum of Squares Error")) {
			((NeuralGraph) canvas.graph).setErrorType(NeuralGraph.SUMSQUARES);
			return;
		} else if (arg.equals("Show Parameters")) {
			((NeuralGraph) canvas.graph).setShowParam(showParam.getState());
			return;
		} else if (arg.equals("Normalize Inputs")) {
			NeuralGraph g = (NeuralGraph) canvas.graph;
			ExampleList ex = g.getExampleList();
			g.standardized = !(g.standardized);
			if ((ex != null || g.getExampleList().getTrainingArrayList().size() > 0)
				&& (g.getInputNodes().size() > 0 && g.getOutputNodes().size() > 0)) {
				g.initBackProp();
			}
			
			return;
		}
		super.actionPerformed(e);
	}

	/**
	 * Helper function that loads a sample example with the Neural Wizard.
	 * This is analogous to load(String, String)
	 */
	public void loadWizard(String dir, String filename) {
		try {
			this.fileName = filename;
			if (applet != null) {
				loadWizard(new BufferedReader(new InputStreamReader(new URL(applet.getCodeBase() + dir + filename).openStream())));
			} else {
				loadWizard(new BufferedReader(new FileReader(dir + filename)));
			}
		} catch (Exception e) {
			filename = null;
			super.setPromptLabel("Error" + e.toString());
		}
	}

	/**
	 * Helper function that loads an example with the NeuralWizard
	 * This is analogous to load(BufferedReader)
	 */
	public void loadWizard(BufferedReader file) {
		StringBuffer all = new StringBuffer();
		ExampleList example = null;
		try {
			((NeuralGraph) returnCanvas().graph).disposeWindows();
			String line;
			while ((line = file.readLine()) != null) {
				all.append(line).append("\n");
			}
		} catch (Exception ex) {
			new MessageDialog(this).open("Error Loading File", "There has been an error loading the file. Terminating wizard...");
			return;
		}

		example = new ExampleList();
		String errorMessage = example.parseString(all.toString());
		setPromptLabel("");

		while (errorMessage.equals("You must identify parameters before loading examples")) {
			new ParameterInputDialog((JFrame) canvas.parent, example);
			errorMessage = example.parseString(all.toString());
		}
		if ((errorMessage.length() > 0) && (!errorMessage.equals("OK"))) {
			new MessageDialog(this).open("Error Loading File", errorMessage);
			return;
		}

		resetCanvas();
		wizard = new NeuralWizard(this, (NeuralGraph) canvas.graph, example, fileName);
		if (canvas.getMode()==GraphConsts.SOLVE)
		{
			((NeuralGraph) canvas.graph).setTotalErrorAfterInitialization();
		}
		setTitle(appletTitle + " --- " + fileName);
	}

	//Save applet properties to be restored again later
	protected void saveProperties() {
		super.saveProperties();
		showParameters = showParam.getState();
	}

	//	Restore all saved applet properties
	protected void restoreProperties() {
		super.restoreProperties();
		if(returnCanvas().getMode() == GraphConsts.CREATE){
			createButModes[1].setSelected(true);
			returnCanvas().setSubmode(GraphConsts.C_CREATE_NODE);
			setPromptLabel("Click the canvas to create a node.");		
		}
		showParam.setSelected(showParameters);
		((NeuralGraph) canvas.graph).setShowParam(showParam.getState());
	}
	/**
	 * Helper function for ActionPerformed that checks if the canvas has
	 * a valid graph. If it does, return true, else return false.
	 * The check just looks to see if there is an input node. This
	 * is because input nodes only exist if there is at least 2 nodes
	 * and they are connected
	 */
	private boolean validGraph() {
		NeuralGraph g = (NeuralGraph) canvas.graph;
		if (g.getInputNodes().size() > 0 && g.getOutputNodes().size() > 0) {
			return true;
		}
		return false;
	}

	//
	// The end of the interface ActionListener
	//

	/**
	 * Invoked when the user attempts to close the window from the 
	 * window's system menu.
	 */
	public void windowClosing(WindowEvent e) {
		super.windowClosing(e);
		if (sod != null)
			sod.dispose();
		if (lod != null)
			lod.dispose();
		((NeuralGraph) canvas.graph).disposeWindows();
		dispose();
	}

	public static void main(String[] args) {
		NeuralWindow window = new NeuralWindow(null);
		
		if (args.length > 0) {
			// The first argument is used to load a problem from a URL.
			String url = args[0];
			try {
				window.fileName = url.substring(url.lastIndexOf("/") + 1);
				window.load(new BufferedReader(new InputStreamReader(new URL(url).openStream())));
			} catch (Exception exc) {
				window.showMessage("Error", exc.toString());
			}
		}
	}
}