/**
 * PlotCanvas.java
 * Aug 14, 2000
 */


package AIspace.neural;


import java.awt.Canvas;
import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.PrintJob;
import java.awt.Toolkit;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.JFrame;

import AIspace.neural.intList.IntList;


/**
 * @author Shinjiro Sueda
 */
public class PlotCanvas extends Canvas implements MouseListener
{

   //
   // Fields
   //

	private Font font;				// font used
	private FontMetrics fm;			// FontMetrics to calculate font size
	private int width;				// width of the canvas
	private int height;				// height of the canvas
	private int width_adjusted;		// width of the canvas minus left/right pads
	private int height_adjusted;	// height of the canvas minus top/bottom pads
	private float xStep;			// # of pixels for each x
	private float yStep;			// # of pixels for each y
	
	private IntList xVals;			// x values of points to draw
	private IntList yVals;			// y values of points to draw
	private IntList xTestVals;
	private IntList yTestVals;
	private int maxTestX;
	
	private int maxY;				// maximum y value
	private int maxX;				// maximum x value
	private int skipX;				// grid fineness for x
	private int skipY;				// grid fineness for y

	private final static int LEFT_PAD = 100;
	private final static int RIGHT_PAD = 100;
	private final static int TOP_PAD = 50;
	private final static int BOTTOM_PAD = 50;

	private boolean useLogScale = false;
	private int mouseX;
	private int mouseY;
	private String mouseStr;
	private int redX;
	private int redY;

	/**
	 * Constructor.
	 */
	public PlotCanvas() {
		setBackground( Color.white );
		addMouseListener( this );
		font = new Font( "SansSerif", Font.PLAIN, 12 );
		fm = getFontMetrics( font );
		xVals = new IntList();
		yVals = new IntList();
		xTestVals = new IntList();
		yTestVals = new IntList();
		maxTestX = 1;
		xStep = 5;
		maxX = 1;
		mouseX = -100;
		mouseY = -100;
		mouseStr = "";
		redX = -100;
		redY = -100;
	}


	/**
	 * Sets the step sizes.
	 */
	public void resetSteps( int maxY ) {
		updateSize();

		this.maxY = maxY;
		if( maxY <= 0 ) {
			maxY = 1;
			yStep = 50;
		} else {
			yStep = height_adjusted/(float)maxY;
		}

		skipX = round( 50.0/xStep );
		skipY = round( 20.0/yStep );
	}


	/** resets the number of steps to the default */
	public void resetSteps() {
		resetSteps( maxY );
	}


	/**
	 * Round the number to {10, 25, 50, or 100}.
	 */
	private int round( double d ) {
		int digits = 0;
		double tmp = d;
		while( tmp >= 1 ) {
			tmp /= 10;
			digits++;
		}
		if( digits <= 0 ) 
			digits = 1;
		
		int mult = (int)Math.pow( 10, digits );
		if( d/mult < 0.1 ) 
			return (int)(0.1*mult);
		else if( d/mult < 0.25 ) 
			return (int)(0.25*mult);
		else if( d/mult < 0.5 ) 
			return (int)(0.5*mult);
		else 
			return mult;
	}
  

	/**
	 * Updates the width/height vars.
	 */
	public void updateSize() {
		width = getSize().width;
		height = getSize().height;
		width_adjusted = getWidth() - LEFT_PAD - RIGHT_PAD;
		height_adjusted = getHeight() - TOP_PAD - BOTTOM_PAD;
	}


	/**
	 * returns the current width of this canvas.
	 */
	public int getWidth() {
		return width;
	}


	/**
	 * Returns the current height of this canvas.
	 */
	public int getHeight() {
		return height;
	}


	/**
	 * Clears the points.
	 */
	public void clear() {
		xVals = null;
		yVals = null;
		xTestVals = null;
		yTestVals = null;
		maxY = 1;
		repaint();
	}


	/**
	 * Toggles logarithmic scale on and off.
	 */
	public void useLogScale( boolean bool ) {
		useLogScale = bool;
		repaint();
	}


	/**
	 * Redraws the plot.  Updates the maxY and maxX values.
	 */
	public void redraw( IntList[] vals, IntList[] testVals ) {
		if( vals == null ) 
			return;
		if (testVals == null) 
			return;
		
		xVals = vals[0];
		yVals = vals[1];
		xTestVals = testVals[0];
		yTestVals = testVals[1];

		if( xVals.first() == -1 ) 
			return;
		
		maxX = xVals.last();
		xStep = (float)width_adjusted/(float)maxX;
		maxY = yVals.first();
		int tmp = yVals.next();
		while( tmp != -1 ) {
			if( tmp > maxY ) maxY = tmp;
			tmp = yVals.next();
		}

		if( xTestVals.first() == -1 ) 
			return;
		
		maxTestX = xTestVals.last();
		xStep = (float)width_adjusted/(float)maxTestX;
		tmp = yTestVals.next();
		while( tmp != -1 ) {
			if( tmp > maxY ) 
				maxY = tmp;
			
			tmp = yTestVals.next();
		}

		resetSteps();
		repaint();
	}


	/**
	 * Called by the graph class after each step, etc.
	 */
	public void paint( IntList[] vals, IntList[] testVals ) {
		if( vals == null ) return;
		if (testVals == null) return;
		
		xTestVals = testVals[0];
		yTestVals = testVals[1];
		
		xVals = vals[0];
		yVals = vals[1];

		maxX = xVals.last();
		if( maxX == -1 ) maxX = 1;
		xStep = (float)width_adjusted/(float)maxX;
		if( maxX == 1 )
		{
			if( yVals.first() > maxY ) maxY = yVals.first();
		}
		if( yVals.last() > maxY ) maxY = yVals.last();
		
		maxTestX = xTestVals.last();
		if( maxTestX == -1 ) maxTestX = 1;
		xStep = (float)width_adjusted/(float)maxTestX;
		if( maxTestX == 1 )
		{
			if( yTestVals.first() > maxY ) maxY = yTestVals.first();
		}
		if( yTestVals.last() > maxY ) maxY = yTestVals.last();		
		
		resetSteps();
		
		repaint();
	}	


	/**
	 * Overrides update() in component to avoid flickering.
	 */
	public void update( Graphics g ) {
		paint( g );
	}


	/**
	 * Paints the components.
	 */
	public void paint( Graphics g ) {
		g.clearRect( 0, 0, width, height );

		drawGrid( g );
		drawPoints( g );
		drawLabels( g );

		g.setColor( Color.red );
		g.drawOval( (int)(redX*xStep) + LEFT_PAD - 1, height - BOTTOM_PAD - (int)(redY*yStep), 2, 2 );
		
		g.setColor( Color.white );
		g.fillRect( mouseX, mouseY - fm.getAscent(), fm.stringWidth( mouseStr ), fm.getHeight() );
		g.setColor( Color.black );
		g.drawString( mouseStr, mouseX, mouseY );
	}


	/**
	 * Draws the grid lines.
	 */
	private void drawGrid( Graphics graphics ) {
		graphics.setColor( Color.black );

		// Draw the x axis
		int xStart = LEFT_PAD;
		if( xStart < 0 ) xStart = 0;
		int xEnd = width - RIGHT_PAD;
		graphics.drawLine( xStart, height - BOTTOM_PAD, xEnd, height - BOTTOM_PAD );

		// Draw the y axis
		graphics.drawLine( LEFT_PAD, height - BOTTOM_PAD, LEFT_PAD, TOP_PAD );

		// Draw the vertical grid lines
		graphics.setColor( Color.lightGray );
		int i = 0;
		float percent;
		float x2;
		float x = LEFT_PAD + xStep*skipX;
		while( x <= LEFT_PAD + width_adjusted + 1 ) {
			i += skipX;
			x2 = x - LEFT_PAD;
		
			if( useLogScale ) {
				percent = x2/width_adjusted;
				x2 = x2*percent*percent;
		 
			}
			x2 = x2 + LEFT_PAD;
			if( x >= 0 ) {
				graphics.drawLine( (int)x2, TOP_PAD, (int)x2, height - BOTTOM_PAD );
				graphics.drawString( ( i + "" ),
											(int)x2 - fm.stringWidth( i + "" )/2,
											height - BOTTOM_PAD + fm.getHeight() );
			}
			x += xStep*skipX;
		}

		// Draw the horizontal grid lines
		i = 0;
		String yStr;
		for( float y = height - BOTTOM_PAD - yStep*skipY;  y >= TOP_PAD; y -= yStep*skipY ) {
			xStart = LEFT_PAD;
			graphics.drawLine( xStart, (int)y, width - RIGHT_PAD, (int)y );
			i += skipY;
			yStr = PlotCanvas.doubleToString( i/100.0 );
			if( xStart == 0 ) {
				xStart = width - RIGHT_PAD + fm.stringWidth( " " );
			} else {
				xStart = xStart - fm.stringWidth( yStr + " " );
			}
			graphics.drawString( yStr, xStart, (int)y + fm.getHeight()/2 );
		}
	}


	/**
	 * Draws the points.
	 */
	private void drawPoints( Graphics graphics ) {
		if( xVals == null || yVals == null ) 
			return;
		
		int x = xVals.first();
		int y = yVals.first();
		int xPrev = -1;
		int yPrev = -1;
		int r = 1;

		int xStart = LEFT_PAD;
		int yStart = height - BOTTOM_PAD;
		float percent = 1;
		graphics.setColor( Color.blue );
		
		while( x != -1 && y != -1 ) {
			if( useLogScale ) {
				percent = percent( x );
				x = (int)(x*xStep*percent*percent);
			} else {
				x = (int)(x*xStep);
			}
			x = x + xStart;
			y = (int)(y*yStep);
			y = yStart - y;
			graphics.drawRect( x - r, y - r, r*2, r*2 );
			if( xPrev != -1 && yPrev != -1 ) {
				graphics.drawLine( xPrev, yPrev, x, y );
			}
			xPrev = x;
			yPrev = y;
			x = xVals.next();
			y = yVals.next();
		}
		
		if( xTestVals == null || yTestVals == null ) 
			return;
		
		int xTest = xTestVals.first();
		int yTest = yTestVals.first();
		int xTestPrev = -1;
		int yTestPrev = -1;
		int rTest = 1;

		int xTestStart = LEFT_PAD;
		int yTestStart = height - BOTTOM_PAD;
		float testPercent = 1;
		graphics.setColor( Color.orange);
		
		while( xTest != -1 && yTest != -1 ) {
			if( useLogScale ) {
				testPercent = percent( xTest );
				xTest = (int)(xTest*xStep*testPercent*testPercent);
			} else { 
				xTest = (int)(xTest*xStep);
			}
			xTest = xTest + xTestStart;
			yTest = (int)(yTest*yStep);
			yTest = yTestStart - yTest;
			graphics.drawRect( xTest - rTest, yTest - rTest, rTest*2, rTest*2 );
			if( xTestPrev != -1 && yTestPrev != -1 ) {
				graphics.drawLine( xTestPrev, yTestPrev, xTest, yTest );
			}
			xTestPrev = xTest;
			yTestPrev = yTest;
			xTest = xTestVals.next();
			yTest = yTestVals.next();
		}
	}


	/**
	 * Draws the labels like axis names, etc.
	 */
	private void drawLabels( Graphics graphics ) {
		String xName = "Step Count";
		String yName = "Total Error";
		graphics.setColor( Color.black );
		graphics.drawString( xName, width/2 - fm.stringWidth( xName )/2, height - fm.getHeight() );
		int xStart = LEFT_PAD/2;
		int yNameWidth = fm.stringWidth( yName );
		if( xStart + yNameWidth < 0 ) return;
		graphics.drawString( yName, xStart - yNameWidth/2, height/2 - fm.getHeight()/2 );

		xStart = width - LEFT_PAD + 5;
		if( yVals != null && yVals.last() != -1 ) {
			graphics.setColor(Color.blue);
			graphics.drawLine(xStart, height/3-fm.getHeight()/2, xStart+55, height/3-fm.getHeight()/2);
			graphics.drawRect(xStart+27, height/3-fm.getHeight()/2-2, 4, 4);
			int y = height/3 - fm.getHeight();
			 graphics.setColor(Color.black);
	
			String err = PlotCanvas.doubleToString( yVals.last()/100.000 );
			graphics.drawString( err+" Train", xStart, y );
		}
		
		if( yTestVals != null && yTestVals.last() != -1 ) {
			graphics.setColor(Color.orange);
			graphics.drawLine(xStart, (2*height/3)-fm.getHeight()/2, xStart+55, (2*height/3)-fm.getHeight()/2);
			graphics.drawRect(xStart+27, (2*height/3)-fm.getHeight()/2-2, 4, 4);
			int y = (2*height/3) - fm.getHeight();
			String err = PlotCanvas.doubleToString( yTestVals.last()/100.000 );
			graphics.setColor(Color.black);
			graphics.drawString( err+" Test", xStart, y );
		}
	}


	/**
	 * Gets the percentage of this x point relative to the max point.
	 */
	private float percent( int x ) {
		float percent;
		if( maxX == 0 ) percent = 0;
		else percent = (float)x/(float)xVals.last();
		return percent;
	}
	

	/**
	 * Returns a double up to 2 decimal places.
	 */
	public static String doubleToString( double d ) {
		double tmp = Math.round( d*100 )/100.0;
		return tmp + "";
	}


	/**
	 * Prints the graph.
	 */
	public void print( JFrame parent ) {
		// Get the print job object
		PrintJob pj = Toolkit.getDefaultToolkit().getPrintJob(parent, "Title", null );

		if( pj == null ) {
			System.out.println( "The print request was cancelled" );
			return;
		}

		// Get graphics context for page
		Graphics g = pj.getGraphics();

		// Get the image to print
		Image im = createImage( width, height );
		Graphics graphics = im.getGraphics();
		drawGrid( graphics );
		drawPoints( graphics );
		drawLabels( graphics );
		
		// Paint the image
		g.drawImage( im, 0, 0, this );
		
		g.dispose();
		
		// Start printing.
		pj.end();
	}


	public void mousePressed( MouseEvent e ) {
		if( xVals == null ) return;
		if( useLogScale ) return;
		
		mouseX = e.getX();
		mouseY = e.getY();
		float clickedX = (mouseX - LEFT_PAD)/xStep;
		int x = xVals.first();
		int y = yVals.first();
		int prevX = x;
		int prevY = y;
		while( x < clickedX && x != -1 ) {
			prevX = x;
			prevY = y;
			x = xVals.next();
			y = yVals.next();
		}
		if( Math.abs( clickedX - x ) > Math.abs( clickedX - prevX ) ) {
			x = prevX;
			y = prevY;
		}
		if( x == -1 ) {
			mouseX = -100;
			mouseY = -100;
			mouseStr = "";
			redX = -100;
			redY = -100;
			return;
		}
		redX = x;
		redY = y;
		mouseStr = "(" + x + ", " + doubleToString( y/100.0 ) + ")";
		repaint();
	}


	public void mouseReleased( MouseEvent e ) {
		mouseX = -100;
		mouseY = -100;
		mouseStr = "";
		redX = -100;
		redY = -100;
		repaint();
	}

	public void mouseClicked( MouseEvent e ) {}
	public void mouseEntered( MouseEvent e ) {}
	public void mouseExited( MouseEvent e ) {}
}
