/**
 * NodeDialog.java
 * Jul 17, 2000
 */

package AIspace.neural.dialogs;

import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.util.ArrayList;

import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;

import AIspace.graphToolKit.dialogs.BasicDialog;
import AIspace.graphToolKit.elements.Node;
import AIspace.neural.NeuralGraph;
import AIspace.neural.elements.NeuralEdge;
import AIspace.neural.elements.NeuralNode;

/**
 * dialog box for entering the node names and node heuristics
 * @author Leslie Tung & Shinjiro Sueda
 */
public class NodeDialog extends BasicDialog {

   //
   // Fields
   //
   
	public boolean isCancelled = true;
	private boolean isEditable;

	private JTextField textNodeName;
	private JTextField textParameters[];
	private JLabel labelError;

	private int nodeIndex;
	private NeuralGraph graph;
	private NeuralNode selected;
	private int numParents;
	private ArrayList<Node> parents;
	
	private JCheckBox sigmoid;
	private JCheckBox linear;
	private JCheckBox tanh;
	private JCheckBox exp;

	//
	// Methods
	//
	
	/**
	 * constructs the dialog box
	 *
	 * @param parent				 the SearchWindow
	 * @param g						the GraphCanvas
	 * @param selectedNodeIndex  the index of the node to which the
	 *									properties apply
	 * @param title				  the title of this dialog box
	 */
	public NodeDialog( JFrame parent, NeuralGraph g, int selectedNodeIndex, String title, boolean edit, NeuralNode node) {
		super(parent, title, true);
		graph = g;
		nodeIndex = selectedNodeIndex;
		selected = node;
		numParents = selected.getNumParentNodes();
		parents = selected.getParentNodes();
		isEditable = edit;

		String tmpString = new String(" ");
		int parameterIndex;

		String nodeName = selected.getLabel();
		if (nodeName.equals("")) {
			
			// Generate a unique node name.
			boolean nameFound = false;
			int nameCount = 0;
			nodeName = "Node "+nameCount;
			// Continue iterating through the nodes until no duplicate name exists.
			while((!nameFound)&&(graph.numNodes()>0))
			{
				for (int count = 0; count < graph.numNodes(); count++)
				{
					// Make sure a node with the same name does not already exist.
					if (graph.nodeAt(count).getLabel().equals(nodeName))
					{
						nameCount++;
						nodeName = "Node "+nameCount;
						break;
					}
					if (count == graph.numNodes()-1) nameFound = true;
				}
			}
		}
		textNodeName = new JTextField(nodeName, 20);
		textNodeName.setEditable(isEditable);
		textNodeName.addActionListener(this);

		JLabel[] labelParameters = new JLabel[numParents + 1];
		textParameters = new JTextField[numParents + 1];

		if (!isEditable) {
			parameterIndex = selected.getParameterIndex();
			if (parameterIndex > -1) {
				tmpString = new String(" w_" + parameterIndex + " ");
			}
		}
		if (numParents > 0) {
			labelParameters[0] = new JLabel("Parameter" + tmpString + ":");
		} else {
			labelParameters[0] = new JLabel("");
		}
		textParameters[0] = new JTextField((new Double(selected.getCurrentParaValue())).toString(), 20);
		textParameters[0].setEditable(isEditable);

		for (int i = 1; i < numParents + 1; i++) {
			tmpString = new String(" ");
			NeuralNode tempParentNode = (NeuralNode)parents.get(i - 1);
			nodeName = tempParentNode.getLabel();
			if (nodeName.equals("")) {
				nodeName = new String("Node " + tempParentNode.getIndex());
			}
			if (!isEditable) {
				NeuralEdge edge = (NeuralEdge) graph.getEdge(tempParentNode.getIndex(), selected.getIndex());
				if (edge != null) {
					parameterIndex = edge.getParameterIndex();
					if (parameterIndex != -1) {
						tmpString = new String(" w_" + parameterIndex + " ");
					}
				}
			}
			labelParameters[i] = new JLabel("Parameter" + tmpString + "(for " + nodeName + "):");
			NeuralEdge edge = (NeuralEdge) graph.getEdge(tempParentNode.getIndex(), nodeIndex);
			if (edge != null) {
				textParameters[i] = new JTextField((new Double(edge.getCurrentParaValue())).toString(), 20);
			} else {
				textParameters[i] = new JTextField("Error", 20);
			}
			textParameters[i].setEditable(isEditable);
		}

		JButton buttonOk = new JButton("OK");
		buttonOk.addActionListener(this);
		JButton buttonCancel = new JButton("Cancel");
		buttonCancel.addActionListener(this);
		labelError = new JLabel("");
		JPanel panel = new JPanel();
		if (numParents != 0)
		{
			panel.setLayout(new GridLayout(numParents + 5, 2));
		}
		else
		{
			panel.setLayout(new GridLayout(2, 2));
		}
		panel.add(new JLabel("Node name:"));
		panel.add(textNodeName);
		if (numParents > 0) {
			for (int i = 0; i < numParents + 1; i++) {
				panel.add(labelParameters[i]);
				panel.add(textParameters[i]);
			}
		}
		ButtonGroup functionGroup = new ButtonGroup();  
		
		setFunction();
		functionGroup.add(sigmoid);
		functionGroup.add(linear);
		functionGroup.add(tanh);
		functionGroup.add(exp);
	 
		if (numParents != 0)
		{
			panel.add(sigmoid);
			panel.add(linear);
			panel.add(tanh);
			panel.add(exp);
		}

		panel.add(buttonOk);
		panel.add(buttonCancel);
		getContentPane().add(panel);
		getContentPane().add("South", labelError);
		pack();
		centerWindow();
		setVisible(true);
	}

	private void setFunction() {
		sigmoid = new JCheckBox("Sigmoid Function");
		linear = new JCheckBox("Linear Function");
		tanh = new JCheckBox("Tanh Function");
		exp = new JCheckBox("Exponential Function");
		
		if(!isEditable) {
			sigmoid.setEnabled(false);
			linear.setEnabled(false);
			tanh.setEnabled(false);
			exp.setEnabled(false);
		}
		selectFunction();
	}
	
	private void selectFunction() {
		switch(selected.functionType) {
			case NeuralNode.SIGMOID:
				sigmoid.setSelected(true);
				break;
			case NeuralNode.LINEAR:
				linear.setSelected(true);
				break;
			case NeuralNode.TANH:
				tanh.setSelected(true);
				break;
			case NeuralNode.EXP:
				exp.setSelected(true);
				break;
		}
	}

	public void open(NeuralNode node) {
		isCancelled = true;
		selected = node;
		
		textNodeName.setText(selected.getLabel());
		selectFunction();
		setVisible( true );
	}
	
	/**
	 * set the node name and parameters in the node object and in the edges
	 * directed to it
	 */
	public boolean setProperties() {
		try {
			// Check the node name.
			if(textNodeName.getText().trim().equals("")) {
				graph.showMessage("Error in Name", "Name cannot be blank.");
				return false;
			}
			Double[] tempParaValues = new Double[numParents + 1];
			for (int i = 0; i < numParents + 1; i++) {
				tempParaValues[i] = Double.valueOf(textParameters[i].getText());
			}

			if (!selected.getLabel().equals(textNodeName.getText().trim())) {
				if (!graph.isNodeNameAllowed(textNodeName.getText().trim(), selected)) {
					graph.showMessage("Error in Name", "Name has already been used.");
					return false;
				}
			}
			selected.setLabel(textNodeName.getText().trim());
			((NeuralGraph) graph).rebuildData();
			selected.setInitialParaValue(tempParaValues[0].doubleValue());
			selected.setCurrentParaValue(tempParaValues[0].doubleValue());
					 
			for (int i = 1; i < numParents + 1; i++) {
				NeuralNode tempParentNode = (NeuralNode) parents.get(i - 1);
				NeuralEdge edge = (NeuralEdge) graph.getEdge(tempParentNode.getIndex(), nodeIndex);
				edge.setInitialParaValue(tempParaValues[i].doubleValue());
				edge.setCurrentParaValue(tempParaValues[i].doubleValue());
			}
			
			if (sigmoid.isSelected()) {
				selected.functionType = NeuralNode.SIGMOID;	
			} else if (linear.isSelected()) {
				selected.functionType = NeuralNode.LINEAR;	
			} else if (tanh.isSelected()) {
				selected.functionType = NeuralNode.TANH;	
			} else if (exp.isSelected()) {
				selected.functionType = NeuralNode.EXP;	
			}
			return true;
		} catch (NumberFormatException e) {
			labelError.setText("Error: Parameters have invalid real number format!");
			return false;
		}
	}
	
	public boolean isCancelled() {
		return isCancelled;
	}
	
	/** "OK" actionEvent from actionPerformed */
	protected boolean actionOK() {
		if(isEditable) {
			if(!setProperties()) {
				return false;
			}
			isCancelled = false;
		}
		return true;
	}
	
	/** "Cancel" actionEvent from actionPerformed */
	protected boolean actionCancel() {
		return true;
	}
	
	public void actionPerformed (ActionEvent e) {
		if (e.getSource().equals(textNodeName)) {
			textNodeName.transferFocus();
		} else {
			super.actionPerformed(e);
		}
	}
}