/**
 * TestStatisticsDialog.java
 * 
 * This class displays the results of testing all examples in the test set.
 * Examples are sorted into correctly identified, incorrectly identified, and
 * examples with no preditictions. Each of the three types are displayed in
 * a different text area.
 * 
 * @author Wesley Coelho, Joseph Roy Santos
 * 
 */

package AIspace.neural.dialogs;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.Hashtable;
import java.util.ArrayList;

import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.border.EmptyBorder;

import AIspace.graphToolKit.GraphConsts;
import AIspace.graphToolKit.dialogs.BasicDialog;
import AIspace.graphToolKit.elements.Node;
import AIspace.neural.ExampleList;
import AIspace.neural.NeuralGraph;
import AIspace.neural.elements.NeuralNode;

public class TestStatisticsDialog extends BasicDialog implements ItemListener, FocusListener {
	protected NeuralGraph graph;
	protected ExampleList exampleList;
	protected Font font;
	
	protected ArrayList<Integer> correctExamples = null;
	protected ArrayList<Integer> noPredictionExamples = null;
	protected ArrayList<Integer> incorrectExamples = null;
	
	protected JTextArea correctTextArea;
	protected JTextArea noPredictionTextArea;
	protected JTextArea incorrectTextArea;
	protected JTextField rangeField;
	protected JPanel centerPanel;
	protected JPanel mainPanel;
	protected JPanel bottomPanel;
	protected JLabel correct;
	protected JLabel incorrect;
	
	protected ButtonGroup outputSelection;

	private int currentItem;
	private double rangeFieldValue;

	/**
	 * Constructs and displays a new TestStatisticsDialog.
	 * First the data is collected, then the components are created and arranged.
	 * 
	 * @param parent - The 'window' object that is opening this dialog.
	 * @param g - the NeuralGraph object of the neural network
	 * 
	 */
	public TestStatisticsDialog(JFrame parent, NeuralGraph g) {
		super(parent, "Test Results", true);
		super.setCancelString("Close");
		graph = g; 
		rangeFieldValue = 0.5;
		exampleList = graph.getExampleList();
		
		font = new Font("courier", Font.PLAIN, 13);	       

		getData();
		
		mainPanel = makeMainPanel();
		mainPanel.setBorder(new EmptyBorder(3,3,3,3));
		getContentPane().add(mainPanel);
		currentItem = 0;	
			
		pack();
		centerWindow();
		setVisible(true);
	}


	/**
	 * Builds the user interface for this dialog.  The main panel is the 
	 * container for all other panels and components.
	 */
	protected JPanel makeMainPanel() {
		JPanel mainPanel = new JPanel(new BorderLayout());		
		
		mainPanel.add(BorderLayout.NORTH, makeCenterPanel());
		mainPanel.add(BorderLayout.SOUTH, makeBottomPanel());
		return mainPanel;
	}


	/**
	 * Builds the main display panel that contains the three text areas of
	 * examples.
	 */
	protected JPanel makeCenterPanel() {
		centerPanel = new JPanel();
		int panelCount = 0;
		
		if (correctTextArea.getText().length() < 1 && noPredictionTextArea.getText().length() < 1 && incorrectTextArea.getText().length() < 1) {
			centerPanel.add(new JLabel("There are no test examples"));
		} else {
			//Only add textAreas with data
			if (correctTextArea.getText().length() > 1) {
				JPanel correctPanel = new JPanel(new BorderLayout());
				JLabel correctLabel = new JLabel("Correctly Predicted Examples (" + correctExamples.size() + "):" );
				correctTextArea.setRows(10);
				correctTextArea.setFont(font);
				correctTextArea.setEditable(false);
				correctTextArea.setCursor(GraphConsts.DEFAULT_CURSOR);
				JScrollPane correctScrollPane = new JScrollPane(correctTextArea,JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
															 JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
	 
				
				correctPanel.add(BorderLayout.NORTH, correctLabel);
				correctPanel.add(BorderLayout.CENTER, correctScrollPane);
				centerPanel.add(correctPanel);	
				panelCount++;
			}
			if (noPredictionTextArea.getText().length() > 1) {
				JPanel noPredictionPanel = new JPanel(new BorderLayout());
				JLabel noPredictionLabel = new JLabel("Examples With No Prediction (" + noPredictionExamples.size() + "):" );
				noPredictionTextArea.setRows(10);
				noPredictionTextArea.setFont(font);
				noPredictionTextArea.setEditable(false);
				noPredictionTextArea.setCursor(GraphConsts.DEFAULT_CURSOR);
				
				JScrollPane noPredictionScrollPane = new JScrollPane(noPredictionTextArea,JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
																			 JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);				
				
				noPredictionPanel.add(BorderLayout.NORTH, noPredictionLabel);
				noPredictionPanel.add(BorderLayout.CENTER, noPredictionScrollPane);
				centerPanel.add(noPredictionPanel);
				panelCount++;	
			}
			if (incorrectTextArea.getText().length() > 1) {
				JPanel incorrectPanel = new JPanel(new BorderLayout());
				JLabel incorrectLabel = new JLabel("Incorrectly Predicted Examples (" + incorrectExamples.size() + "):" );
				incorrectTextArea.setRows(10);
				incorrectTextArea.setFont(font);
				incorrectTextArea.setEditable(false);
				incorrectTextArea.setCursor(GraphConsts.DEFAULT_CURSOR);
				JScrollPane incorrectScrollPane = new JScrollPane(incorrectTextArea,JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
																		 JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
				
				incorrectPanel.add(BorderLayout.NORTH, incorrectLabel);
				incorrectPanel.add(BorderLayout.CENTER, incorrectScrollPane);
				centerPanel.add(incorrectPanel);
				panelCount++;
			}
		}
		centerPanel.setLayout(new GridLayout(panelCount,1));
		return centerPanel;
	}

	/**
	 * updates the main display panel that contains the three text areas of
	 * examples.
	 */
	protected JPanel updateCenterPanel() {
		if (centerPanel == null) {
			centerPanel = new JPanel();
		} else {
			centerPanel.removeAll();	
		}
		
		int panelCount = 0;
		
		if (correctTextArea.getText().length() < 1 && noPredictionTextArea.getText().length() < 1 && incorrectTextArea.getText().length() < 1) {
			centerPanel.add(new JLabel("There are no test examples"));
		} else {
			//Only add textAreas with data
			if (correctTextArea.getText().length() > 1) {
				JPanel correctPanel = new JPanel(new BorderLayout());
				JLabel correctLabel = new JLabel("Correctly Predicted Examples (" + correctExamples.size() + "):" );
				correctTextArea.setRows(10);
				correctTextArea.setFont(font);
				correctTextArea.setEditable(false);
				correctTextArea.setCursor(GraphConsts.DEFAULT_CURSOR);
				correctPanel.add(BorderLayout.NORTH, correctLabel);
				correctPanel.add(BorderLayout.CENTER, correctTextArea);
				centerPanel.add(correctPanel);	
				panelCount++;
			}
			if (noPredictionTextArea.getText().length() > 1) {
				JPanel noPredictionPanel = new JPanel(new BorderLayout());
				JLabel noPredictionLabel = new JLabel("Examples With No Prediction (" + noPredictionExamples.size() + "):" );
				noPredictionTextArea.setRows(10);
				noPredictionTextArea.setFont(font);
				noPredictionTextArea.setEditable(false);
				noPredictionTextArea.setCursor(GraphConsts.DEFAULT_CURSOR);
				noPredictionPanel.add(BorderLayout.NORTH, noPredictionLabel);
				noPredictionPanel.add(BorderLayout.CENTER, noPredictionTextArea);
				centerPanel.add(noPredictionPanel);
				panelCount++;	
			}
			if (incorrectTextArea.getText().length() > 1) {
				JPanel incorrectPanel = new JPanel(new BorderLayout());
				JLabel incorrectLabel = new JLabel("Incorrectly Predicted Examples (" + incorrectExamples.size() + "):" );
				incorrectTextArea.setRows(10);
				incorrectTextArea.setFont(font);
				incorrectTextArea.setEditable(false);
				incorrectTextArea.setCursor(GraphConsts.DEFAULT_CURSOR);
				incorrectPanel.add(BorderLayout.NORTH, incorrectLabel);
				incorrectPanel.add(BorderLayout.CENTER, incorrectTextArea);
				centerPanel.add(incorrectPanel);
				panelCount++;
			}
		}
		centerPanel.setLayout(new GridLayout(panelCount,1));
		return centerPanel;
	}
	
	/**
	 * This method updates the lower panel of the dialog window that contains
	 * the percentage of examples that are correct or incorrect and the
	 * close button.
	 */
	protected void updateBottomLabels() {
		int numCorrect = 0;
		if (correctExamples != null) {
			numCorrect += correctExamples.size();
		}
		
		int numIncorrect = 0;
		if (exampleList.getTestArrayList().size() > 0) {
			numCorrect = numCorrect*100/(exampleList.getTestArrayList().size());
			numIncorrect = 100 - numCorrect;
		}
		
      	correct.setText("Predicted Correctly: "+ numCorrect + "%" );
      	incorrect.setText("Predicted Incorrectly: "+ numIncorrect +"%");		
	}

	/**
	 * Fills a TextArea with exaples identified by a offsetPoint of indices.
	 * This method is a bit on the nasty side after undergoing some extensions...
	 * It should be simplified.
	 * PRE: targetTextArea has been initialized.
	 * 
	 * @param targetTextArea		The textArea object to fill with examples
	 * @param selectedIndices		The indices of examples to add to the textArea
	 * @param allExamples			The data source for the examples (ArrayList of Hash Tables) 
	 * @param showPrediction		Shows the predicted value if true
	 * 
	 */
	private void fillTextArea(JTextArea targetTextArea, ArrayList<Integer> selectedIndices, ArrayList<Hashtable<Integer, String>> allExamples, boolean showPrediction) {
		targetTextArea.setText("");
		int maxColWidth = 0; // Used to set the size of the text area.
				
		//Build Title String, return if no data
		String heading = new String("");
		if (selectedIndices == null || selectedIndices.size() < 1 || exampleList.getNumParameters() < 2) {
			return;
		}		
		
		int numParameters = exampleList.getNumParameters();
		//Add a spot for the predicted value
		if (showPrediction) {
			numParameters += 1;	
		}
		
		//Build heading String
		String[] parameters = exampleList.getParameters();
		
		if (showPrediction) {
			String newParameters[] = new String[parameters.length + 1];
			for (int i = 0; i < parameters.length; i++) {
				newParameters[i] = parameters[i];	
			}
			newParameters[newParameters.length - 1] = "Predicted Value";
			parameters = newParameters;
		}
		
		
		int[] colPos = new int[parameters.length]; // Record starting positions of each column
		for (int i = 0; i < numParameters; i++) {
			colPos[i] = heading.length(); //Record current position
			heading += parameters[i];
			for (int j = 0; j < 5; j++) {
				heading += " ";
			}
		}
		targetTextArea.setText( targetTextArea.getText() + heading); 
		
		
		//Add Example Rows      
		for (int i = 0; i < selectedIndices.size(); i++) {
			String display = new String("");
			Hashtable<Integer, String> oneExample = allExamples.get( selectedIndices.get(i).intValue()  );
			for (int j = 0; j < numParameters; j++) {
				String value = "";
				
				if (showPrediction && j == numParameters - 1) {
					//Add the predicted value (Get it from the graph)
					
					//Reformat to call the search method
					String[] values = new String[oneExample.size()];
					for (int k = 0; k < oneExample.size(); k++) {
						values[k] = oneExample.get((new Integer(k)));
					}
					
					value = graph.outputValue(values,currentItem);
				} else {
					value = oneExample.get((new Integer(j)));
				}

				
				if (value != null) {
					while (display.length() < colPos[j]) {
						display += " ";
					}
					display += value;
				}
			}
			if (maxColWidth < display.length()) 
				maxColWidth = heading.length();
			targetTextArea.setText( targetTextArea.getText() + "\n" + display);
		}
		
		targetTextArea.setColumns(maxColWidth);
		targetTextArea.setRows(selectedIndices.size() + 1);
	}


	/**
	 * This method builds the lower panel of the dialog window that contains
	 * the percentage of examples that are correct or incorrect and the
	 * close button.
	 */
	protected JPanel makeBottomPanel() {
		bottomPanel = new JPanel(new BorderLayout());
		JPanel buttonPanel = new JPanel();
		JPanel percentagePanel = new JPanel(new GridLayout(3,2));
		
		int numCorrect = 0;
		if (correctExamples != null) {
			numCorrect += correctExamples.size();
		}
		
		int numIncorrect = 0;
		if (exampleList.getTestArrayList().size() > 0) {
			numCorrect = numCorrect*100/(exampleList.getTestArrayList().size());
			numIncorrect = 100 - numCorrect;
		}
		
      	correct = new JLabel("Predicted Correctly: "+ numCorrect + "%" );
      	incorrect = new JLabel("Predicted Incorrectly: "+ numIncorrect +"%");
      	
      	JPanel rangePanel = new JPanel(new FlowLayout(FlowLayout.LEFT,0,0));
      	JLabel rangeLabel = new JLabel("Input range threshold of classification: ");
      	rangePanel.add(rangeLabel);
      	rangeField = new JTextField("0.5");
      	rangeField.setName("rangeField");
      	rangeField.setColumns(3);
      	rangeField.addFocusListener(this);
      	rangePanel.add(rangeField);
		rangePanel.add(new JLabel("  ")); //filler
      			
		
		percentagePanel.add(rangePanel); //filler
		percentagePanel.add(new JLabel("")); //filler
		percentagePanel.add(correct);
		percentagePanel.add(new JLabel("")); //filler
		percentagePanel.add(incorrect);
		percentagePanel.add(new JLabel("")); //filler
		
		JButton closeButton = new JButton("Close");
		closeButton.addActionListener(this);
		buttonPanel.add(closeButton);
		
		JPanel selectPanel = new JPanel(new BorderLayout());
		JPanel checkboxPanel = new JPanel(new FlowLayout());
		JLabel header = new JLabel("Select an output to analyze:");
		header.setAlignmentX(JLabel.CENTER);
		
		selectPanel.add(BorderLayout.NORTH, header);
		
		JCheckBox outButton;
		ArrayList<Node> outputIndexList = graph.getOutputNodes();
		String nodeName = "";
		
		outputSelection = new ButtonGroup();
		for (int i = 0; i < outputIndexList.size(); i++) {			
			nodeName = ((NeuralNode)outputIndexList.get(i)).getLabel();
			if (i == 0) {
				outButton = new JCheckBox(nodeName,  true);	
				outputSelection.add(outButton);
			} else {
				outButton = new JCheckBox(nodeName, false);		
				outputSelection.add(outButton);		
			}
			outButton.setName(Integer.toString(i));
			outButton.addItemListener(this);
			checkboxPanel.add(outButton);
		}
		selectPanel.add(BorderLayout.CENTER, checkboxPanel);
		
		bottomPanel.add(BorderLayout.NORTH, percentagePanel);
		bottomPanel.add(BorderLayout.CENTER, selectPanel);
		bottomPanel.add(BorderLayout.SOUTH, buttonPanel);
		return bottomPanel;
	}

	/**
	 * This method collects the indices examples to display and stores them
	 * in the class-scope vectors (eg correctExamples). The method to fill the 
	 * text areas is also called here.
	 * 
	 */
	protected void getData() {
		correctExamples = graph.getTestExampleIndices(NeuralGraph.CORRECT, rangeFieldValue, currentItem);
		noPredictionExamples = graph.getTestExampleIndices(NeuralGraph.NO_PREDICTION, rangeFieldValue, currentItem);
		incorrectExamples = graph.getTestExampleIndices(NeuralGraph.INCORRECT, rangeFieldValue, currentItem);
		
		correctTextArea = new JTextArea("",5,5);
		noPredictionTextArea = new JTextArea("",5,5);
		incorrectTextArea = new JTextArea("",5,5);
		
		fillTextArea(correctTextArea, correctExamples, exampleList.getTestArrayList(), true);
		fillTextArea(noPredictionTextArea, noPredictionExamples, exampleList.getTestArrayList(), true);
		fillTextArea(incorrectTextArea, incorrectExamples, exampleList.getTestArrayList(), true);
	}

	/** "OK" actionEvent from actionPerformed */
	protected boolean actionOK() {
		return true;
	}
	
	/** "Close" actionEvent from actionPerformed */
	protected boolean actionCancel() {
		return true;
	}
	
	/**
	 * 	Implementation of ItemListener.
	 * 	Keeps track of the checkboxes for output
	 */
	public void itemStateChanged(ItemEvent e) {
		// find the outputNode index
		// it is set to the name of the selected checkbox
		JCheckBox current = (JCheckBox)e.getItemSelectable();		
		String indexName = current.getName();
		int index = Integer.parseInt(indexName);
		// refresh the examples displayed to the new selection
		// if it is different from the original selection
		if (index != currentItem) {
			currentItem = index;	
			getData();
				
			updateCenterPanel();
			updateBottomLabels();
			
			validate();
		}	
	}
	
	/**
	 * Implementation of FocusListener.
	 * This saves the value of rangeField.
	 */
	public void focusGained(FocusEvent e) {
		if (e.getComponent().getName().equals("rangeField")) {	
			try
			{
				rangeFieldValue = Double.parseDouble(rangeField.getText());
			}
			catch(NumberFormatException nfe)
			{}
		}
	}
	
	/**
	 * Implementation of focusListener. This updates
	 * the TextAreas if the value in rangeField has changed.
	 */
	public void focusLost(FocusEvent e) {
		if (e.getComponent().getName().equals("rangeField")) {	
			try
			{
				double range = Double.parseDouble(rangeField.getText());
				if (rangeFieldValue != range) {
					rangeFieldValue = range;
					getData();
					updateCenterPanel();
					updateBottomLabels();
					validate();
					pack();
				}
			}
			catch(NumberFormatException nfe)
			{}
		}				
	}
}