/**
 * NeuralEdge.java
 * Jul 26, 2000
 */


package AIspace.neural.elements;


import java.awt.Color;
import java.awt.Graphics;

import AIspace.graphToolKit.GraphConsts;
import AIspace.graphToolKit.elements.Edge;
import AIspace.graphToolKit.elements.Node;
import AIspace.neural.BackPropagation;
import AIspace.neural.NeuralGraph;
import AIspace.neural.PlotCanvas;


/**
 * NeuralEdge subclasses CIEdge, implementing functionality for
 * edges which is specific to the neural network applet.
 *
 * @author  Leslie Tung & Shinjiro Sueda
 * @version 1.1
 * @since JDK1.1
 */
public class NeuralEdge extends Edge
{

   //
   // Fields
   //
   
   private double initialParaValue;
   private double currentParaValue;
   private int parameterIndex;


   //
   // Methods
   //

   /**
    * Initializes a new NeuralEdge
    *
    * @param tail The first NeuralNode of the edge (the arc points away
    *             from it)
    * @param head The second NeuralNode of the edge (the arc points to it)
    * @param from The edge object which this edge will be constructed from
    */
   public NeuralEdge(NeuralGraph graph, Node tail, Node head)
   {
	  super(graph, tail, head);
	  if (graph.getInitParamType())
	  {
	      Double tempValue = 2*graph.getInitParamBound()*Math.random()-graph.getInitParamBound();
	      initialParaValue = tempValue;
	      currentParaValue = tempValue;
	  }
	  else
	  {
		  initialParaValue = graph.getInitParamValue();
		  currentParaValue = graph.getInitParamValue();
	  }

	  edgeType = GraphConsts.DIRECTIONAL;
      parameterIndex = -1;
      setLabel( ((NeuralGraph)graph).showParam );
   }
   

   /**
    * access the initial parameter value
    */
   public double getInitialParaValue()
   {
      return initialParaValue;
   }

   
   /**
    * set the initial parameter value
    */
   public void setInitialParaValue(double value)
   {
      initialParaValue = value;
   }

   
   /**
    * access the current parameter value
    */
   public double getCurrentParaValue()
   {
      return currentParaValue;
   }

   
   /**
    * set the current parameter value
    */
   public void setCurrentParaValue(double value)
   {
      currentParaValue = value;
      setLabel( ((NeuralGraph)graph).showParam );
   }

   
   /**
    * set the parameter index of this edge, which is the parameter
    * of the head node from the tail node.
    */
   public void setParameterIndex(int index)
   {
      parameterIndex = index;
   }

   
   /**
    * access the parameter index of this edge
    */
   public int getParameterIndex()
   {
      return parameterIndex;
   }


   /**
    * Sets the label to be the current parameter value.
    */
   public void setLabel( boolean showParam )
   {
      if( showParam )
      {
         setLabel( PlotCanvas.doubleToString( currentParaValue ) );
      }
      else setLabel( "" );
   }

   
   /**
    * draw edges specific to neural network
    */
   public void draw(Graphics offscreen, boolean moving)
   {
      // change the colours according to the parameters
      // (only in problem solution mode)
      if( graph.isInSolveMode() )
      {
         float r_comp = (float)1.0;
         float g_comp = (float)1.0;
         float b_comp = (float)1.0;
         double value = BackPropagation.sigmoidFunction(
            currentParaValue * 0.5);
         if (value < 0.5)
         {
            g_comp = (float) (value * 2);
            b_comp = (float) (1.0 - ((0.5 - value) * 2));
         }
         else
         {
            r_comp = (float) ((1.0 - value) * 2);
            b_comp = (float) (1.0 - ((value - 0.5) * 2));
         }
         color = new Color (r_comp, g_comp, b_comp);
      }
      else color = Color.black;
      super.draw( offscreen, moving );
   }
}
