/**
 * NeuralNode.java
 * Jul 26, 2000
 */


package AIspace.neural.elements;

import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.util.ArrayList;

import AIspace.graphToolKit.GraphConsts;
import AIspace.graphToolKit.elements.Node;
import AIspace.neural.BackPropagation;
import AIspace.neural.NeuralGraph;
import AIspace.neural.PlotCanvas;


/**
 * NeuralNode subclasses CINode, implementing functionality for
 * nodes which is specific to the neural network applet.
 *
 * @author  Leslie Tung & Shinjiro Sueda
 * @version 1.1
 * @since JDK1.1
 */
public class NeuralNode extends Node
{
	// static constants
	public static final int SIGMOID = 0;
	public static final int LINEAR = 1;
	public static final int TANH = 2;
	public static final int EXP = 3;


	//
	// Fields
	//
	
	private ArrayList<Node> parents;
	private ArrayList<Node> children;
	private double initialParaValue;
	private double currentParaValue;
	private int parameterIndex;
	private boolean coloring;
	public int functionType;
	public double[] currentError;
	private double currOutput;


	//
	// Methods
	//

	/**
 	* Initializes a new NeuralNode
 	*/
	public NeuralNode( NeuralGraph graph ) {
		super( graph );
		shape = GraphConsts.RECT;
		wPad = 10;
		hPad = 10;
		parents = new ArrayList<Node>(5);
		children = new ArrayList<Node>(5);
		if (graph.getInitParamType())	
		{
			Double tempValue = 2*graph.getInitParamBound()*Math.random()-graph.getInitParamBound();
		    initialParaValue = tempValue;
		    currentParaValue = tempValue;
		}	
		else
		{
			initialParaValue = graph.getInitParamValue();
			currentParaValue = graph.getInitParamValue();
		}
		parameterIndex = -1;
		coloring = false;
		functionType = NeuralNode.SIGMOID;
		currentError = null;
		currOutput = 0;
	}
	
	
	/**
 	* Accesses the current output of this node.
 	* This is for efficiency in backpropagation
 	*/
	public double getOutputValue() {
		return currOutput;
	}
	
	/**
 	* Sets the current output of this node. 
 	* This is for backprop efficiency.
 	*/
	public void setOutputValue(double out) {
		currOutput = out;
	}

	/**
	 * access the initial parameter value
	 */
	public double getInitialParaValue() {
		return initialParaValue;
	}

	
	/**
	 * set the initial parameter value
	 */
	public void setInitialParaValue(double value) {
		initialParaValue = value;
	}

	
	/**
	 * access the current parameter value
	 */
	public double getCurrentParaValue() {
		return currentParaValue;
	}

	
	/**
	 * set the current parameter value
	 */
	public void setCurrentParaValue(double value) {
		currentParaValue = value;
		setLabel( ((NeuralGraph)graph).showParam );
	}

	
	/**
	 * get the parents of this node
	 */
	public ArrayList<Node> getParentNodes() {
		return parents;
	}

	
	/**
	 * get the children of this node
	 */
	public ArrayList<Node> getChildrenNodes() {
		return children;
	}

	
	/**
	 * get the number of parents of this node
	 */
	public int getNumParentNodes() {
		return parents.size();
	}

	
	/**
	 * get the number of children of this node
	 */
	public int getNumChildrenNodes() {
		return children.size();
	}

	
	/**
	 * add the node to be a parent of this node
	 */
	public void addParent(Node node) {
		if (!parents.contains(new Integer(index))) {
			parents.add(node);
		}
	}

	
	/**
	 * add the node to be a child of this node
	 */
	public void addChild(Node node) {
		if (!children.contains(new Integer(index))) {
			children.add(node);
		}
	}

	
	/**
	 * remove the node as a parent of this node
	 */
	public void removeParent(Node node) {
		parents.remove(node);
	}

	
	/**
	 * remove the node as a child of this node
	 */
	public void removeChild(Node node) {
		children.remove(node);
	}

	
	/**
	 * set the parameter index of this node
	 */
	public void setParameterIndex(int index) {
		parameterIndex = index;
	}

	
	/**
	 * access the parameter index of this node
	 */
	public int getParameterIndex() {
		return parameterIndex;
	}

	
	public void setColoring(boolean value) {
		coloring = value;
	}


	public void setLabel( boolean showParam ) {
		if( showParam && parents.size()!=0 ) {
			label[1] = PlotCanvas.doubleToString( currentParaValue );
		} else {
			label[1] = "";
		}

		updateSize();
	}


	/**
	 * Overrides getWidth() in Node.
	 */
	protected int getWidth() {
		FontMetrics fm = graph.canvas.getFontMetrics();
		try {
			if( label[1].equals( "" ) ) {
				return fm.stringWidth( label[0] ) + wPad;
			}
			int width1, width2;
			if( label[0] == null || label[0].equals( "" ) ) 
				width1 = 0;
			else 
				width1 = fm.stringWidth( label[0] ) + wPad;
			
			if( label[1] == null || label[1].equals( "" ) ) 
				width2 = 0;
			else 
				width2 = fm.stringWidth( label[1] ) + wPad;
			
			return Math.max( width1, width2 );
		} catch( Exception e ) {
			System.out.println( label[0] + ", " + label[1] );
		}
		return 10;
	}
	

	/**
	 * Overrides getHeight() in Node.
	 */
	protected int getHeight() {
		int height = graph.canvas.getFontMetrics().getHeight() + hPad; 
		if( !label[1].equals( "" ) ) {
			height *= 2;
		}
		return height;
	}

	
	public void draw( Graphics offscreen, boolean moving ) {
		if( graph.isInSolveMode() && coloring ) {
			// change the colours according to the parameters
			// (only in problem solution mode)
			float r_comp = (float)1.0;
			float g_comp = (float)1.0;
			float b_comp = (float)1.0;
			double value = BackPropagation.sigmoidFunction(currentParaValue * 0.5);
			if (value < 0.5) {
				g_comp = (float) (value * 2);
				b_comp = (float) (1.0 - ((0.5 - value) * 2));
			} else {
				r_comp = (float) ((1.0 - value) * 2);
				b_comp = (float) (1.0 - ((value - 0.5) * 2));
			}
			Color newColor = new Color (r_comp, g_comp, b_comp);
			color = newColor;
		} else {
			color = Color.black;
		}
		super.draw( offscreen, moving );
	}

	
	/**
	 * Overrides drawLabel() in Node.
	 */
	protected void drawLabel( Graphics offscreen ) {
		FontMetrics fm = graph.canvas.getFontMetrics();
		int h = fm.getAscent();
		int width1 = fm.stringWidth( label[0] ) + wPad;
		int width2 = fm.stringWidth( label[1] ) + wPad;
		if( isBold && color==Color.black) {
			offscreen.setColor( Color.white );
		} else {
			offscreen.setColor( Color.black );
		}
		if( label[1].equals( "" ) ) {
			offscreen.drawString( label[0], (int)pos.x - (width1 - wPad)/2, (int)pos.y + h/2 );
		} else {
			offscreen.drawString( label[0], (int)pos.x - (width1 - wPad)/2, (int)pos.y - (height - hPad)/4 + h/2 );
			offscreen.drawString( label[1], (int)pos.x - (width2 - wPad)/2, (int)pos.y + (height - hPad)/4 + h/2 );
		}
	}


	public String toString() {
		return index + ": " + label[0];
	}
}
